/* Copyright 2019 Luca Fedeli, Maxence Thevenet
 *
 * This file is part of WarpX.
 *
 * License: BSD-3-Clause-LBNL
 */
#include "BreitWheelerEngineWrapper.H"

#include "Utils/TextMsg.H"

#include <AMReX.H>
#include <AMReX_BLassert.H>
#include <AMReX_GpuDevice.H>

#include <picsar_qed/physics/breit_wheeler/breit_wheeler_engine_tables.hpp>
//Functions needed to generate a new table
#ifdef WARPX_QED_TABLE_GEN
#   include <picsar_qed/physics/breit_wheeler/breit_wheeler_engine_tables_generator.hpp>
#endif
#include <picsar_qed/utils/serialization.hpp>

#include <algorithm>
#include <cmath>
#include <cstdint>
#include <initializer_list>
#include <iosfwd>
#include <vector>

using namespace std;
using namespace amrex;
namespace pxr_sr = picsar::multi_physics::utils::serialization;

//This file provides a wrapper around the breit_wheeler engine
//provided by the PICSAR library

// Factory class =============================

BreitWheelerGetOpticalDepth
BreitWheelerEngine::build_optical_depth_functor () const
{
    return {};
}

BreitWheelerEvolveOpticalDepth
BreitWheelerEngine::build_evolve_functor () const
{
    AMREX_ALWAYS_ASSERT(m_lookup_tables_initialized);

    return {m_dndt_table.get_view(), m_bw_minimum_chi_phot};
}

BreitWheelerGeneratePairs
BreitWheelerEngine::build_pair_functor () const
{
    AMREX_ALWAYS_ASSERT(m_lookup_tables_initialized);

    return {m_pair_prod_table.get_view()};
}

bool BreitWheelerEngine::are_lookup_tables_initialized () const
{
    return m_lookup_tables_initialized;
}

bool
BreitWheelerEngine::init_lookup_tables_from_raw_data (
    const vector<char>& raw_data,
    const amrex::ParticleReal bw_minimum_chi_phot)
{
    auto raw_iter = raw_data.begin();
    const auto size_first = pxr_sr::get_out<uint64_t>(raw_iter);
    if(size_first <= 0 || size_first >= raw_data.size() ) { return false; }

    const auto raw_dndt_table = vector<char>{
        raw_iter, raw_iter+static_cast<long>(size_first)};

    const auto raw_pair_prod_table = vector<char>{
        raw_iter+static_cast<long>(size_first), raw_data.end()};

    m_dndt_table = BW_dndt_table{raw_dndt_table};
    m_pair_prod_table = BW_pair_prod_table{raw_pair_prod_table};

    if (!m_dndt_table.is_init() || !m_pair_prod_table.is_init()) {
        return false;
    }

    m_bw_minimum_chi_phot = bw_minimum_chi_phot;

    amrex::Gpu::synchronize();

    m_lookup_tables_initialized = true;

    return true;
}

void BreitWheelerEngine::init_builtin_tables(
    const amrex::ParticleReal bw_minimum_chi_phot)
{
    init_builtin_dndt_table();
    init_builtin_pair_prod_table();
    m_bw_minimum_chi_phot = bw_minimum_chi_phot;

    m_lookup_tables_initialized = true;
}

vector<char> BreitWheelerEngine::export_lookup_tables_data () const
{
    if(!m_lookup_tables_initialized) {
        return vector<char>{};
    }

    const auto data_dndt = m_dndt_table.serialize();
    const auto data_pair_prod = m_pair_prod_table.serialize();

    const uint64_t size_first = data_dndt.size();

    vector<char> res{};
    pxr_sr::put_in(size_first, res);
    for (const auto& tmp : data_dndt) {
        pxr_sr::put_in(tmp, res);
    }
    for (const auto& tmp : data_pair_prod) {
        pxr_sr::put_in(tmp, res);
    }

    return res;
}

PicsarBreitWheelerCtrl
BreitWheelerEngine::get_default_ctrl() const
{
    namespace pxr_bw = picsar::multi_physics::phys::breit_wheeler;
    return PicsarBreitWheelerCtrl{
        pxr_bw::default_dndt_lookup_table_params<amrex::ParticleReal>,
        pxr_bw::default_pair_prod_lookup_table_params<amrex::ParticleReal>
    };
}

amrex::ParticleReal
BreitWheelerEngine::get_minimum_chi_phot() const
{
    return m_bw_minimum_chi_phot;
}

void BreitWheelerEngine::compute_lookup_tables (
    PicsarBreitWheelerCtrl ctrl,
    const amrex::ParticleReal bw_minimum_chi_phot)
{
#ifdef WARPX_QED_TABLE_GEN
    m_dndt_table = BW_dndt_table{ctrl.dndt_params};
    m_dndt_table.generate(true); //Progress bar is displayed
    m_pair_prod_table = BW_pair_prod_table{ctrl.pair_prod_params};
    m_pair_prod_table.generate(true); //Progress bar is displayed
    m_bw_minimum_chi_phot = bw_minimum_chi_phot;

    amrex::Gpu::synchronize();

    m_lookup_tables_initialized = true;
#else
    amrex::ignore_unused(ctrl, bw_minimum_chi_phot);
    WARPX_ABORT_WITH_MESSAGE("WarpX was not compiled with table generation support!");
#endif
}

void BreitWheelerEngine::init_builtin_dndt_table()
{
    constexpr auto default_chi_phot_min = 0.02_prt;
    constexpr auto default_chi_phot_max = 200.0_prt;
    constexpr auto default_chi_phot_how_many = 64;

    BW_dndt_table_params dndt_params;
    dndt_params.chi_phot_min = default_chi_phot_min;
    dndt_params.chi_phot_max = default_chi_phot_max;
    dndt_params.chi_phot_how_many = default_chi_phot_how_many;

    const auto vals = std::vector<amrex::ParticleReal>{
        -1.34808e+02_prt, -1.16674e+02_prt, -1.01006e+02_prt, -8.74694e+01_prt,
        -7.57742e+01_prt, -6.56699e+01_prt, -5.69401e+01_prt, -4.93981e+01_prt,
        -4.28821e+01_prt, -3.72529e+01_prt, -3.23897e+01_prt, -2.81885e+01_prt,
        -2.45593e+01_prt, -2.14243e+01_prt, -1.87166e+01_prt, -1.63779e+01_prt,
        -1.43584e+01_prt, -1.26145e+01_prt, -1.11091e+01_prt, -9.80975e+00_prt,
        -8.68862e+00_prt, -7.72163e+00_prt, -6.88795e+00_prt, -6.16965e+00_prt,
        -5.55120e+00_prt, -5.01921e+00_prt, -4.56210e+00_prt, -4.16989e+00_prt,
        -3.83395e+00_prt, -3.54680e+00_prt, -3.30202e+00_prt, -3.09403e+00_prt,
        -2.91801e+00_prt, -2.76980e+00_prt, -2.64579e+00_prt, -2.54287e+00_prt,
        -2.45833e+00_prt, -2.38982e+00_prt, -2.33532e+00_prt, -2.29307e+00_prt,
        -2.26152e+00_prt, -2.23935e+00_prt, -2.22540e+00_prt, -2.21867e+00_prt,
        -2.21827e+00_prt, -2.22343e+00_prt, -2.23350e+00_prt, -2.24788e+00_prt,
        -2.26607e+00_prt, -2.28761e+00_prt, -2.31212e+00_prt, -2.33926e+00_prt,
        -2.36871e+00_prt, -2.40023e+00_prt, -2.43357e+00_prt, -2.46853e+00_prt,
        -2.50493e+00_prt, -2.54261e+00_prt, -2.58143e+00_prt, -2.62127e+00_prt,
        -2.66201e+00_prt, -2.70357e+00_prt, -2.74585e+00_prt, -2.78877e+00_prt};

    m_dndt_table = BW_dndt_table{dndt_params, vals};
}

void BreitWheelerEngine::init_builtin_pair_prod_table()
{
    constexpr auto default_chi_phot_min = 0.02_prt;
    constexpr auto default_chi_phot_max = 200.0_prt;
    constexpr auto default_chi_phot_how_many = 64;
    constexpr auto default_frac_how_many = 64;

    BW_pair_prod_table_params pair_prod_params;
    pair_prod_params.chi_phot_min = default_chi_phot_min;
    pair_prod_params.chi_phot_max = default_chi_phot_max;
    pair_prod_params.chi_phot_how_many = default_chi_phot_how_many;
    pair_prod_params.frac_how_many = default_frac_how_many;

    const auto vals = std::vector<amrex::ParticleReal>{
        0.00000e+00_prt, 0.00000e+00_prt, 0.00000e+00_prt, 0.00000e+00_prt,
        0.00000e+00_prt, 0.00000e+00_prt, 0.00000e+00_prt, 3.35120e-221_prt,
        1.13067e-188_prt, 2.14228e-163_prt, 3.39948e-143_prt, 1.09215e-126_prt,
        5.94019e-113_prt, 2.37985e-101_prt, 2.00771e-91_prt, 7.70540e-83_prt,
        2.39749e-75_prt, 9.40741e-69_prt, 6.56422e-63_prt, 1.06836e-57_prt,
        5.03881e-53_prt, 8.20929e-49_prt, 5.33414e-45_prt, 1.55651e-41_prt,
        2.25171e-38_prt, 1.75472e-35_prt, 7.90210e-33_prt, 2.18314e-30_prt,
        3.89470e-28_prt, 4.68899e-26_prt, 3.95815e-24_prt, 2.42185e-22_prt,
        1.10577e-20_prt, 3.86465e-19_prt, 1.05740e-17_prt, 2.31035e-16_prt,
        4.10266e-15_prt, 6.01463e-14_prt, 7.38207e-13_prt, 7.68074e-12_prt,
        6.85101e-11_prt, 5.29188e-10_prt, 3.57202e-09_prt, 2.12435e-08_prt,
        1.12141e-07_prt, 5.28995e-07_prt, 2.24354e-06_prt, 8.60245e-06_prt,
        2.99719e-05_prt, 9.53273e-05_prt, 2.77953e-04_prt, 7.45878e-04_prt,
        1.84871e-03_prt, 4.24642e-03_prt, 9.06751e-03_prt, 1.80527e-02_prt,
        3.36054e-02_prt, 5.86505e-02_prt, 9.62278e-02_prt, 1.48823e-01_prt,
        2.17569e-01_prt, 3.01546e-01_prt, 3.97478e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 0.00000e+00_prt, 0.00000e+00_prt, 0.00000e+00_prt,
        0.00000e+00_prt, 0.00000e+00_prt, 4.96543e-229_prt, 1.61201e-191_prt,
        2.10361e-163_prt, 1.48853e-141_prt, 4.31513e-124_prt, 8.03657e-110_prt,
        6.02583e-98_prt, 6.48154e-88_prt, 2.47897e-79_prt, 6.56015e-72_prt,
        1.97905e-65_prt, 9.97097e-60_prt, 1.12913e-54_prt, 3.63334e-50_prt,
        4.00764e-46_prt, 1.76373e-42_prt, 3.50658e-39_prt, 3.48978e-36_prt,
        1.89363e-33_prt, 6.01935e-31_prt, 1.19104e-28_prt, 1.54482e-26_prt,
        1.37287e-24_prt, 8.68465e-23_prt, 4.04195e-21_prt, 1.42438e-19_prt,
        3.89736e-18_prt, 8.46436e-17_prt, 1.48776e-15_prt, 2.15299e-14_prt,
        2.60458e-13_prt, 2.66996e-12_prt, 2.34743e-11_prt, 1.78939e-10_prt,
        1.19414e-09_prt, 7.03769e-09_prt, 3.69190e-08_prt, 1.73618e-07_prt,
        7.36647e-07_prt, 2.83643e-06_prt, 9.96391e-06_prt, 3.20868e-05_prt,
        9.51418e-05_prt, 2.60801e-04_prt, 6.63347e-04_prt, 1.57087e-03_prt,
        3.47431e-03_prt, 7.19766e-03_prt, 1.40053e-02_prt, 2.56621e-02_prt,
        4.43869e-02_prt, 7.26462e-02_prt, 1.12767e-01_prt, 1.66410e-01_prt,
        2.34012e-01_prt, 3.14373e-01_prt, 4.04524e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 0.00000e+00_prt, 0.00000e+00_prt, 0.00000e+00_prt,
        0.00000e+00_prt, 1.05570e-243_prt, 2.71557e-198_prt, 7.19045e-166_prt,
        1.44533e-141_prt, 1.11742e-122_prt, 1.39503e-107_prt, 3.03452e-95_prt,
        5.61714e-85_prt, 2.65318e-76_prt, 7.00707e-69_prt, 1.83945e-62_prt,
        7.38996e-57_prt, 6.32013e-52_prt, 1.48739e-47_prt, 1.17966e-43_prt,
        3.70790e-40_prt, 5.26675e-37_prt, 3.76381e-34_prt, 1.47876e-31_prt,
        3.43911e-29_prt, 5.03750e-27_prt, 4.89780e-25_prt, 3.30531e-23_prt,
        1.60870e-21_prt, 5.83599e-20_prt, 1.62380e-18_prt, 3.55242e-17_prt,
        6.24487e-16_prt, 8.99102e-15_prt, 1.07814e-13_prt, 1.09288e-12_prt,
        9.48909e-12_prt, 7.14037e-11_prt, 4.70552e-10_prt, 2.74127e-09_prt,
        1.42365e-08_prt, 6.64113e-08_prt, 2.80174e-07_prt, 1.07555e-06_prt,
        3.77812e-06_prt, 1.22054e-05_prt, 3.64294e-05_prt, 1.00878e-04_prt,
        2.60162e-04_prt, 6.27062e-04_prt, 1.41706e-03_prt, 3.01137e-03_prt,
        6.03424e-03_prt, 1.14306e-02_prt, 2.05179e-02_prt, 3.49780e-02_prt,
        5.67520e-02_prt, 8.78195e-02_prt, 1.29868e-01_prt, 1.83904e-01_prt,
        2.49888e-01_prt, 3.26495e-01_prt, 4.11098e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 0.00000e+00_prt, 0.00000e+00_prt, 0.00000e+00_prt,
        7.26024e-270_prt, 5.56502e-211_prt, 9.89467e-172_prt, 1.05165e-143_prt,
        1.07452e-122_prt, 2.29464e-106_prt, 2.58816e-93_prt, 1.20441e-82_prt,
        9.10565e-74_prt, 2.88914e-66_prt, 7.57859e-60_prt, 2.70227e-54_prt,
        1.90184e-49_prt, 3.51404e-45_prt, 2.12810e-41_prt, 5.03275e-38_prt,
        5.34704e-35_prt, 2.85882e-32_prt, 8.43982e-30_prt, 1.48542e-27_prt,
        1.66142e-25_prt, 1.24599e-23_prt, 6.55645e-22_prt, 2.51605e-20_prt,
        7.27852e-19_prt, 1.63315e-17_prt, 2.91337e-16_prt, 4.22167e-15_prt,
        5.06352e-14_prt, 5.11048e-13_prt, 4.40374e-12_prt, 3.28180e-11_prt,
        2.13937e-10_prt, 1.23238e-09_prt, 6.33036e-09_prt, 2.92316e-08_prt,
        1.22231e-07_prt, 4.65857e-07_prt, 1.62791e-06_prt, 5.24366e-06_prt,
        1.56444e-05_prt, 4.34216e-05_prt, 1.12566e-04_prt, 2.73554e-04_prt,
        6.25248e-04_prt, 1.34820e-03_prt, 2.75018e-03_prt, 5.32098e-03_prt,
        9.78768e-03_prt, 1.71549e-02_prt, 2.87089e-02_prt, 4.59637e-02_prt,
        7.05331e-02_prt, 1.03927e-01_prt, 1.47294e-01_prt, 2.01146e-01_prt,
        2.65135e-01_prt, 3.37925e-01_prt, 4.17229e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 0.00000e+00_prt, 0.00000e+00_prt, 0.00000e+00_prt,
        1.33696e-233_prt, 1.05001e-182_prt, 8.88418e-149_prt, 1.50181e-124_prt,
        2.18820e-106_prt, 2.87640e-92_prt, 5.55963e-81_prt, 9.33671e-72_prt,
        4.45389e-64_prt, 1.37058e-57_prt, 4.88856e-52_prt, 3.10614e-47_prt,
        4.85334e-43_prt, 2.38628e-39_prt, 4.47261e-36_prt, 3.71812e-33_prt,
        1.54748e-30_prt, 3.55680e-28_prt, 4.89187e-26_prt, 4.30188e-24_prt,
        2.55620e-22_prt, 1.07505e-20_prt, 3.32812e-19_prt, 7.84181e-18_prt,
        1.44715e-16_prt, 2.14389e-15_prt, 2.60471e-14_prt, 2.64398e-13_prt,
        2.27907e-12_prt, 1.69219e-11_prt, 1.09596e-10_prt, 6.26073e-10_prt,
        3.18581e-09_prt, 1.45679e-08_prt, 6.03332e-08_prt, 2.27904e-07_prt,
        7.90163e-07_prt, 2.52880e-06_prt, 7.50871e-06_prt, 2.07815e-05_prt,
        5.38356e-05_prt, 1.31037e-04_prt, 3.00711e-04_prt, 6.52692e-04_prt,
        1.34377e-03_prt, 2.63113e-03_prt, 4.91155e-03_prt, 8.76045e-03_prt,
        1.49612e-02_prt, 2.45118e-02_prt, 3.85956e-02_prt, 5.85053e-02_prt,
        8.55165e-02_prt, 1.20719e-01_prt, 1.64829e-01_prt, 2.18003e-01_prt,
        2.79711e-01_prt, 3.48679e-01_prt, 4.22943e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 0.00000e+00_prt, 0.00000e+00_prt, 1.43228e-275_prt,
        2.90526e-202_prt, 2.85407e-158_prt, 6.08538e-129_prt, 5.37574e-108_prt,
        2.72332e-92_prt, 4.40352e-80_prt, 2.53579e-70_prt, 2.41452e-62_prt,
        1.05824e-55_prt, 4.33779e-50_prt, 2.75960e-45_prt, 3.95059e-41_prt,
        1.68168e-37_prt, 2.63427e-34_prt, 1.79233e-31_prt, 6.03771e-29_prt,
        1.11818e-26_prt, 1.23925e-24_prt, 8.80945e-23_prt, 4.25371e-21_prt,
        1.46338e-19_prt, 3.73368e-18_prt, 7.30873e-17_prt, 1.12985e-15_prt,
        1.41392e-14_prt, 1.46326e-13_prt, 1.27566e-12_prt, 9.52018e-12_prt,
        6.16825e-11_prt, 3.51273e-10_prt, 1.77753e-09_prt, 8.06971e-09_prt,
        3.31493e-08_prt, 1.24157e-07_prt, 4.26866e-07_prt, 1.35544e-06_prt,
        3.99674e-06_prt, 1.09977e-05_prt, 2.83656e-05_prt, 6.88522e-05_prt,
        1.57855e-04_prt, 3.42959e-04_prt, 7.08237e-04_prt, 1.39397e-03_prt,
        2.62155e-03_prt, 4.72160e-03_prt, 8.16139e-03_prt, 1.35653e-02_prt,
        2.17204e-02_prt, 3.35592e-02_prt, 5.01121e-02_prt, 7.24274e-02_prt,
        1.01462e-01_prt, 1.37952e-01_prt, 1.82280e-01_prt, 2.34366e-01_prt,
        2.93590e-01_prt, 3.58779e-01_prt, 4.28265e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 0.00000e+00_prt, 0.00000e+00_prt, 1.51593e-238_prt,
        3.48684e-175_prt, 3.72191e-137_prt, 8.44188e-112_prt, 1.08918e-93_prt,
        4.14348e-80_prt, 1.49893e-69_prt, 4.14701e-61_prt, 3.30780e-54_prt,
        1.84292e-48_prt, 1.32364e-43_prt, 1.89875e-39_prt, 7.50022e-36_prt,
        1.03805e-32_prt, 6.05229e-30_prt, 1.71569e-27_prt, 2.64805e-25_prt,
        2.43622e-23_prt, 1.43764e-21_prt, 5.77804e-20_prt, 1.66200e-18_prt,
        3.56565e-17_prt, 5.90700e-16_prt, 7.78145e-15_prt, 8.35746e-14_prt,
        7.47669e-13_prt, 5.67517e-12_prt, 3.71387e-11_prt, 2.12468e-10_prt,
        1.07564e-09_prt, 4.87056e-09_prt, 1.99124e-08_prt, 7.41165e-08_prt,
        2.53026e-07_prt, 7.97501e-07_prt, 2.33436e-06_prt, 6.37912e-06_prt,
        1.63518e-05_prt, 3.94847e-05_prt, 9.01614e-05_prt, 1.95367e-04_prt,
        4.02983e-04_prt, 7.93547e-04_prt, 1.49569e-03_prt, 2.70473e-03_prt,
        4.70293e-03_prt, 7.87848e-03_prt, 1.27392e-02_prt, 1.99162e-02_prt,
        3.01520e-02_prt, 4.42696e-02_prt, 6.31203e-02_prt, 8.75113e-02_prt,
        1.18119e-01_prt, 1.55396e-01_prt, 1.99486e-01_prt, 2.50156e-01_prt,
        3.06761e-01_prt, 3.68251e-01_prt, 4.33220e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 0.00000e+00_prt, 0.00000e+00_prt, 1.49266e-206_prt,
        8.76968e-152_prt, 6.58450e-119_prt, 5.51350e-97_prt, 2.52491e-81_prt,
        1.40295e-69_prt, 1.90378e-60_prt, 3.82392e-53_prt, 3.57979e-47_prt,
        3.35629e-42_prt, 5.35264e-38_prt, 2.12010e-34_prt, 2.75294e-31_prt,
        1.44327e-28_prt, 3.58262e-26_prt, 4.76628e-24_prt, 3.74791e-22_prt,
        1.88382e-20_prt, 6.44864e-19_prt, 1.58345e-17_prt, 2.91116e-16_prt,
        4.15302e-15_prt, 4.73727e-14_prt, 4.43185e-13_prt, 3.47475e-12_prt,
        2.32589e-11_prt, 1.35057e-10_prt, 6.89789e-10_prt, 3.13627e-09_prt,
        1.28287e-08_prt, 4.76470e-08_prt, 1.62000e-07_prt, 5.07870e-07_prt,
        1.47750e-06_prt, 4.01156e-06_prt, 1.02170e-05_prt, 2.45213e-05_prt,
        5.56861e-05_prt, 1.20099e-04_prt, 2.46812e-04_prt, 4.84774e-04_prt,
        9.12525e-04_prt, 1.65029e-03_prt, 2.87391e-03_prt, 4.82923e-03_prt,
        7.84506e-03_prt, 1.23420e-02_prt, 1.88336e-02_prt, 2.79182e-02_prt,
        4.02565e-02_prt, 5.65371e-02_prt, 7.74272e-02_prt, 1.03514e-01_prt,
        1.35243e-01_prt, 1.72849e-01_prt, 2.16312e-01_prt, 2.65312e-01_prt,
        3.19222e-01_prt, 3.77120e-01_prt, 4.37832e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 0.00000e+00_prt, 1.58793e-273_prt, 6.61011e-179_prt,
        1.46284e-131_prt, 3.87939e-103_prt, 3.51382e-84_prt, 1.23011e-70_prt,
        1.77378e-60_prt, 1.41107e-52_prt, 2.93962e-46_prt, 4.32235e-41_prt,
        8.68976e-37_prt, 3.77603e-33_prt, 4.91941e-30_prt, 2.44280e-27_prt,
        5.53569e-25_prt, 6.57043e-23_prt, 4.54673e-21_prt, 1.99639e-19_prt,
        5.95176e-18_prt, 1.27267e-16_prt, 2.04147e-15_prt, 2.54936e-14_prt,
        2.55622e-13_prt, 2.11212e-12_prt, 1.47004e-11_prt, 8.78122e-11_prt,
        4.57466e-10_prt, 2.10737e-09_prt, 8.68763e-09_prt, 3.23866e-08_prt,
        1.10177e-07_prt, 3.44792e-07_prt, 9.99597e-07_prt, 2.70151e-06_prt,
        6.84404e-06_prt, 1.63337e-05_prt, 3.68843e-05_prt, 7.91224e-05_prt,
        1.61808e-04_prt, 3.16473e-04_prt, 5.93691e-04_prt, 1.07105e-03_prt,
        1.86257e-03_prt, 3.12902e-03_prt, 5.08803e-03_prt, 8.02264e-03_prt,
        1.22864e-02_prt, 1.83032e-02_prt, 2.65600e-02_prt, 3.75911e-02_prt,
        5.19532e-02_prt, 7.01924e-02_prt, 9.28038e-02_prt, 1.20187e-01_prt,
        1.52603e-01_prt, 1.90133e-01_prt, 2.32648e-01_prt, 2.79792e-01_prt,
        3.30980e-01_prt, 3.85415e-01_prt, 4.42123e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 0.00000e+00_prt, 8.41483e-237_prt, 5.13669e-155_prt,
        4.37898e-114_prt, 1.65375e-89_prt, 4.10179e-73_prt, 2.12988e-61_prt,
        1.30962e-52_prt, 8.98269e-46_prt, 2.64139e-40_prt, 7.84631e-36_prt,
        4.17035e-32_prt, 5.89256e-29_prt, 2.93719e-26_prt, 6.35624e-24_prt,
        6.97817e-22_prt, 4.37819e-20_prt, 1.72232e-18_prt, 4.57083e-17_prt,
        8.67713e-16_prt, 1.23555e-14_prt, 1.37184e-13_prt, 1.22642e-12_prt,
        9.06732e-12_prt, 5.66989e-11_prt, 3.05620e-10_prt, 1.44321e-09_prt,
        6.05407e-09_prt, 2.28311e-08_prt, 7.82116e-08_prt, 2.45582e-07_prt,
        7.12411e-07_prt, 1.92258e-06_prt, 4.85639e-06_prt, 1.15442e-05_prt,
        2.59492e-05_prt, 5.53927e-05_prt, 1.12722e-04_prt, 2.19423e-04_prt,
        4.09838e-04_prt, 7.36549e-04_prt, 1.27685e-03_prt, 2.14000e-03_prt,
        3.47470e-03_prt, 5.47600e-03_prt, 8.39063e-03_prt, 1.25195e-02_prt,
        1.82164e-02_prt, 2.58816e-02_prt, 3.59497e-02_prt, 4.88719e-02_prt,
        6.50927e-02_prt, 8.50218e-02_prt, 1.09003e-01_prt, 1.37286e-01_prt,
        1.69993e-01_prt, 2.07098e-01_prt, 2.48412e-01_prt, 2.93572e-01_prt,
        3.42050e-01_prt, 3.93165e-01_prt, 4.46113e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 0.00000e+00_prt, 4.55844e-205_prt, 2.27106e-134_prt,
        5.55927e-99_prt, 9.97211e-78_prt, 1.51108e-63_prt, 2.06349e-53_prt,
        8.30931e-46_prt, 6.85779e-40_prt, 3.71472e-35_prt, 2.77086e-31_prt,
        4.66556e-28_prt, 2.49733e-25_prt, 5.42730e-23_prt, 5.73037e-21_prt,
        3.36330e-19_prt, 1.21642e-17_prt, 2.93735e-16_prt, 5.04529e-15_prt,
        6.48460e-14_prt, 6.49787e-13_prt, 5.24973e-12_prt, 3.51590e-11_prt,
        1.99764e-10_prt, 9.81792e-10_prt, 4.24316e-09_prt, 1.63536e-08_prt,
        5.68862e-08_prt, 1.80455e-07_prt, 5.26740e-07_prt, 1.42588e-06_prt,
        3.60409e-06_prt, 8.55743e-06_prt, 1.91878e-05_prt, 4.08205e-05_prt,
        8.27388e-05_prt, 1.60372e-04_prt, 2.98248e-04_prt, 5.33760e-04_prt,
        9.21711e-04_prt, 1.53947e-03_prt, 2.49239e-03_prt, 3.91911e-03_prt,
        5.99602e-03_prt, 8.94020e-03_prt, 1.30102e-02_prt, 1.85040e-02_prt,
        2.57530e-02_prt, 3.51130e-02_prt, 4.69503e-02_prt, 6.16258e-02_prt,
        7.94746e-02_prt, 1.00785e-01_prt, 1.25778e-01_prt, 1.54584e-01_prt,
        1.87231e-01_prt, 2.23623e-01_prt, 2.63542e-01_prt, 3.06640e-01_prt,
        3.52450e-01_prt, 4.00398e-01_prt, 4.49822e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 0.00000e+00_prt, 1.20482e-177_prt, 1.57838e-116_prt,
        6.29271e-86_prt, 1.51798e-67_prt, 2.81107e-55_prt, 1.65577e-46_prt,
        6.32592e-40_prt, 8.37836e-35_prt, 1.05225e-30_prt, 2.37828e-27_prt,
        1.48323e-24_prt, 3.43190e-22_prt, 3.64024e-20_prt, 2.06740e-18_prt,
        7.06337e-17_prt, 1.58711e-15_prt, 2.51381e-14_prt, 2.96472e-13_prt,
        2.72023e-12_prt, 2.01198e-11_prt, 1.23499e-10_prt, 6.44398e-10_prt,
        2.91607e-09_prt, 1.16386e-08_prt, 4.15573e-08_prt, 1.34372e-07_prt,
        3.97561e-07_prt, 1.08599e-06_prt, 2.76028e-06_prt, 6.57236e-06_prt,
        1.47471e-05_prt, 3.13448e-05_prt, 6.34004e-05_prt, 1.22532e-04_prt,
        2.27096e-04_prt, 4.04916e-04_prt, 6.96571e-04_prt, 1.15913e-03_prt,
        1.87012e-03_prt, 2.93148e-03_prt, 4.47305e-03_prt, 6.65524e-03_prt,
        9.67030e-03_prt, 1.37419e-02_prt, 1.91225e-02_prt, 2.60882e-02_prt,
        3.49313e-02_prt, 4.59500e-02_prt, 5.94364e-02_prt, 7.56627e-02_prt,
        9.48661e-02_prt, 1.17234e-01_prt, 1.42889e-01_prt, 1.71879e-01_prt,
        2.04164e-01_prt, 2.39612e-01_prt, 2.77996e-01_prt, 3.18995e-01_prt,
        3.62205e-01_prt, 4.07143e-01_prt, 4.53269e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 0.00000e+00_prt, 5.99544e-154_prt, 4.14924e-101_prt,
        1.20684e-74_prt, 9.62151e-59_prt, 3.96369e-48_prt, 1.54418e-40_prt,
        7.71031e-35_prt, 2.10168e-30_prt, 7.46541e-27_prt, 6.01388e-24_prt,
        1.59000e-21_prt, 1.78233e-19_prt, 1.01722e-17_prt, 3.38066e-16_prt,
        7.23646e-15_prt, 1.07766e-13_prt, 1.18561e-12_prt, 1.01039e-11_prt,
        6.92812e-11_prt, 3.94155e-10_prt, 1.90798e-09_prt, 8.02357e-09_prt,
        2.98251e-08_prt, 9.94357e-08_prt, 3.01028e-07_prt, 8.36260e-07_prt,
        2.15109e-06_prt, 5.16335e-06_prt, 1.16436e-05_prt, 2.48126e-05_prt,
        5.02246e-05_prt, 9.70014e-05_prt, 1.79467e-04_prt, 3.19203e-04_prt,
        5.47500e-04_prt, 9.08129e-04_prt, 1.46030e-03_prt, 2.28159e-03_prt,
        3.47064e-03_prt, 5.14926e-03_prt, 7.46373e-03_prt, 1.05849e-02_prt,
        1.47071e-02_prt, 2.00450e-02_prt, 2.68295e-02_prt, 3.53013e-02_prt,
        4.57031e-02_prt, 5.82712e-02_prt, 7.32249e-02_prt, 9.07564e-02_prt,
        1.11020e-01_prt, 1.34124e-01_prt, 1.60119e-01_prt, 1.88994e-01_prt,
        2.20669e-01_prt, 2.54993e-01_prt, 2.91748e-01_prt, 3.30646e-01_prt,
        3.71339e-01_prt, 4.13428e-01_prt, 4.56472e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 3.09018e-270_prt, 1.80390e-133_prt, 8.80988e-88_prt,
        6.82879e-65_prt, 3.91287e-51_prt, 6.01679e-42_prt, 2.24300e-35_prt,
        1.93029e-30_prt, 1.34253e-26_prt, 1.60141e-23_prt, 5.29015e-21_prt,
        6.66294e-19_prt, 3.99298e-17_prt, 1.33392e-15_prt, 2.79016e-14_prt,
        3.98728e-13_prt, 4.16160e-12_prt, 3.34150e-11_prt, 2.15053e-10_prt,
        1.14642e-09_prt, 5.19866e-09_prt, 2.04954e-08_prt, 7.15253e-08_prt,
        2.24297e-07_prt, 6.40075e-07_prt, 1.68005e-06_prt, 4.09311e-06_prt,
        9.32854e-06_prt, 2.00230e-05_prt, 4.07133e-05_prt, 7.88211e-05_prt,
        1.45942e-04_prt, 2.59447e-04_prt, 4.44372e-04_prt, 7.35533e-04_prt,
        1.17976e-03_prt, 1.83811e-03_prt, 2.78792e-03_prt, 4.12440e-03_prt,
        5.96172e-03_prt, 8.43331e-03_prt, 1.16912e-02_prt, 1.59043e-02_prt,
        2.12556e-02_prt, 2.79383e-02_prt, 3.61506e-02_prt, 4.60903e-02_prt,
        5.79473e-02_prt, 7.18973e-02_prt, 8.80942e-02_prt, 1.06663e-01_prt,
        1.27692e-01_prt, 1.51229e-01_prt, 1.77275e-01_prt, 2.05781e-01_prt,
        2.36645e-01_prt, 2.69715e-01_prt, 3.04786e-01_prt, 3.41606e-01_prt,
        3.79881e-01_prt, 4.19280e-01_prt, 4.59446e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 5.35199e-234_prt, 8.99239e-116_prt, 2.90990e-76_prt,
        1.84101e-56_prt, 1.48341e-44_prt, 1.33146e-36_prt, 6.53485e-31_prt,
        1.23103e-26_prt, 2.63040e-23_prt, 1.22212e-20_prt, 1.86879e-18_prt,
        1.24017e-16_prt, 4.32619e-15_prt, 9.09794e-14_prt, 1.27546e-12_prt,
        1.28549e-11_prt, 9.86818e-11_prt, 6.03541e-10_prt, 3.04728e-09_prt,
        1.30681e-08_prt, 4.87097e-08_prt, 1.60813e-07_prt, 4.77647e-07_prt,
        1.29308e-06_prt, 3.22569e-06_prt, 7.48383e-06_prt, 1.62764e-05_prt,
        3.34090e-05_prt, 6.50994e-05_prt, 1.21031e-04_prt, 2.15646e-04_prt,
        3.69643e-04_prt, 6.11643e-04_prt, 9.79912e-04_prt, 1.52406e-03_prt,
        2.30657e-03_prt, 3.40403e-03_prt, 4.90794e-03_prt, 6.92498e-03_prt,
        9.57657e-03_prt, 1.29977e-02_prt, 1.73352e-02_prt, 2.27448e-02_prt,
        2.93879e-02_prt, 3.74277e-02_prt, 4.70243e-02_prt, 5.83301e-02_prt,
        7.14846e-02_prt, 8.66092e-02_prt, 1.03802e-01_prt, 1.23135e-01_prt,
        1.44647e-01_prt, 1.68343e-01_prt, 1.94190e-01_prt, 2.22119e-01_prt,
        2.52019e-01_prt, 2.83744e-01_prt, 3.17108e-01_prt, 3.51895e-01_prt,
        3.87858e-01_prt, 4.24726e-01_prt, 4.62208e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 1.10366e-202_prt, 1.77441e-100_prt, 2.63415e-66_prt,
        3.57775e-49_prt, 7.23954e-39_prt, 5.58144e-32_prt, 4.75140e-27_prt,
        2.40872e-23_prt, 1.85636e-20_prt, 3.81877e-18_prt, 3.00131e-16_prt,
        1.14488e-14_prt, 2.50208e-13_prt, 3.52728e-12_prt, 3.49933e-11_prt,
        2.60808e-10_prt, 1.53525e-09_prt, 7.42117e-09_prt, 3.03789e-08_prt,
        1.07940e-07_prt, 3.39608e-07_prt, 9.61749e-07_prt, 2.48489e-06_prt,
        5.92398e-06_prt, 1.31553e-05_prt, 2.74316e-05_prt, 5.40808e-05_prt,
        1.01397e-04_prt, 1.81718e-04_prt, 3.12656e-04_prt, 5.18437e-04_prt,
        8.31268e-04_prt, 1.29265e-03_prt, 1.95454e-03_prt, 2.88028e-03_prt,
        4.14512e-03_prt, 5.83644e-03_prt, 8.05333e-03_prt, 1.09058e-02_prt,
        1.45133e-02_prt, 1.90030e-02_prt, 2.45068e-02_prt, 3.11594e-02_prt,
        3.90942e-02_prt, 4.84403e-02_prt, 5.93190e-02_prt, 7.18397e-02_prt,
        8.60969e-02_prt, 1.02167e-01_prt, 1.20103e-01_prt, 1.39937e-01_prt,
        1.61673e-01_prt, 1.85287e-01_prt, 2.10728e-01_prt, 2.37914e-01_prt,
        2.66737e-01_prt, 2.97060e-01_prt, 3.28721e-01_prt, 3.61536e-01_prt,
        3.95300e-01_prt, 4.29789e-01_prt, 4.64771e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 1.26685e-175_prt, 2.94133e-87_prt, 1.06599e-57_prt,
        7.16756e-43_prt, 6.01109e-34_prt, 5.55779e-28_prt, 1.04131e-23_prt,
        1.69811e-20_prt, 5.42194e-18_prt, 5.51650e-16_prt, 2.43927e-14_prt,
        5.76631e-13_prt, 8.41263e-12_prt, 8.39366e-11_prt, 6.17584e-10_prt,
        3.54609e-09_prt, 1.65933e-08_prt, 6.54488e-08_prt, 2.23480e-07_prt,
        6.74877e-07_prt, 1.83390e-06_prt, 4.54829e-06_prt, 1.04169e-05_prt,
        2.22481e-05_prt, 4.46771e-05_prt, 8.49453e-05_prt, 1.53828e-04_prt,
        2.66678e-04_prt, 4.44534e-04_prt, 7.15225e-04_prt, 1.11441e-03_prt,
        1.68646e-03_prt, 2.48513e-03_prt, 3.57389e-03_prt, 5.02603e-03_prt,
        6.92425e-03_prt, 9.35997e-03_prt, 1.24322e-02_prt, 1.62459e-02_prt,
        2.09105e-02_prt, 2.65373e-02_prt, 3.32377e-02_prt, 4.11201e-02_prt,
        5.02877e-02_prt, 6.08358e-02_prt, 7.28492e-02_prt, 8.64000e-02_prt,
        1.01545e-01_prt, 1.18325e-01_prt, 1.36760e-01_prt, 1.56853e-01_prt,
        1.78584e-01_prt, 2.01912e-01_prt, 2.26777e-01_prt, 2.53095e-01_prt,
        2.80762e-01_prt, 3.09658e-01_prt, 3.39641e-01_prt, 3.70556e-01_prt,
        4.02235e-01_prt, 4.34495e-01_prt, 4.67149e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 3.06823e-152_prt, 7.84694e-76_prt, 2.94312e-50_prt,
        2.01640e-37_prt, 1.08049e-29_prt, 1.59848e-24_prt, 8.09748e-21_prt,
        4.95566e-18_prt, 7.38924e-16_prt, 4.09157e-14_prt, 1.10057e-12_prt,
        1.72052e-11_prt, 1.77014e-10_prt, 1.31001e-09_prt, 7.44426e-09_prt,
        3.41152e-08_prt, 1.30907e-07_prt, 4.33080e-07_prt, 1.26417e-06_prt,
        3.31680e-06_prt, 7.94014e-06_prt, 1.75580e-05_prt, 3.62307e-05_prt,
        7.03591e-05_prt, 1.29511e-04_prt, 2.27332e-04_prt, 3.82492e-04_prt,
        6.19605e-04_prt, 9.70054e-04_prt, 1.47267e-03_prt, 2.17418e-03_prt,
        3.12943e-03_prt, 4.40132e-03_prt, 6.06038e-03_prt, 8.18413e-03_prt,
        1.08561e-02_prt, 1.41646e-02_prt, 1.82011e-02_prt, 2.30590e-02_prt,
        2.88314e-02_prt, 3.56093e-02_prt, 4.34800e-02_prt, 5.25249e-02_prt,
        6.28178e-02_prt, 7.44230e-02_prt, 8.73941e-02_prt, 1.01772e-01_prt,
        1.17585e-01_prt, 1.34846e-01_prt, 1.53553e-01_prt, 1.73689e-01_prt,
        1.95219e-01_prt, 2.18095e-01_prt, 2.42252e-01_prt, 2.67610e-01_prt,
        2.94075e-01_prt, 3.21540e-01_prt, 3.49886e-01_prt, 3.78983e-01_prt,
        4.08691e-01_prt, 4.38866e-01_prt, 4.69354e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 4.95800e-132_prt, 5.90844e-66_prt, 7.98796e-44_prt,
        1.04041e-32_prt, 5.17752e-26_prt, 1.57228e-21_prt, 2.57237e-18_prt,
        6.74966e-16_prt, 5.21207e-14_prt, 1.70592e-12_prt, 2.98623e-11_prt,
        3.26542e-10_prt, 2.48412e-09_prt, 1.41997e-08_prt, 6.45408e-08_prt,
        2.43387e-07_prt, 7.86698e-07_prt, 2.23553e-06_prt, 5.69801e-06_prt,
        1.32378e-05_prt, 2.83996e-05_prt, 5.68658e-05_prt, 1.07218e-04_prt,
        1.91761e-04_prt, 3.27358e-04_prt, 5.36214e-04_prt, 8.46544e-04_prt,
        1.29309e-03_prt, 1.91741e-03_prt, 2.76798e-03_prt, 3.89998e-03_prt,
        5.37485e-03_prt, 7.25969e-03_prt, 9.62628e-03_prt, 1.25501e-02_prt,
        1.61089e-02_prt, 2.03818e-02_prt, 2.54473e-02_prt, 3.13823e-02_prt,
        3.82604e-02_prt, 4.61506e-02_prt, 5.51159e-02_prt, 6.52122e-02_prt,
        7.64871e-02_prt, 8.89788e-02_prt, 1.02716e-01_prt, 1.17715e-01_prt,
        1.33982e-01_prt, 1.51513e-01_prt, 1.70289e-01_prt, 1.90281e-01_prt,
        2.11449e-01_prt, 2.33740e-01_prt, 2.57091e-01_prt, 2.81428e-01_prt,
        3.06668e-01_prt, 3.32719e-01_prt, 3.59480e-01_prt, 3.86844e-01_prt,
        4.14698e-01_prt, 4.42924e-01_prt, 4.71399e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 1.44533e-114_prt, 2.04002e-57_prt, 2.92203e-38_prt,
        1.24009e-28_prt, 7.91692e-23_prt, 6.11998e-19_prt, 3.77057e-16_prt,
        4.75895e-14_prt, 2.08099e-12_prt, 4.32387e-11_prt, 5.22135e-10_prt,
        4.19201e-09_prt, 2.45648e-08_prt, 1.12319e-07_prt, 4.20867e-07_prt,
        1.34091e-06_prt, 3.73661e-06_prt, 9.30974e-06_prt, 2.11026e-05_prt,
        4.41295e-05_prt, 8.61053e-05_prt, 1.58221e-04_prt, 2.75907e-04_prt,
        4.59522e-04_prt, 7.34912e-04_prt, 1.13379e-03_prt, 1.69391e-03_prt,
        2.45902e-03_prt, 3.47858e-03_prt, 4.80725e-03_prt, 6.50424e-03_prt,
        8.63244e-03_prt, 1.12574e-02_prt, 1.44465e-02_prt, 1.82673e-02_prt,
        2.27868e-02_prt, 2.80705e-02_prt, 3.41806e-02_prt, 4.11754e-02_prt,
        4.91083e-02_prt, 5.80271e-02_prt, 6.79724e-02_prt, 7.89781e-02_prt,
        9.10699e-02_prt, 1.04265e-01_prt, 1.18573e-01_prt, 1.33994e-01_prt,
        1.50518e-01_prt, 1.68128e-01_prt, 1.86798e-01_prt, 2.06493e-01_prt,
        2.27168e-01_prt, 2.48773e-01_prt, 2.71249e-01_prt, 2.94530e-01_prt,
        3.18543e-01_prt, 3.43210e-01_prt, 3.68448e-01_prt, 3.94169e-01_prt,
        4.20281e-01_prt, 4.46688e-01_prt, 4.73294e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 1.79517e-99_prt, 4.91237e-50_prt, 1.89220e-33_prt,
        4.16777e-25_prt, 4.51239e-20_prt, 1.06937e-16_prt, 2.83302e-14_prt,
        1.89962e-12_prt, 5.08130e-11_prt, 7.12850e-10_prt, 6.24477e-09_prt,
        3.83849e-08_prt, 1.79496e-07_prt, 6.76626e-07_prt, 2.14545e-06_prt,
        5.90819e-06_prt, 1.44810e-05_prt, 3.21995e-05_prt, 6.59426e-05_prt,
        1.25896e-04_prt, 2.26284e-04_prt, 3.86005e-04_prt, 6.29103e-04_prt,
        9.85056e-04_prt, 1.48883e-03_prt, 2.18072e-03_prt, 3.10596e-03_prt,
        4.31421e-03_prt, 5.85876e-03_prt, 7.79575e-03_prt, 1.01832e-02_prt,
        1.30801e-02_prt, 1.65454e-02_prt, 2.06369e-02_prt, 2.54106e-02_prt,
        3.09194e-02_prt, 3.72129e-02_prt, 4.43359e-02_prt, 5.23287e-02_prt,
        6.12256e-02_prt, 7.10552e-02_prt, 8.18396e-02_prt, 9.35946e-02_prt,
        1.06329e-01_prt, 1.20045e-01_prt, 1.34737e-01_prt, 1.50394e-01_prt,
        1.66999e-01_prt, 1.84525e-01_prt, 2.02942e-01_prt, 2.22214e-01_prt,
        2.42297e-01_prt, 2.63144e-01_prt, 2.84700e-01_prt, 3.06910e-01_prt,
        3.29711e-01_prt, 3.53037e-01_prt, 3.76820e-01_prt, 4.00988e-01_prt,
        4.25466e-01_prt, 4.50180e-01_prt, 4.75050e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 1.99614e-86_prt, 1.18517e-43_prt, 2.74527e-29_prt,
        4.69191e-22_prt, 1.09641e-17_prt, 9.35339e-15_prt, 1.19488e-12_prt,
        4.63800e-11_prt, 8.11222e-10_prt, 8.10414e-09_prt, 5.37917e-08_prt,
        2.62460e-07_prt, 1.00976e-06_prt, 3.22103e-06_prt, 8.83961e-06_prt,
        2.14581e-05_prt, 4.70665e-05_prt, 9.48423e-05_prt, 1.77895e-04_prt,
        3.13887e-04_prt, 5.25459e-04_prt, 8.40437e-04_prt, 1.29179e-03_prt,
        1.91736e-03_prt, 2.75939e-03_prt, 3.86386e-03_prt, 5.27975e-03_prt,
        7.05813e-03_prt, 9.25132e-03_prt, 1.19119e-02_prt, 1.50919e-02_prt,
        1.88418e-02_prt, 2.32102e-02_prt, 2.82423e-02_prt, 3.39802e-02_prt,
        4.04618e-02_prt, 4.77204e-02_prt, 5.57847e-02_prt, 6.46781e-02_prt,
        7.44188e-02_prt, 8.50196e-02_prt, 9.64880e-02_prt, 1.08826e-01_prt,
        1.22030e-01_prt, 1.36091e-01_prt, 1.50996e-01_prt, 1.66726e-01_prt,
        1.83257e-01_prt, 2.00562e-01_prt, 2.18609e-01_prt, 2.37361e-01_prt,
        2.56778e-01_prt, 2.76817e-01_prt, 2.97432e-01_prt, 3.18573e-01_prt,
        3.40189e-01_prt, 3.62224e-01_prt, 3.84623e-01_prt, 4.07328e-01_prt,
        4.30279e-01_prt, 4.53416e-01_prt, 4.76677e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 3.77431e-75_prt, 3.91782e-38_prt, 1.09378e-25_prt,
        2.05310e-19_prt, 1.27532e-15_prt, 4.49930e-13_prt, 3.06002e-11_prt,
        7.40487e-10_prt, 8.97080e-09_prt, 6.68206e-08_prt, 3.48911e-07_prt,
        1.39420e-06_prt, 4.53089e-06_prt, 1.25092e-05_prt, 3.02964e-05_prt,
        6.59394e-05_prt, 1.31379e-04_prt, 2.43107e-04_prt, 4.22591e-04_prt,
        6.96427e-04_prt, 1.09621e-03_prt, 1.65813e-03_prt, 2.42244e-03_prt,
        3.43262e-03_prt, 4.73461e-03_prt, 6.37581e-03_prt, 8.40429e-03_prt,
        1.08678e-02_prt, 1.38130e-02_prt, 1.72847e-02_prt, 2.13252e-02_prt,
        2.59738e-02_prt, 3.12662e-02_prt, 3.72342e-02_prt, 4.39056e-02_prt,
        5.13035e-02_prt, 5.94468e-02_prt, 6.83497e-02_prt, 7.80219e-02_prt,
        8.84686e-02_prt, 9.96904e-02_prt, 1.11684e-01_prt, 1.24442e-01_prt,
        1.37952e-01_prt, 1.52199e-01_prt, 1.67164e-01_prt, 1.82824e-01_prt,
        1.99155e-01_prt, 2.16128e-01_prt, 2.33711e-01_prt, 2.51871e-01_prt,
        2.70572e-01_prt, 2.89776e-01_prt, 3.09443e-01_prt, 3.29532e-01_prt,
        3.49998e-01_prt, 3.70799e-01_prt, 3.91888e-01_prt, 4.13218e-01_prt,
        4.34743e-01_prt, 4.56414e-01_prt, 4.78182e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 2.11236e-65_prt, 2.32556e-33_prt, 1.42680e-22_prt,
        3.97114e-17_prt, 7.84976e-14_prt, 1.29110e-11_prt, 5.09207e-10_prt,
        8.19002e-09_prt, 7.22268e-08_prt, 4.17383e-07_prt, 1.77099e-06_prt,
        5.95385e-06_prt, 1.67202e-05_prt, 4.07382e-05_prt, 8.85495e-05_prt,
        1.75349e-04_prt, 3.21469e-04_prt, 5.52521e-04_prt, 8.99188e-04_prt,
        1.39672e-03_prt, 2.08420e-03_prt, 3.00366e-03_prt, 4.19911e-03_prt,
        5.71555e-03_prt, 7.59803e-03_prt, 9.89077e-03_prt, 1.26364e-02_prt,
        1.58752e-02_prt, 1.96447e-02_prt, 2.39791e-02_prt, 2.89089e-02_prt,
        3.44608e-02_prt, 4.06575e-02_prt, 4.75177e-02_prt, 5.50559e-02_prt,
        6.32824e-02_prt, 7.22038e-02_prt, 8.18227e-02_prt, 9.21382e-02_prt,
        1.03146e-01_prt, 1.14838e-01_prt, 1.27203e-01_prt, 1.40227e-01_prt,
        1.53895e-01_prt, 1.68186e-01_prt, 1.83080e-01_prt, 1.98553e-01_prt,
        2.14580e-01_prt, 2.31133e-01_prt, 2.48184e-01_prt, 2.65702e-01_prt,
        2.83656e-01_prt, 3.02014e-01_prt, 3.20741e-01_prt, 3.39803e-01_prt,
        3.59165e-01_prt, 3.78790e-01_prt, 3.98643e-01_prt, 4.18685e-01_prt,
        4.38880e-01_prt, 4.59190e-01_prt, 4.79576e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 5.64885e-57_prt, 3.13115e-29_prt, 7.09352e-20_prt,
        3.79397e-15_prt, 2.78788e-12_prt, 2.37089e-10_prt, 5.83808e-09_prt,
        6.59592e-08_prt, 4.42016e-07_prt, 2.05080e-06_prt, 7.27176e-06_prt,
        2.10524e-05_prt, 5.21029e-05_prt, 1.13928e-04_prt, 2.25496e-04_prt,
        4.11446e-04_prt, 7.01839e-04_prt, 1.13153e-03_prt, 1.73923e-03_prt,
        2.56646e-03_prt, 3.65640e-03_prt, 5.05278e-03_prt, 6.79885e-03_prt,
        8.93646e-03_prt, 1.15053e-02_prt, 1.45424e-02_prt, 1.80813e-02_prt,
        2.21521e-02_prt, 2.67812e-02_prt, 3.19909e-02_prt, 3.77995e-02_prt,
        4.42216e-02_prt, 5.12676e-02_prt, 5.89446e-02_prt, 6.72559e-02_prt,
        7.62016e-02_prt, 8.57786e-02_prt, 9.59811e-02_prt, 1.06800e-01_prt,
        1.18225e-01_prt, 1.30242e-01_prt, 1.42837e-01_prt, 1.55990e-01_prt,
        1.69685e-01_prt, 1.83900e-01_prt, 1.98614e-01_prt, 2.13804e-01_prt,
        2.29445e-01_prt, 2.45513e-01_prt, 2.61983e-01_prt, 2.78828e-01_prt,
        2.96021e-01_prt, 3.13534e-01_prt, 3.31340e-01_prt, 3.49410e-01_prt,
        3.67716e-01_prt, 3.86228e-01_prt, 4.04918e-01_prt, 4.23756e-01_prt,
        4.42713e-01_prt, 4.61759e-01_prt, 4.80865e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 1.09171e-49_prt, 1.17015e-25_prt, 1.53257e-17_prt,
        1.97065e-13_prt, 6.15660e-11_prt, 2.96033e-09_prt, 4.85087e-08_prt,
        4.03814e-07_prt, 2.13401e-06_prt, 8.18820e-06_prt, 2.48564e-05_prt,
        6.32274e-05_prt, 1.40265e-04_prt, 2.79269e-04_prt, 5.09674e-04_prt,
        8.66300e-04_prt, 1.38820e-03_prt, 2.11734e-03_prt, 3.09715e-03_prt,
        4.37128e-03_prt, 5.98236e-03_prt, 7.97107e-03_prt, 1.03753e-02_prt,
        1.32296e-02_prt, 1.65649e-02_prt, 2.04079e-02_prt, 2.47814e-02_prt,
        2.97039e-02_prt, 3.51902e-02_prt, 4.12509e-02_prt, 4.78932e-02_prt,
        5.51205e-02_prt, 6.29333e-02_prt, 7.13291e-02_prt, 8.03025e-02_prt,
        8.98461e-02_prt, 9.99498e-02_prt, 1.10602e-01_prt, 1.21789e-01_prt,
        1.33495e-01_prt, 1.45705e-01_prt, 1.58401e-01_prt, 1.71564e-01_prt,
        1.85175e-01_prt, 1.99214e-01_prt, 2.13660e-01_prt, 2.28491e-01_prt,
        2.43687e-01_prt, 2.59225e-01_prt, 2.75082e-01_prt, 2.91238e-01_prt,
        3.07668e-01_prt, 3.24350e-01_prt, 3.41261e-01_prt, 3.58379e-01_prt,
        3.75680e-01_prt, 3.93142e-01_prt, 4.10742e-01_prt, 4.28456e-01_prt,
        4.46262e-01_prt, 4.64136e-01_prt, 4.82057e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 2.17999e-43_prt, 1.44503e-22_prt, 1.61168e-15_prt,
        6.04581e-12_prt, 9.01779e-10_prt, 2.64847e-08_prt, 3.05151e-07_prt,
        1.95048e-06_prt, 8.39295e-06_prt, 2.73221e-05_prt, 7.25035e-05_prt,
        1.64880e-04_prt, 3.32701e-04_prt, 6.10751e-04_prt, 1.03895e-03_prt,
        1.66060e-03_prt, 2.52064e-03_prt, 3.66392e-03_prt, 5.13381e-03_prt,
        6.97107e-03_prt, 9.21300e-03_prt, 1.18928e-02_prt, 1.50395e-02_prt,
        1.86772e-02_prt, 2.28260e-02_prt, 2.75012e-02_prt, 3.27141e-02_prt,
        3.84722e-02_prt, 4.47790e-02_prt, 5.16352e-02_prt, 5.90381e-02_prt,
        6.69827e-02_prt, 7.54613e-02_prt, 8.44644e-02_prt, 9.39806e-02_prt,
        1.03997e-01_prt, 1.14500e-01_prt, 1.25473e-01_prt, 1.36901e-01_prt,
        1.48766e-01_prt, 1.61051e-01_prt, 1.73738e-01_prt, 1.86809e-01_prt,
        2.00245e-01_prt, 2.14027e-01_prt, 2.28137e-01_prt, 2.42554e-01_prt,
        2.57261e-01_prt, 2.72239e-01_prt, 2.87468e-01_prt, 3.02930e-01_prt,
        3.18607e-01_prt, 3.34479e-01_prt, 3.50529e-01_prt, 3.66738e-01_prt,
        3.83088e-01_prt, 3.99562e-01_prt, 4.16141e-01_prt, 4.32808e-01_prt,
        4.49545e-01_prt, 4.66334e-01_prt, 4.83158e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 6.12543e-38_prt, 6.85566e-20_prt, 9.09852e-14_prt,
        1.17683e-10_prt, 9.26345e-09_prt, 1.77633e-07_prt, 1.50917e-06_prt,
        7.67527e-06_prt, 2.76457e-05_prt, 7.80584e-05_prt, 1.84367e-04_prt,
        3.80484e-04_prt, 7.07218e-04_prt, 1.21002e-03_prt, 1.93657e-03_prt,
        2.93453e-03_prt, 4.24963e-03_prt, 5.92428e-03_prt, 7.99658e-03_prt,
        1.04997e-02_prt, 1.34615e-02_prt, 1.69048e-02_prt, 2.08471e-02_prt,
        2.53013e-02_prt, 3.02755e-02_prt, 3.57742e-02_prt, 4.17979e-02_prt,
        4.83441e-02_prt, 5.54072e-02_prt, 6.29794e-02_prt, 7.10509e-02_prt,
        7.96100e-02_prt, 8.86436e-02_prt, 9.81373e-02_prt, 1.08076e-01_prt,
        1.18444e-01_prt, 1.29224e-01_prt, 1.40401e-01_prt, 1.51955e-01_prt,
        1.63871e-01_prt, 1.76131e-01_prt, 1.88717e-01_prt, 2.01613e-01_prt,
        2.14801e-01_prt, 2.28264e-01_prt, 2.41986e-01_prt, 2.55951e-01_prt,
        2.70141e-01_prt, 2.84543e-01_prt, 2.99139e-01_prt, 3.13914e-01_prt,
        3.28855e-01_prt, 3.43945e-01_prt, 3.59170e-01_prt, 3.74517e-01_prt,
        3.89970e-01_prt, 4.05517e-01_prt, 4.21143e-01_prt, 4.36836e-01_prt,
        4.52581e-01_prt, 4.68365e-01_prt, 4.84176e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 3.16256e-33_prt, 1.42325e-17_prt, 3.00080e-12_prt,
        1.54610e-09_prt, 7.00286e-08_prt, 9.28729e-07_prt, 6.06220e-06_prt,
        2.52969e-05_prt, 7.81137e-05_prt, 1.94977e-04_prt, 4.16312e-04_prt,
        7.89827e-04_prt, 1.36715e-03_prt, 2.20057e-03_prt, 3.34031e-03_prt,
        4.83252e-03_prt, 6.71797e-03_prt, 9.03129e-03_prt, 1.18008e-02_prt,
        1.50485e-02_prt, 1.87904e-02_prt, 2.30372e-02_prt, 2.77946e-02_prt,
        3.30639e-02_prt, 3.88427e-02_prt, 4.51252e-02_prt, 5.19032e-02_prt,
        5.91659e-02_prt, 6.69010e-02_prt, 7.50945e-02_prt, 8.37314e-02_prt,
        9.27957e-02_prt, 1.02271e-01_prt, 1.12140e-01_prt, 1.22387e-01_prt,
        1.32993e-01_prt, 1.43941e-01_prt, 1.55215e-01_prt, 1.66797e-01_prt,
        1.78671e-01_prt, 1.90822e-01_prt, 2.03232e-01_prt, 2.15886e-01_prt,
        2.28770e-01_prt, 2.41869e-01_prt, 2.55168e-01_prt, 2.68655e-01_prt,
        2.82314e-01_prt, 2.96134e-01_prt, 3.10102e-01_prt, 3.24206e-01_prt,
        3.38434e-01_prt, 3.52773e-01_prt, 3.67214e-01_prt, 3.81745e-01_prt,
        3.96356e-01_prt, 4.11036e-01_prt, 4.25774e-01_prt, 4.40561e-01_prt,
        4.55387e-01_prt, 4.70242e-01_prt, 4.85116e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 3.77813e-29_prt, 1.44680e-15_prt, 6.22031e-11_prt,
        1.44615e-08_prt, 4.06135e-07_prt, 3.91507e-06_prt, 2.03425e-05_prt,
        7.15233e-05_prt, 1.93276e-04_prt, 4.33552e-04_prt, 8.48163e-04_prt,
        1.49588e-03_prt, 2.43410e-03_prt, 3.71562e-03_prt, 5.38660e-03_prt,
        7.48549e-03_prt, 1.00427e-02_prt, 1.30811e-02_prt, 1.66160e-02_prt,
        2.06565e-02_prt, 2.52060e-02_prt, 3.02631e-02_prt, 3.58223e-02_prt,
        4.18747e-02_prt, 4.84089e-02_prt, 5.54111e-02_prt, 6.28663e-02_prt,
        7.07580e-02_prt, 7.90688e-02_prt, 8.77809e-02_prt, 9.68763e-02_prt,
        1.06337e-01_prt, 1.16144e-01_prt, 1.26280e-01_prt, 1.36728e-01_prt,
        1.47470e-01_prt, 1.58490e-01_prt, 1.69771e-01_prt, 1.81297e-01_prt,
        1.93054e-01_prt, 2.05027e-01_prt, 2.17202e-01_prt, 2.29565e-01_prt,
        2.42104e-01_prt, 2.54806e-01_prt, 2.67660e-01_prt, 2.80655e-01_prt,
        2.93779e-01_prt, 3.07022e-01_prt, 3.20374e-01_prt, 3.33826e-01_prt,
        3.47369e-01_prt, 3.60993e-01_prt, 3.74691e-01_prt, 3.88454e-01_prt,
        4.02275e-01_prt, 4.16145e-01_prt, 4.30057e-01_prt, 4.44004e-01_prt,
        4.57979e-01_prt, 4.71975e-01_prt, 4.85984e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 1.27452e-25_prt, 7.93605e-14_prt, 8.63184e-10_prt,
        1.00847e-07_prt, 1.87306e-06_prt, 1.36999e-05_prt, 5.84249e-05_prt,
        1.77091e-04_prt, 4.26306e-04_prt, 8.71675e-04_prt, 1.58062e-03_prt,
        2.61657e-03_prt, 4.03518e-03_prt, 5.88232e-03_prt, 8.19354e-03_prt,
        1.09944e-02_prt, 1.43012e-02_prt, 1.81224e-02_prt, 2.24595e-02_prt,
        2.73084e-02_prt, 3.26606e-02_prt, 3.85039e-02_prt, 4.48233e-02_prt,
        5.16019e-02_prt, 5.88213e-02_prt, 6.64621e-02_prt, 7.45046e-02_prt,
        8.29285e-02_prt, 9.17139e-02_prt, 1.00841e-01_prt, 1.10290e-01_prt,
        1.20043e-01_prt, 1.30080e-01_prt, 1.40385e-01_prt, 1.50940e-01_prt,
        1.61729e-01_prt, 1.72737e-01_prt, 1.83949e-01_prt, 1.95351e-01_prt,
        2.06930e-01_prt, 2.18673e-01_prt, 2.30567e-01_prt, 2.42603e-01_prt,
        2.54769e-01_prt, 2.67055e-01_prt, 2.79452e-01_prt, 2.91951e-01_prt,
        3.04543e-01_prt, 3.17220e-01_prt, 3.29974e-01_prt, 3.42800e-01_prt,
        3.55688e-01_prt, 3.68635e-01_prt, 3.81632e-01_prt, 3.94674e-01_prt,
        4.07756e-01_prt, 4.20872e-01_prt, 4.34016e-01_prt, 4.47185e-01_prt,
        4.60372e-01_prt, 4.73574e-01_prt, 4.86785e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 1.44202e-22_prt, 2.55451e-12_prt, 8.46798e-09_prt,
        5.45600e-07_prt, 7.08563e-06_prt, 4.08072e-05_prt, 1.46658e-04_prt,
        3.90884e-04_prt, 8.51202e-04_prt, 1.60605e-03_prt, 2.72653e-03_prt,
        4.27184e-03_prt, 6.28755e-03_prt, 8.80575e-03_prt, 1.18463e-02_prt,
        1.54184e-02_prt, 1.95227e-02_prt, 2.41527e-02_prt, 2.92967e-02_prt,
        3.49387e-02_prt, 4.10601e-02_prt, 4.76399e-02_prt, 5.46561e-02_prt,
        6.20859e-02_prt, 6.99065e-02_prt, 7.80948e-02_prt, 8.66286e-02_prt,
        9.54859e-02_prt, 1.04646e-01_prt, 1.14088e-01_prt, 1.23792e-01_prt,
        1.33741e-01_prt, 1.43918e-01_prt, 1.54304e-01_prt, 1.64885e-01_prt,
        1.75647e-01_prt, 1.86575e-01_prt, 1.97657e-01_prt, 2.08880e-01_prt,
        2.20233e-01_prt, 2.31706e-01_prt, 2.43289e-01_prt, 2.54973e-01_prt,
        2.66748e-01_prt, 2.78608e-01_prt, 2.90545e-01_prt, 3.02551e-01_prt,
        3.14621e-01_prt, 3.26749e-01_prt, 3.38928e-01_prt, 3.51154e-01_prt,
        3.63422e-01_prt, 3.75728e-01_prt, 3.88066e-01_prt, 4.00434e-01_prt,
        4.12826e-01_prt, 4.25241e-01_prt, 4.37673e-01_prt, 4.50121e-01_prt,
        4.62580e-01_prt, 4.75049e-01_prt, 4.87523e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 6.34899e-20_prt, 5.18780e-11_prt, 6.15571e-08_prt,
        2.37013e-06_prt, 2.25825e-05_prt, 1.05734e-04_prt, 3.27625e-04_prt,
        7.81043e-04_prt, 1.55910e-03_prt, 2.74338e-03_prt, 4.39812e-03_prt,
        6.56886e-03_prt, 9.28382e-03_prt, 1.25563e-02_prt, 1.63876e-02_prt,
        2.07697e-02_prt, 2.56876e-02_prt, 3.11213e-02_prt, 3.70476e-02_prt,
        4.34410e-02_prt, 5.02751e-02_prt, 5.75228e-02_prt, 6.51575e-02_prt,
        7.31527e-02_prt, 8.14833e-02_prt, 9.01246e-02_prt, 9.90536e-02_prt,
        1.08248e-01_prt, 1.17687e-01_prt, 1.27352e-01_prt, 1.37223e-01_prt,
        1.47284e-01_prt, 1.57520e-01_prt, 1.67914e-01_prt, 1.78453e-01_prt,
        1.89125e-01_prt, 1.99918e-01_prt, 2.10821e-01_prt, 2.21824e-01_prt,
        2.32917e-01_prt, 2.44093e-01_prt, 2.55342e-01_prt, 2.66659e-01_prt,
        2.78036e-01_prt, 2.89468e-01_prt, 3.00948e-01_prt, 3.12472e-01_prt,
        3.24035e-01_prt, 3.35634e-01_prt, 3.47263e-01_prt, 3.58920e-01_prt,
        3.70601e-01_prt, 3.82304e-01_prt, 3.94025e-01_prt, 4.05762e-01_prt,
        4.17513e-01_prt, 4.29276e-01_prt, 4.41049e-01_prt, 4.52830e-01_prt,
        4.64617e-01_prt, 4.76409e-01_prt, 4.88204e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 1.23682e-17_prt, 7.07629e-10_prt, 3.45337e-07_prt,
        8.51604e-06_prt, 6.20521e-05_prt, 2.42815e-04_prt, 6.61580e-04_prt,
        1.43165e-03_prt, 2.64983e-03_prt, 4.38816e-03_prt, 6.69315e-03_prt,
        9.58883e-03_prt, 1.30811e-02_prt, 1.71623e-02_prt, 2.18148e-02_prt,
        2.70141e-02_prt, 3.27317e-02_prt, 3.89366e-02_prt, 4.55966e-02_prt,
        5.26795e-02_prt, 6.01537e-02_prt, 6.79886e-02_prt, 7.61550e-02_prt,
        8.46252e-02_prt, 9.33733e-02_prt, 1.02375e-01_prt, 1.11607e-01_prt,
        1.21049e-01_prt, 1.30681e-01_prt, 1.40485e-01_prt, 1.50445e-01_prt,
        1.60546e-01_prt, 1.70773e-01_prt, 1.81113e-01_prt, 1.91555e-01_prt,
        2.02089e-01_prt, 2.12704e-01_prt, 2.23391e-01_prt, 2.34143e-01_prt,
        2.44952e-01_prt, 2.55812e-01_prt, 2.66716e-01_prt, 2.77660e-01_prt,
        2.88637e-01_prt, 2.99645e-01_prt, 3.10678e-01_prt, 3.21734e-01_prt,
        3.32810e-01_prt, 3.43902e-01_prt, 3.55008e-01_prt, 3.66127e-01_prt,
        3.77255e-01_prt, 3.88392e-01_prt, 3.99537e-01_prt, 4.10687e-01_prt,
        4.21842e-01_prt, 4.33001e-01_prt, 4.44163e-01_prt, 4.55328e-01_prt,
        4.66494e-01_prt, 4.77662e-01_prt, 4.88831e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 1.19109e-15_prt, 6.84291e-09_prt, 1.54848e-06_prt,
        2.59650e-05_prt, 1.49964e-04_prt, 5.02268e-04_prt, 1.22395e-03_prt,
        2.43500e-03_prt, 4.22038e-03_prt, 6.62970e-03_prt, 9.68294e-03_prt,
        1.33777e-02_prt, 1.76958e-02_prt, 2.26088e-02_prt, 2.80820e-02_prt,
        3.40772e-02_prt, 4.05551e-02_prt, 4.74768e-02_prt, 5.48042e-02_prt,
        6.25010e-02_prt, 7.05331e-02_prt, 7.88685e-02_prt, 8.74775e-02_prt,
        9.63328e-02_prt, 1.05409e-01_prt, 1.14683e-01_prt, 1.24134e-01_prt,
        1.33743e-01_prt, 1.43491e-01_prt, 1.53363e-01_prt, 1.63345e-01_prt,
        1.73424e-01_prt, 1.83587e-01_prt, 1.93824e-01_prt, 2.04125e-01_prt,
        2.14482e-01_prt, 2.24887e-01_prt, 2.35333e-01_prt, 2.45813e-01_prt,
        2.56323e-01_prt, 2.66857e-01_prt, 2.77411e-01_prt, 2.87981e-01_prt,
        2.98564e-01_prt, 3.09157e-01_prt, 3.19757e-01_prt, 3.30363e-01_prt,
        3.40972e-01_prt, 3.51582e-01_prt, 3.62194e-01_prt, 3.72805e-01_prt,
        3.83415e-01_prt, 3.94024e-01_prt, 4.04631e-01_prt, 4.15235e-01_prt,
        4.25837e-01_prt, 4.36437e-01_prt, 4.47034e-01_prt, 4.57630e-01_prt,
        4.68224e-01_prt, 4.78817e-01_prt, 4.89409e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 6.24054e-14_prt, 4.91529e-08_prt, 5.72022e-06_prt,
        6.86778e-05_prt, 3.24283e-04_prt, 9.48942e-04_prt, 2.09875e-03_prt,
        3.88104e-03_prt, 6.35297e-03_prt, 9.53137e-03_prt, 1.34045e-02_prt,
        1.79422e-02_prt, 2.31033e-02_prt, 2.88412e-02_prt, 3.51073e-02_prt,
        4.18532e-02_prt, 4.90324e-02_prt, 5.66011e-02_prt, 6.45182e-02_prt,
        7.27460e-02_prt, 8.12500e-02_prt, 8.99985e-02_prt, 9.89631e-02_prt,
        1.08118e-01_prt, 1.17439e-01_prt, 1.26907e-01_prt, 1.36501e-01_prt,
        1.46205e-01_prt, 1.56004e-01_prt, 1.65883e-01_prt, 1.75832e-01_prt,
        1.85839e-01_prt, 1.95894e-01_prt, 2.05989e-01_prt, 2.16116e-01_prt,
        2.26268e-01_prt, 2.36440e-01_prt, 2.46627e-01_prt, 2.56824e-01_prt,
        2.67026e-01_prt, 2.77231e-01_prt, 2.87436e-01_prt, 2.97638e-01_prt,
        3.07836e-01_prt, 3.18027e-01_prt, 3.28210e-01_prt, 3.38385e-01_prt,
        3.48550e-01_prt, 3.58706e-01_prt, 3.68851e-01_prt, 3.78987e-01_prt,
        3.89112e-01_prt, 3.99228e-01_prt, 4.09334e-01_prt, 4.19431e-01_prt,
        4.29521e-01_prt, 4.39603e-01_prt, 4.49680e-01_prt, 4.59750e-01_prt,
        4.69817e-01_prt, 4.79880e-01_prt, 4.89940e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 1.93231e-12_prt, 2.73037e-07_prt, 1.78694e-05_prt,
        1.60620e-04_prt, 6.36814e-04_prt, 1.65732e-03_prt, 3.36918e-03_prt,
        5.84631e-03_prt, 9.10512e-03_prt, 1.31238e-02_prt, 1.78577e-02_prt,
        2.32510e-02_prt, 2.92429e-02_prt, 3.57726e-02_prt, 4.27817e-02_prt,
        5.02155e-02_prt, 5.80236e-02_prt, 6.61603e-02_prt, 7.45842e-02_prt,
        8.32580e-02_prt, 9.21486e-02_prt, 1.01226e-01_prt, 1.10465e-01_prt,
        1.19840e-01_prt, 1.29333e-01_prt, 1.38923e-01_prt, 1.48595e-01_prt,
        1.58334e-01_prt, 1.68127e-01_prt, 1.77965e-01_prt, 1.87835e-01_prt,
        1.97731e-01_prt, 2.07644e-01_prt, 2.17567e-01_prt, 2.27496e-01_prt,
        2.37425e-01_prt, 2.47350e-01_prt, 2.57267e-01_prt, 2.67173e-01_prt,
        2.77067e-01_prt, 2.86945e-01_prt, 2.96806e-01_prt, 3.06650e-01_prt,
        3.16474e-01_prt, 3.26280e-01_prt, 3.36065e-01_prt, 3.45830e-01_prt,
        3.55576e-01_prt, 3.65303e-01_prt, 3.75011e-01_prt, 3.84701e-01_prt,
        3.94374e-01_prt, 4.04031e-01_prt, 4.13672e-01_prt, 4.23300e-01_prt,
        4.32916e-01_prt, 4.42520e-01_prt, 4.52115e-01_prt, 4.61702e-01_prt,
        4.71282e-01_prt, 4.80858e-01_prt, 4.90430e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 3.79786e-11_prt, 1.21439e-06_prt, 4.82824e-05_prt,
        3.37662e-04_prt, 1.15028e-03_prt, 2.70359e-03_prt, 5.10744e-03_prt,
        8.38478e-03_prt, 1.25037e-02_prt, 1.74026e-02_prt, 2.30071e-02_prt,
        2.92397e-02_prt, 3.60251e-02_prt, 4.32929e-02_prt, 5.09793e-02_prt,
        5.90267e-02_prt, 6.73840e-02_prt, 7.60060e-02_prt, 8.48530e-02_prt,
        9.38900e-02_prt, 1.03086e-01_prt, 1.12415e-01_prt, 1.21853e-01_prt,
        1.31380e-01_prt, 1.40978e-01_prt, 1.50630e-01_prt, 1.60324e-01_prt,
        1.70049e-01_prt, 1.79792e-01_prt, 1.89547e-01_prt, 1.99304e-01_prt,
        2.09058e-01_prt, 2.18804e-01_prt, 2.28535e-01_prt, 2.38249e-01_prt,
        2.47942e-01_prt, 2.57611e-01_prt, 2.67255e-01_prt, 2.76871e-01_prt,
        2.86458e-01_prt, 2.96015e-01_prt, 3.05543e-01_prt, 3.15040e-01_prt,
        3.24507e-01_prt, 3.33944e-01_prt, 3.43351e-01_prt, 3.52730e-01_prt,
        3.62080e-01_prt, 3.71405e-01_prt, 3.80703e-01_prt, 3.89978e-01_prt,
        3.99229e-01_prt, 4.08460e-01_prt, 4.17671e-01_prt, 4.26864e-01_prt,
        4.36042e-01_prt, 4.45205e-01_prt, 4.54356e-01_prt, 4.63498e-01_prt,
        4.72630e-01_prt, 4.81757e-01_prt, 4.90880e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 5.03976e-10_prt, 4.45642e-06_prt, 1.15052e-04_prt,
        6.47172e-04_prt, 1.93239e-03_prt, 4.15645e-03_prt, 7.36609e-03_prt,
        1.15221e-02_prt, 1.65428e-02_prt, 2.23308e-02_prt, 2.87872e-02_prt,
        3.58187e-02_prt, 4.33409e-02_prt, 5.12788e-02_prt, 5.95667e-02_prt,
        6.81476e-02_prt, 7.69720e-02_prt, 8.59975e-02_prt, 9.51872e-02_prt,
        1.04510e-01_prt, 1.13938e-01_prt, 1.23448e-01_prt, 1.33020e-01_prt,
        1.42637e-01_prt, 1.52284e-01_prt, 1.61949e-01_prt, 1.71620e-01_prt,
        1.81288e-01_prt, 1.90946e-01_prt, 2.00586e-01_prt, 2.10205e-01_prt,
        2.19795e-01_prt, 2.29355e-01_prt, 2.38881e-01_prt, 2.48370e-01_prt,
        2.57821e-01_prt, 2.67232e-01_prt, 2.76602e-01_prt, 2.85931e-01_prt,
        2.95219e-01_prt, 3.04465e-01_prt, 3.13670e-01_prt, 3.22835e-01_prt,
        3.31961e-01_prt, 3.41048e-01_prt, 3.50099e-01_prt, 3.59113e-01_prt,
        3.68093e-01_prt, 3.77041e-01_prt, 3.85957e-01_prt, 3.94845e-01_prt,
        4.03706e-01_prt, 4.12541e-01_prt, 4.21354e-01_prt, 4.30145e-01_prt,
        4.38918e-01_prt, 4.47675e-01_prt, 4.56418e-01_prt, 4.65148e-01_prt,
        4.73870e-01_prt, 4.82584e-01_prt, 4.91293e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 4.76242e-09_prt, 1.38470e-05_prt, 2.45877e-04_prt,
        1.14480e-03_prt, 3.04803e-03_prt, 6.06887e-03_prt, 1.01723e-02_prt,
        1.52539e-02_prt, 2.11867e-02_prt, 2.78440e-02_prt, 3.51097e-02_prt,
        4.28814e-02_prt, 5.10706e-02_prt, 5.96018e-02_prt, 6.84111e-02_prt,
        7.74442e-02_prt, 8.66556e-02_prt, 9.60064e-02_prt, 1.05464e-01_prt,
        1.15002e-01_prt, 1.24595e-01_prt, 1.34225e-01_prt, 1.43875e-01_prt,
        1.53531e-01_prt, 1.63181e-01_prt, 1.72816e-01_prt, 1.82427e-01_prt,
        1.92008e-01_prt, 2.01552e-01_prt, 2.11056e-01_prt, 2.20516e-01_prt,
        2.29928e-01_prt, 2.39291e-01_prt, 2.48603e-01_prt, 2.57862e-01_prt,
        2.67069e-01_prt, 2.76224e-01_prt, 2.85325e-01_prt, 2.94374e-01_prt,
        3.03372e-01_prt, 3.12318e-01_prt, 3.21215e-01_prt, 3.30064e-01_prt,
        3.38867e-01_prt, 3.47624e-01_prt, 3.56338e-01_prt, 3.65011e-01_prt,
        3.73644e-01_prt, 3.82241e-01_prt, 3.90802e-01_prt, 3.99330e-01_prt,
        4.07828e-01_prt, 4.16298e-01_prt, 4.24742e-01_prt, 4.33163e-01_prt,
        4.41564e-01_prt, 4.49946e-01_prt, 4.58312e-01_prt, 4.66665e-01_prt,
        4.75008e-01_prt, 4.83343e-01_prt, 4.91673e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 3.35560e-08_prt, 3.72534e-05_prt, 4.78125e-04_prt,
        1.88884e-03_prt, 4.55138e-03_prt, 8.47206e-03_prt, 1.35259e-02_prt,
        1.95478e-02_prt, 2.63739e-02_prt, 3.38573e-02_prt, 4.18721e-02_prt,
        5.03127e-02_prt, 5.90915e-02_prt, 6.81360e-02_prt, 7.73863e-02_prt,
        8.67932e-02_prt, 9.63157e-02_prt, 1.05920e-01_prt, 1.15579e-01_prt,
        1.25268e-01_prt, 1.34970e-01_prt, 1.44667e-01_prt, 1.54348e-01_prt,
        1.64000e-01_prt, 1.73616e-01_prt, 1.83188e-01_prt, 1.92711e-01_prt,
        2.02180e-01_prt, 2.11590e-01_prt, 2.20941e-01_prt, 2.30229e-01_prt,
        2.39453e-01_prt, 2.48613e-01_prt, 2.57708e-01_prt, 2.66738e-01_prt,
        2.75704e-01_prt, 2.84606e-01_prt, 2.93446e-01_prt, 3.02224e-01_prt,
        3.10943e-01_prt, 3.19603e-01_prt, 3.28207e-01_prt, 3.36757e-01_prt,
        3.45254e-01_prt, 3.53701e-01_prt, 3.62100e-01_prt, 3.70454e-01_prt,
        3.78764e-01_prt, 3.87033e-01_prt, 3.95264e-01_prt, 4.03460e-01_prt,
        4.11622e-01_prt, 4.19754e-01_prt, 4.27858e-01_prt, 4.35938e-01_prt,
        4.43994e-01_prt, 4.52032e-01_prt, 4.60052e-01_prt, 4.68059e-01_prt,
        4.76054e-01_prt, 4.84041e-01_prt, 4.92022e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 1.83437e-07_prt, 8.84690e-05_prt, 8.56627e-04_prt,
        2.93343e-03_prt, 6.47963e-03_prt, 1.13727e-02_prt, 1.74006e-02_prt,
        2.43477e-02_prt, 3.20245e-02_prt, 4.02730e-02_prt, 4.89649e-02_prt,
        5.79966e-02_prt, 6.72849e-02_prt, 7.67632e-02_prt, 8.63775e-02_prt,
        9.60845e-02_prt, 1.05849e-01_prt, 1.15643e-01_prt, 1.25444e-01_prt,
        1.35231e-01_prt, 1.44992e-01_prt, 1.54713e-01_prt, 1.64385e-01_prt,
        1.74000e-01_prt, 1.83552e-01_prt, 1.93036e-01_prt, 2.02448e-01_prt,
        2.11787e-01_prt, 2.21050e-01_prt, 2.30236e-01_prt, 2.39345e-01_prt,
        2.48377e-01_prt, 2.57331e-01_prt, 2.66210e-01_prt, 2.75013e-01_prt,
        2.83743e-01_prt, 2.92401e-01_prt, 3.00988e-01_prt, 3.09507e-01_prt,
        3.17960e-01_prt, 3.26348e-01_prt, 3.34675e-01_prt, 3.42943e-01_prt,
        3.51154e-01_prt, 3.59310e-01_prt, 3.67415e-01_prt, 3.75470e-01_prt,
        3.83480e-01_prt, 3.91446e-01_prt, 3.99371e-01_prt, 4.07258e-01_prt,
        4.15111e-01_prt, 4.22931e-01_prt, 4.30721e-01_prt, 4.38486e-01_prt,
        4.46227e-01_prt, 4.53947e-01_prt, 4.61649e-01_prt, 4.69337e-01_prt,
        4.77013e-01_prt, 4.84680e-01_prt, 4.92342e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 8.05120e-07_prt, 1.88562e-04_prt, 1.42938e-03_prt,
        4.32199e-03_prt, 8.84910e-03_prt, 1.47535e-02_prt, 2.17477e-02_prt,
        2.95801e-02_prt, 3.80469e-02_prt, 4.69880e-02_prt, 5.62785e-02_prt,
        6.58213e-02_prt, 7.55404e-02_prt, 8.53767e-02_prt, 9.52835e-02_prt,
        1.05224e-01_prt, 1.15169e-01_prt, 1.25097e-01_prt, 1.34988e-01_prt,
        1.44829e-01_prt, 1.54608e-01_prt, 1.64317e-01_prt, 1.73949e-01_prt,
        1.83499e-01_prt, 1.92964e-01_prt, 2.02341e-01_prt, 2.11628e-01_prt,
        2.20825e-01_prt, 2.29931e-01_prt, 2.38947e-01_prt, 2.47873e-01_prt,
        2.56711e-01_prt, 2.65462e-01_prt, 2.74128e-01_prt, 2.82710e-01_prt,
        2.91212e-01_prt, 2.99634e-01_prt, 3.07980e-01_prt, 3.16252e-01_prt,
        3.24452e-01_prt, 3.32584e-01_prt, 3.40650e-01_prt, 3.48652e-01_prt,
        3.56595e-01_prt, 3.64480e-01_prt, 3.72310e-01_prt, 3.80089e-01_prt,
        3.87820e-01_prt, 3.95504e-01_prt, 4.03147e-01_prt, 4.10749e-01_prt,
        4.18315e-01_prt, 4.25848e-01_prt, 4.33350e-01_prt, 4.40825e-01_prt,
        4.48275e-01_prt, 4.55704e-01_prt, 4.63115e-01_prt, 4.70510e-01_prt,
        4.77893e-01_prt, 4.85267e-01_prt, 4.92635e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 2.92242e-06_prt, 3.65922e-04_prt, 2.24206e-03_prt,
        6.08232e-03_prt, 1.16544e-02_prt, 1.85759e-02_prt, 2.65021e-02_prt,
        3.51606e-02_prt, 4.43448e-02_prt, 5.38996e-02_prt, 6.37086e-02_prt,
        7.36840e-02_prt, 8.37593e-02_prt, 9.38837e-02_prt, 1.04018e-01_prt,
        1.14133e-01_prt, 1.24206e-01_prt, 1.34219e-01_prt, 1.44158e-01_prt,
        1.54014e-01_prt, 1.63779e-01_prt, 1.73447e-01_prt, 1.83015e-01_prt,
        1.92480e-01_prt, 2.01841e-01_prt, 2.11097e-01_prt, 2.20249e-01_prt,
        2.29296e-01_prt, 2.38241e-01_prt, 2.47084e-01_prt, 2.55827e-01_prt,
        2.64474e-01_prt, 2.73025e-01_prt, 2.81484e-01_prt, 2.89853e-01_prt,
        2.98135e-01_prt, 3.06333e-01_prt, 3.14449e-01_prt, 3.22487e-01_prt,
        3.30449e-01_prt, 3.38339e-01_prt, 3.46160e-01_prt, 3.53915e-01_prt,
        3.61607e-01_prt, 3.69240e-01_prt, 3.76815e-01_prt, 3.84338e-01_prt,
        3.91810e-01_prt, 3.99234e-01_prt, 4.06615e-01_prt, 4.13955e-01_prt,
        4.21258e-01_prt, 4.28526e-01_prt, 4.35762e-01_prt, 4.42971e-01_prt,
        4.50154e-01_prt, 4.57316e-01_prt, 4.64458e-01_prt, 4.71585e-01_prt,
        4.78700e-01_prt, 4.85805e-01_prt, 4.92904e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 8.99958e-06_prt, 6.54588e-04_prt, 3.33256e-03_prt,
        8.22390e-03_prt, 1.48698e-02_prt, 2.27846e-02_prt, 3.15873e-02_prt,
        4.09999e-02_prt, 5.08222e-02_prt, 6.09102e-02_prt, 7.11594e-02_prt,
        8.14935e-02_prt, 9.18562e-02_prt, 1.02206e-01_prt, 1.12511e-01_prt,
        1.22750e-01_prt, 1.32904e-01_prt, 1.42962e-01_prt, 1.52914e-01_prt,
        1.62754e-01_prt, 1.72479e-01_prt, 1.82084e-01_prt, 1.91570e-01_prt,
        2.00935e-01_prt, 2.10180e-01_prt, 2.19307e-01_prt, 2.28316e-01_prt,
        2.37211e-01_prt, 2.45992e-01_prt, 2.54663e-01_prt, 2.63227e-01_prt,
        2.71686e-01_prt, 2.80044e-01_prt, 2.88304e-01_prt, 2.96468e-01_prt,
        3.04541e-01_prt, 3.12525e-01_prt, 3.20424e-01_prt, 3.28241e-01_prt,
        3.35980e-01_prt, 3.43644e-01_prt, 3.51237e-01_prt, 3.58761e-01_prt,
        3.66220e-01_prt, 3.73617e-01_prt, 3.80957e-01_prt, 3.88241e-01_prt,
        3.95474e-01_prt, 4.02659e-01_prt, 4.09799e-01_prt, 4.16897e-01_prt,
        4.23957e-01_prt, 4.30981e-01_prt, 4.37974e-01_prt, 4.44938e-01_prt,
        4.51876e-01_prt, 4.58792e-01_prt, 4.65689e-01_prt, 4.72571e-01_prt,
        4.79439e-01_prt, 4.86298e-01_prt, 4.93151e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 2.40362e-05_prt, 1.09102e-03_prt, 4.72659e-03_prt,
        1.07377e-02_prt, 1.84532e-02_prt, 2.73132e-02_prt, 3.69221e-02_prt,
        4.70091e-02_prt, 5.73883e-02_prt, 6.79309e-02_prt, 7.85465e-02_prt,
        8.91714e-02_prt, 9.97598e-02_prt, 1.10279e-01_prt, 1.20706e-01_prt,
        1.31025e-01_prt, 1.41223e-01_prt, 1.51293e-01_prt, 1.61230e-01_prt,
        1.71030e-01_prt, 1.80694e-01_prt, 1.90220e-01_prt, 1.99610e-01_prt,
        2.08865e-01_prt, 2.17987e-01_prt, 2.26979e-01_prt, 2.35844e-01_prt,
        2.44584e-01_prt, 2.53204e-01_prt, 2.61706e-01_prt, 2.70095e-01_prt,
        2.78373e-01_prt, 2.86545e-01_prt, 2.94614e-01_prt, 3.02583e-01_prt,
        3.10458e-01_prt, 3.18240e-01_prt, 3.25935e-01_prt, 3.33545e-01_prt,
        3.41075e-01_prt, 3.48528e-01_prt, 3.55907e-01_prt, 3.63216e-01_prt,
        3.70459e-01_prt, 3.77639e-01_prt, 3.84760e-01_prt, 3.91825e-01_prt,
        3.98837e-01_prt, 4.05801e-01_prt, 4.12719e-01_prt, 4.19594e-01_prt,
        4.26431e-01_prt, 4.33232e-01_prt, 4.40000e-01_prt, 4.46740e-01_prt,
        4.53454e-01_prt, 4.60145e-01_prt, 4.66817e-01_prt, 4.73473e-01_prt,
        4.80116e-01_prt, 4.86749e-01_prt, 4.93376e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 5.67455e-05_prt, 1.70998e-03_prt, 6.43514e-03_prt,
        1.35980e-02_prt, 2.23497e-02_prt, 3.20889e-02_prt, 4.24247e-02_prt,
        5.31040e-02_prt, 6.39605e-02_prt, 7.48834e-02_prt, 8.57977e-02_prt,
        9.66524e-02_prt, 1.07412e-01_prt, 1.18054e-01_prt, 1.28561e-01_prt,
        1.38923e-01_prt, 1.49134e-01_prt, 1.59189e-01_prt, 1.69089e-01_prt,
        1.78831e-01_prt, 1.88419e-01_prt, 1.97854e-01_prt, 2.07139e-01_prt,
        2.16278e-01_prt, 2.25272e-01_prt, 2.34128e-01_prt, 2.42848e-01_prt,
        2.51436e-01_prt, 2.59897e-01_prt, 2.68236e-01_prt, 2.76455e-01_prt,
        2.84560e-01_prt, 2.92554e-01_prt, 3.00442e-01_prt, 3.08227e-01_prt,
        3.15914e-01_prt, 3.23508e-01_prt, 3.31011e-01_prt, 3.38428e-01_prt,
        3.45762e-01_prt, 3.53018e-01_prt, 3.60199e-01_prt, 3.67309e-01_prt,
        3.74352e-01_prt, 3.81331e-01_prt, 3.88250e-01_prt, 3.95112e-01_prt,
        4.01921e-01_prt, 4.08681e-01_prt, 4.15395e-01_prt, 4.22065e-01_prt,
        4.28697e-01_prt, 4.35293e-01_prt, 4.41856e-01_prt, 4.48390e-01_prt,
        4.54898e-01_prt, 4.61383e-01_prt, 4.67849e-01_prt, 4.74298e-01_prt,
        4.80735e-01_prt, 4.87162e-01_prt, 4.93583e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 1.20376e-04_prt, 2.54037e-03_prt, 8.45388e-03_prt,
        1.67654e-02_prt, 2.64972e-02_prt, 3.70377e-02_prt, 4.80170e-02_prt,
        5.92074e-02_prt, 7.04660e-02_prt, 8.17011e-02_prt, 9.28535e-02_prt,
        1.03885e-01_prt, 1.14770e-01_prt, 1.25493e-01_prt, 1.36045e-01_prt,
        1.46421e-01_prt, 1.56620e-01_prt, 1.66640e-01_prt, 1.76485e-01_prt,
        1.86156e-01_prt, 1.95657e-01_prt, 2.04993e-01_prt, 2.14168e-01_prt,
        2.23186e-01_prt, 2.32052e-01_prt, 2.40771e-01_prt, 2.49349e-01_prt,
        2.57789e-01_prt, 2.66096e-01_prt, 2.74276e-01_prt, 2.82334e-01_prt,
        2.90273e-01_prt, 2.98099e-01_prt, 3.05816e-01_prt, 3.13428e-01_prt,
        3.20940e-01_prt, 3.28355e-01_prt, 3.35680e-01_prt, 3.42916e-01_prt,
        3.50069e-01_prt, 3.57142e-01_prt, 3.64140e-01_prt, 3.71065e-01_prt,
        3.77923e-01_prt, 3.84716e-01_prt, 3.91449e-01_prt, 3.98125e-01_prt,
        4.04747e-01_prt, 4.11319e-01_prt, 4.17845e-01_prt, 4.24328e-01_prt,
        4.30772e-01_prt, 4.37179e-01_prt, 4.43554e-01_prt, 4.49899e-01_prt,
        4.56219e-01_prt, 4.62515e-01_prt, 4.68792e-01_prt, 4.75053e-01_prt,
        4.81301e-01_prt, 4.87540e-01_prt, 4.93771e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 2.32725e-04_prt, 3.60193e-03_prt, 1.07643e-02_prt,
        2.01911e-02_prt, 3.08302e-02_prt, 4.20879e-02_prt, 5.36273e-02_prt,
        6.52517e-02_prt, 7.68429e-02_prt, 8.83291e-02_prt, 9.96663e-02_prt,
        1.10828e-01_prt, 1.21798e-01_prt, 1.32570e-01_prt, 1.43139e-01_prt,
        1.53506e-01_prt, 1.63672e-01_prt, 1.73641e-01_prt, 1.83418e-01_prt,
        1.93008e-01_prt, 2.02416e-01_prt, 2.11648e-01_prt, 2.20710e-01_prt,
        2.29607e-01_prt, 2.38345e-01_prt, 2.46931e-01_prt, 2.55369e-01_prt,
        2.63666e-01_prt, 2.71826e-01_prt, 2.79855e-01_prt, 2.87759e-01_prt,
        2.95541e-01_prt, 3.03208e-01_prt, 3.10764e-01_prt, 3.18213e-01_prt,
        3.25561e-01_prt, 3.32811e-01_prt, 3.39969e-01_prt, 3.47038e-01_prt,
        3.54022e-01_prt, 3.60926e-01_prt, 3.67754e-01_prt, 3.74509e-01_prt,
        3.81196e-01_prt, 3.87818e-01_prt, 3.94380e-01_prt, 4.00884e-01_prt,
        4.07334e-01_prt, 4.13734e-01_prt, 4.20088e-01_prt, 4.26399e-01_prt,
        4.32670e-01_prt, 4.38905e-01_prt, 4.45107e-01_prt, 4.51280e-01_prt,
        4.57427e-01_prt, 4.63551e-01_prt, 4.69655e-01_prt, 4.75744e-01_prt,
        4.81819e-01_prt, 4.87885e-01_prt, 4.93944e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 4.15090e-04_prt, 4.90330e-03_prt, 1.33363e-02_prt,
        2.38205e-02_prt, 3.52837e-02_prt, 4.71726e-02_prt, 5.91921e-02_prt,
        7.11790e-02_prt, 8.30406e-02_prt, 9.47240e-02_prt, 1.06200e-01_prt,
        1.17453e-01_prt, 1.28476e-01_prt, 1.39268e-01_prt, 1.49832e-01_prt,
        1.60170e-01_prt, 1.70290e-01_prt, 1.80196e-01_prt, 1.89897e-01_prt,
        1.99399e-01_prt, 2.08709e-01_prt, 2.17835e-01_prt, 2.26783e-01_prt,
        2.35561e-01_prt, 2.44174e-01_prt, 2.52630e-01_prt, 2.60934e-01_prt,
        2.69093e-01_prt, 2.77113e-01_prt, 2.84999e-01_prt, 2.92756e-01_prt,
        3.00392e-01_prt, 3.07909e-01_prt, 3.15314e-01_prt, 3.22612e-01_prt,
        3.29806e-01_prt, 3.36903e-01_prt, 3.43906e-01_prt, 3.50819e-01_prt,
        3.57647e-01_prt, 3.64395e-01_prt, 3.71066e-01_prt, 3.77664e-01_prt,
        3.84194e-01_prt, 3.90659e-01_prt, 3.97062e-01_prt, 4.03409e-01_prt,
        4.09701e-01_prt, 4.15943e-01_prt, 4.22139e-01_prt, 4.28292e-01_prt,
        4.34405e-01_prt, 4.40483e-01_prt, 4.46527e-01_prt, 4.52542e-01_prt,
        4.58531e-01_prt, 4.64497e-01_prt, 4.70444e-01_prt, 4.76375e-01_prt,
        4.82292e-01_prt, 4.88200e-01_prt, 4.94102e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 6.90267e-04_prt, 6.44151e-03_prt, 1.61308e-02_prt,
        2.75971e-02_prt, 3.97962e-02_prt, 5.22323e-02_prt, 6.46571e-02_prt,
        7.69417e-02_prt, 8.90185e-02_prt, 1.00853e-01_prt, 1.12429e-01_prt,
        1.23741e-01_prt, 1.34790e-01_prt, 1.45581e-01_prt, 1.56120e-01_prt,
        1.66416e-01_prt, 1.76478e-01_prt, 1.86314e-01_prt, 1.95932e-01_prt,
        2.05343e-01_prt, 2.14554e-01_prt, 2.23573e-01_prt, 2.32410e-01_prt,
        2.41070e-01_prt, 2.49562e-01_prt, 2.57892e-01_prt, 2.66068e-01_prt,
        2.74096e-01_prt, 2.81983e-01_prt, 2.89733e-01_prt, 2.97354e-01_prt,
        3.04851e-01_prt, 3.12229e-01_prt, 3.19493e-01_prt, 3.26649e-01_prt,
        3.33702e-01_prt, 3.40655e-01_prt, 3.47515e-01_prt, 3.54284e-01_prt,
        3.60969e-01_prt, 3.67572e-01_prt, 3.74099e-01_prt, 3.80552e-01_prt,
        3.86937e-01_prt, 3.93257e-01_prt, 3.99516e-01_prt, 4.05718e-01_prt,
        4.11865e-01_prt, 4.17963e-01_prt, 4.24014e-01_prt, 4.30023e-01_prt,
        4.35991e-01_prt, 4.41924e-01_prt, 4.47824e-01_prt, 4.53695e-01_prt,
        4.59539e-01_prt, 4.65361e-01_prt, 4.71164e-01_prt, 4.76951e-01_prt,
        4.82724e-01_prt, 4.88488e-01_prt, 4.94246e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 1.07999e-03_prt, 8.20281e-03_prt, 1.91038e-02_prt,
        3.14654e-02_prt, 4.43117e-02_prt, 5.72152e-02_prt, 6.99770e-02_prt,
        8.25021e-02_prt, 9.47464e-02_prt, 1.06692e-01_prt, 1.18335e-01_prt,
        1.29679e-01_prt, 1.40733e-01_prt, 1.51505e-01_prt, 1.62007e-01_prt,
        1.72251e-01_prt, 1.82247e-01_prt, 1.92006e-01_prt, 2.01540e-01_prt,
        2.10857e-01_prt, 2.19969e-01_prt, 2.28884e-01_prt, 2.37610e-01_prt,
        2.46157e-01_prt, 2.54532e-01_prt, 2.62744e-01_prt, 2.70798e-01_prt,
        2.78702e-01_prt, 2.86463e-01_prt, 2.94086e-01_prt, 3.01579e-01_prt,
        3.08946e-01_prt, 3.16194e-01_prt, 3.23327e-01_prt, 3.30352e-01_prt,
        3.37272e-01_prt, 3.44094e-01_prt, 3.50821e-01_prt, 3.57458e-01_prt,
        3.64009e-01_prt, 3.70480e-01_prt, 3.76873e-01_prt, 3.83194e-01_prt,
        3.89446e-01_prt, 3.95633e-01_prt, 4.01759e-01_prt, 4.07828e-01_prt,
        4.13843e-01_prt, 4.19808e-01_prt, 4.25727e-01_prt, 4.31603e-01_prt,
        4.37440e-01_prt, 4.43240e-01_prt, 4.49008e-01_prt, 4.54747e-01_prt,
        4.60460e-01_prt, 4.66150e-01_prt, 4.71821e-01_prt, 4.77476e-01_prt,
        4.83118e-01_prt, 4.88751e-01_prt, 4.94377e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 1.60237e-03_prt, 1.01645e-02_prt, 2.22089e-02_prt,
        3.53733e-02_prt, 4.87807e-02_prt, 6.20782e-02_prt, 7.51161e-02_prt,
        8.78315e-02_prt, 1.00202e-01_prt, 1.12226e-01_prt, 1.23909e-01_prt,
        1.35265e-01_prt, 1.46305e-01_prt, 1.57046e-01_prt, 1.67501e-01_prt,
        1.77684e-01_prt, 1.87610e-01_prt, 1.97290e-01_prt, 2.06737e-01_prt,
        2.15962e-01_prt, 2.24975e-01_prt, 2.33788e-01_prt, 2.42409e-01_prt,
        2.50847e-01_prt, 2.59111e-01_prt, 2.67209e-01_prt, 2.75148e-01_prt,
        2.82935e-01_prt, 2.90578e-01_prt, 2.98083e-01_prt, 3.05456e-01_prt,
        3.12702e-01_prt, 3.19829e-01_prt, 3.26841e-01_prt, 3.33744e-01_prt,
        3.40543e-01_prt, 3.47242e-01_prt, 3.53846e-01_prt, 3.60361e-01_prt,
        3.66790e-01_prt, 3.73139e-01_prt, 3.79410e-01_prt, 3.85609e-01_prt,
        3.91739e-01_prt, 3.97804e-01_prt, 4.03808e-01_prt, 4.09755e-01_prt,
        4.15649e-01_prt, 4.21493e-01_prt, 4.27291e-01_prt, 4.33046e-01_prt,
        4.38762e-01_prt, 4.44442e-01_prt, 4.50089e-01_prt, 4.55707e-01_prt,
        4.61300e-01_prt, 4.66870e-01_prt, 4.72421e-01_prt, 4.77956e-01_prt,
        4.83478e-01_prt, 4.88991e-01_prt, 4.94497e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 2.26985e-03_prt, 1.22974e-02_prt, 2.53998e-02_prt,
        3.92737e-02_prt, 5.31612e-02_prt, 6.67863e-02_prt, 8.00466e-02_prt,
        9.29094e-02_prt, 1.05372e-01_prt, 1.17447e-01_prt, 1.29149e-01_prt,
        1.40498e-01_prt, 1.51513e-01_prt, 1.62211e-01_prt, 1.72612e-01_prt,
        1.82731e-01_prt, 1.92583e-01_prt, 2.02182e-01_prt, 2.11543e-01_prt,
        2.20677e-01_prt, 2.29596e-01_prt, 2.38310e-01_prt, 2.46830e-01_prt,
        2.55164e-01_prt, 2.63323e-01_prt, 2.71313e-01_prt, 2.79144e-01_prt,
        2.86822e-01_prt, 2.94354e-01_prt, 3.01748e-01_prt, 3.09009e-01_prt,
        3.16144e-01_prt, 3.23159e-01_prt, 3.30059e-01_prt, 3.36849e-01_prt,
        3.43535e-01_prt, 3.50121e-01_prt, 3.56613e-01_prt, 3.63016e-01_prt,
        3.69332e-01_prt, 3.75568e-01_prt, 3.81728e-01_prt, 3.87814e-01_prt,
        3.93833e-01_prt, 3.99786e-01_prt, 4.05679e-01_prt, 4.11515e-01_prt,
        4.17298e-01_prt, 4.23031e-01_prt, 4.28718e-01_prt, 4.34362e-01_prt,
        4.39968e-01_prt, 4.45537e-01_prt, 4.51075e-01_prt, 4.56583e-01_prt,
        4.62066e-01_prt, 4.67527e-01_prt, 4.72968e-01_prt, 4.78394e-01_prt,
        4.83806e-01_prt, 4.89209e-01_prt, 4.94606e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 3.08797e-03_prt, 1.45678e-02_prt, 2.86330e-02_prt,
        4.31255e-02_prt, 5.74186e-02_prt, 7.13124e-02_prt, 8.47486e-02_prt,
        9.77223e-02_prt, 1.10249e-01_prt, 1.22352e-01_prt, 1.34055e-01_prt,
        1.45385e-01_prt, 1.56364e-01_prt, 1.67014e-01_prt, 1.77355e-01_prt,
        1.87406e-01_prt, 1.97184e-01_prt, 2.06704e-01_prt, 2.15980e-01_prt,
        2.25025e-01_prt, 2.33852e-01_prt, 2.42472e-01_prt, 2.50895e-01_prt,
        2.59132e-01_prt, 2.67191e-01_prt, 2.75081e-01_prt, 2.82810e-01_prt,
        2.90386e-01_prt, 2.97816e-01_prt, 3.05106e-01_prt, 3.12264e-01_prt,
        3.19296e-01_prt, 3.26206e-01_prt, 3.33002e-01_prt, 3.39689e-01_prt,
        3.46271e-01_prt, 3.52754e-01_prt, 3.59142e-01_prt, 3.65441e-01_prt,
        3.71655e-01_prt, 3.77787e-01_prt, 3.83844e-01_prt, 3.89828e-01_prt,
        3.95744e-01_prt, 4.01595e-01_prt, 4.07386e-01_prt, 4.13120e-01_prt,
        4.18802e-01_prt, 4.24433e-01_prt, 4.30019e-01_prt, 4.35563e-01_prt,
        4.41068e-01_prt, 4.46537e-01_prt, 4.51974e-01_prt, 4.57382e-01_prt,
        4.62765e-01_prt, 4.68125e-01_prt, 4.73467e-01_prt, 4.78792e-01_prt,
        4.84105e-01_prt, 4.89409e-01_prt, 4.94706e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 4.05505e-03_prt, 1.69406e-02_prt, 3.18685e-02_prt,
        4.68937e-02_prt, 6.15255e-02_prt, 7.56363e-02_prt, 8.92090e-02_prt,
        1.02263e-01_prt, 1.14830e-01_prt, 1.26943e-01_prt, 1.38635e-01_prt,
        1.49935e-01_prt, 1.60871e-01_prt, 1.71468e-01_prt, 1.81747e-01_prt,
        1.91730e-01_prt, 2.01433e-01_prt, 2.10874e-01_prt, 2.20068e-01_prt,
        2.29028e-01_prt, 2.37767e-01_prt, 2.46298e-01_prt, 2.54630e-01_prt,
        2.62775e-01_prt, 2.70740e-01_prt, 2.78536e-01_prt, 2.86171e-01_prt,
        2.93651e-01_prt, 3.00986e-01_prt, 3.08180e-01_prt, 3.15242e-01_prt,
        3.22178e-01_prt, 3.28993e-01_prt, 3.35693e-01_prt, 3.42284e-01_prt,
        3.48770e-01_prt, 3.55158e-01_prt, 3.61451e-01_prt, 3.67655e-01_prt,
        3.73774e-01_prt, 3.79813e-01_prt, 3.85775e-01_prt, 3.91665e-01_prt,
        3.97487e-01_prt, 4.03245e-01_prt, 4.08943e-01_prt, 4.14584e-01_prt,
        4.20173e-01_prt, 4.25712e-01_prt, 4.31206e-01_prt, 4.36657e-01_prt,
        4.42070e-01_prt, 4.47448e-01_prt, 4.52793e-01_prt, 4.58110e-01_prt,
        4.63402e-01_prt, 4.68671e-01_prt, 4.73921e-01_prt, 4.79156e-01_prt,
        4.84378e-01_prt, 4.89590e-01_prt, 4.94797e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 5.16268e-03_prt, 1.93807e-02_prt, 3.50712e-02_prt,
        5.05499e-02_prt, 6.54611e-02_prt, 7.97445e-02_prt, 9.34202e-02_prt,
        1.06530e-01_prt, 1.19118e-01_prt, 1.31227e-01_prt, 1.42897e-01_prt,
        1.54160e-01_prt, 1.65049e-01_prt, 1.75589e-01_prt, 1.85806e-01_prt,
        1.95719e-01_prt, 2.05350e-01_prt, 2.14714e-01_prt, 2.23829e-01_prt,
        2.32708e-01_prt, 2.41364e-01_prt, 2.49810e-01_prt, 2.58057e-01_prt,
        2.66114e-01_prt, 2.73993e-01_prt, 2.81701e-01_prt, 2.89247e-01_prt,
        2.96640e-01_prt, 3.03885e-01_prt, 3.10992e-01_prt, 3.17965e-01_prt,
        3.24812e-01_prt, 3.31539e-01_prt, 3.38151e-01_prt, 3.44654e-01_prt,
        3.51053e-01_prt, 3.57353e-01_prt, 3.63559e-01_prt, 3.69676e-01_prt,
        3.75708e-01_prt, 3.81660e-01_prt, 3.87536e-01_prt, 3.93341e-01_prt,
        3.99077e-01_prt, 4.04750e-01_prt, 4.10362e-01_prt, 4.15919e-01_prt,
        4.21423e-01_prt, 4.26878e-01_prt, 4.32287e-01_prt, 4.37655e-01_prt,
        4.42984e-01_prt, 4.48278e-01_prt, 4.53540e-01_prt, 4.58773e-01_prt,
        4.63982e-01_prt, 4.69168e-01_prt, 4.74335e-01_prt, 4.79487e-01_prt,
        4.84626e-01_prt, 4.89756e-01_prt, 4.94880e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 6.39685e-03_prt, 2.18546e-02_prt, 3.82111e-02_prt,
        5.40718e-02_prt, 6.92104e-02_prt, 8.36286e-02_prt, 9.73795e-02_prt,
        1.10524e-01_prt, 1.23118e-01_prt, 1.35213e-01_prt, 1.46852e-01_prt,
        1.58074e-01_prt, 1.68912e-01_prt, 1.79395e-01_prt, 1.89548e-01_prt,
        1.99395e-01_prt, 2.08955e-01_prt, 2.18246e-01_prt, 2.27285e-01_prt,
        2.36086e-01_prt, 2.44664e-01_prt, 2.53030e-01_prt, 2.61197e-01_prt,
        2.69173e-01_prt, 2.76971e-01_prt, 2.84597e-01_prt, 2.92062e-01_prt,
        2.99372e-01_prt, 3.06536e-01_prt, 3.13561e-01_prt, 3.20453e-01_prt,
        3.27218e-01_prt, 3.33864e-01_prt, 3.40395e-01_prt, 3.46817e-01_prt,
        3.53135e-01_prt, 3.59355e-01_prt, 3.65481e-01_prt, 3.71519e-01_prt,
        3.77472e-01_prt, 3.83345e-01_prt, 3.89142e-01_prt, 3.94868e-01_prt,
        4.00526e-01_prt, 4.06121e-01_prt, 4.11656e-01_prt, 4.17135e-01_prt,
        4.22561e-01_prt, 4.27939e-01_prt, 4.33272e-01_prt, 4.38563e-01_prt,
        4.43816e-01_prt, 4.49033e-01_prt, 4.54219e-01_prt, 4.59377e-01_prt,
        4.64510e-01_prt, 4.69620e-01_prt, 4.74712e-01_prt, 4.79788e-01_prt,
        4.84852e-01_prt, 4.89907e-01_prt, 4.94955e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 7.73935e-03_prt, 2.43318e-02_prt, 4.12635e-02_prt,
        5.74425e-02_prt, 7.27638e-02_prt, 8.72850e-02_prt, 1.01088e-01_prt,
        1.14250e-01_prt, 1.26839e-01_prt, 1.38911e-01_prt, 1.50515e-01_prt,
        1.61692e-01_prt, 1.72478e-01_prt, 1.82903e-01_prt, 1.92995e-01_prt,
        2.02776e-01_prt, 2.12267e-01_prt, 2.21488e-01_prt, 2.30456e-01_prt,
        2.39184e-01_prt, 2.47688e-01_prt, 2.55980e-01_prt, 2.64071e-01_prt,
        2.71973e-01_prt, 2.79694e-01_prt, 2.87245e-01_prt, 2.94634e-01_prt,
        3.01869e-01_prt, 3.08958e-01_prt, 3.15907e-01_prt, 3.22724e-01_prt,
        3.29415e-01_prt, 3.35985e-01_prt, 3.42442e-01_prt, 3.48790e-01_prt,
        3.55035e-01_prt, 3.61181e-01_prt, 3.67234e-01_prt, 3.73199e-01_prt,
        3.79079e-01_prt, 3.84880e-01_prt, 3.90605e-01_prt, 3.96259e-01_prt,
        4.01846e-01_prt, 4.07370e-01_prt, 4.12834e-01_prt, 4.18242e-01_prt,
        4.23598e-01_prt, 4.28906e-01_prt, 4.34169e-01_prt, 4.39390e-01_prt,
        4.44573e-01_prt, 4.49722e-01_prt, 4.54838e-01_prt, 4.59927e-01_prt,
        4.64990e-01_prt, 4.70032e-01_prt, 4.75055e-01_prt, 4.80063e-01_prt,
        4.85058e-01_prt, 4.90044e-01_prt, 4.95023e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 9.16934e-03_prt, 2.67855e-02_prt, 4.42088e-02_prt,
        6.06502e-02_prt, 7.61162e-02_prt, 9.07138e-02_prt, 1.04550e-01_prt,
        1.17718e-01_prt, 1.30292e-01_prt, 1.42335e-01_prt, 1.53900e-01_prt,
        1.65031e-01_prt, 1.75764e-01_prt, 1.86132e-01_prt, 1.96163e-01_prt,
        2.05881e-01_prt, 2.15308e-01_prt, 2.24462e-01_prt, 2.33362e-01_prt,
        2.42022e-01_prt, 2.50457e-01_prt, 2.58679e-01_prt, 2.66701e-01_prt,
        2.74532e-01_prt, 2.82184e-01_prt, 2.89665e-01_prt, 2.96984e-01_prt,
        3.04149e-01_prt, 3.11168e-01_prt, 3.18048e-01_prt, 3.24796e-01_prt,
        3.31418e-01_prt, 3.37920e-01_prt, 3.44309e-01_prt, 3.50589e-01_prt,
        3.56766e-01_prt, 3.62845e-01_prt, 3.68832e-01_prt, 3.74729e-01_prt,
        3.80544e-01_prt, 3.86278e-01_prt, 3.91938e-01_prt, 3.97527e-01_prt,
        4.03048e-01_prt, 4.08507e-01_prt, 4.13906e-01_prt, 4.19250e-01_prt,
        4.24543e-01_prt, 4.29786e-01_prt, 4.34985e-01_prt, 4.40143e-01_prt,
        4.45263e-01_prt, 4.50348e-01_prt, 4.55402e-01_prt, 4.60427e-01_prt,
        4.65428e-01_prt, 4.70407e-01_prt, 4.75367e-01_prt, 4.80312e-01_prt,
        4.85245e-01_prt, 4.90168e-01_prt, 4.95086e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 1.06648e-02_prt, 2.91925e-02_prt, 4.70320e-02_prt,
        6.36874e-02_prt, 7.92659e-02_prt, 9.39181e-02_prt, 1.07773e-01_prt,
        1.20936e-01_prt, 1.33488e-01_prt, 1.45499e-01_prt, 1.57023e-01_prt,
        1.68106e-01_prt, 1.78787e-01_prt, 1.89099e-01_prt, 1.99072e-01_prt,
        2.08730e-01_prt, 2.18095e-01_prt, 2.27187e-01_prt, 2.36023e-01_prt,
        2.44619e-01_prt, 2.52990e-01_prt, 2.61147e-01_prt, 2.69104e-01_prt,
        2.76871e-01_prt, 2.84458e-01_prt, 2.91874e-01_prt, 2.99129e-01_prt,
        3.06230e-01_prt, 3.13185e-01_prt, 3.20001e-01_prt, 3.26686e-01_prt,
        3.33245e-01_prt, 3.39684e-01_prt, 3.46010e-01_prt, 3.52228e-01_prt,
        3.58344e-01_prt, 3.64361e-01_prt, 3.70286e-01_prt, 3.76124e-01_prt,
        3.81877e-01_prt, 3.87552e-01_prt, 3.93151e-01_prt, 3.98680e-01_prt,
        4.04143e-01_prt, 4.09542e-01_prt, 4.14883e-01_prt, 4.20168e-01_prt,
        4.25402e-01_prt, 4.30587e-01_prt, 4.35728e-01_prt, 4.40828e-01_prt,
        4.45890e-01_prt, 4.50918e-01_prt, 4.55914e-01_prt, 4.60882e-01_prt,
        4.65826e-01_prt, 4.70748e-01_prt, 4.75651e-01_prt, 4.80540e-01_prt,
        4.85415e-01_prt, 4.90282e-01_prt, 4.95143e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 1.22036e-02_prt, 3.15337e-02_prt, 4.97224e-02_prt,
        6.65502e-02_prt, 8.22144e-02_prt, 9.69035e-02_prt, 1.10765e-01_prt,
        1.23915e-01_prt, 1.36441e-01_prt, 1.48416e-01_prt, 1.59898e-01_prt,
        1.70933e-01_prt, 1.81564e-01_prt, 1.91823e-01_prt, 2.01740e-01_prt,
        2.11341e-01_prt, 2.20648e-01_prt, 2.29681e-01_prt, 2.38457e-01_prt,
        2.46994e-01_prt, 2.55304e-01_prt, 2.63402e-01_prt, 2.71299e-01_prt,
        2.79006e-01_prt, 2.86533e-01_prt, 2.93890e-01_prt, 3.01085e-01_prt,
        3.08127e-01_prt, 3.15024e-01_prt, 3.21782e-01_prt, 3.28408e-01_prt,
        3.34909e-01_prt, 3.41291e-01_prt, 3.47560e-01_prt, 3.53722e-01_prt,
        3.59780e-01_prt, 3.65742e-01_prt, 3.71611e-01_prt, 3.77393e-01_prt,
        3.83091e-01_prt, 3.88711e-01_prt, 3.94256e-01_prt, 3.99730e-01_prt,
        4.05139e-01_prt, 4.10484e-01_prt, 4.15771e-01_prt, 4.21003e-01_prt,
        4.26183e-01_prt, 4.31316e-01_prt, 4.36404e-01_prt, 4.41451e-01_prt,
        4.46461e-01_prt, 4.51436e-01_prt, 4.56380e-01_prt, 4.61296e-01_prt,
        4.66188e-01_prt, 4.71058e-01_prt, 4.75910e-01_prt, 4.80746e-01_prt,
        4.85570e-01_prt, 4.90385e-01_prt, 4.95194e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 1.37649e-02_prt, 3.37939e-02_prt, 5.22729e-02_prt,
        6.92381e-02_prt, 8.49658e-02_prt, 9.96773e-02_prt, 1.13537e-01_prt,
        1.26668e-01_prt, 1.39164e-01_prt, 1.51102e-01_prt, 1.62541e-01_prt,
        1.73530e-01_prt, 1.84111e-01_prt, 1.94319e-01_prt, 2.04184e-01_prt,
        2.13731e-01_prt, 2.22983e-01_prt, 2.31961e-01_prt, 2.40683e-01_prt,
        2.49163e-01_prt, 2.57418e-01_prt, 2.65461e-01_prt, 2.73302e-01_prt,
        2.80954e-01_prt, 2.88426e-01_prt, 2.95728e-01_prt, 3.02869e-01_prt,
        3.09857e-01_prt, 3.16700e-01_prt, 3.23404e-01_prt, 3.29977e-01_prt,
        3.36426e-01_prt, 3.42755e-01_prt, 3.48972e-01_prt, 3.55081e-01_prt,
        3.61089e-01_prt, 3.66999e-01_prt, 3.72817e-01_prt, 3.78548e-01_prt,
        3.84196e-01_prt, 3.89765e-01_prt, 3.95261e-01_prt, 4.00686e-01_prt,
        4.06045e-01_prt, 4.11341e-01_prt, 4.16579e-01_prt, 4.21762e-01_prt,
        4.26894e-01_prt, 4.31979e-01_prt, 4.37019e-01_prt, 4.42018e-01_prt,
        4.46980e-01_prt, 4.51907e-01_prt, 4.56804e-01_prt, 4.61673e-01_prt,
        4.66517e-01_prt, 4.71340e-01_prt, 4.76145e-01_prt, 4.80934e-01_prt,
        4.85711e-01_prt, 4.90479e-01_prt, 4.95241e-01_prt, 5.00000e-01_prt,
        0.00000e+00_prt, 1.53294e-02_prt, 3.59616e-02_prt, 5.46799e-02_prt,
        7.17528e-02_prt, 8.75258e-02_prt, 1.02248e-01_prt, 1.16098e-01_prt,
        1.29206e-01_prt, 1.41671e-01_prt, 1.53571e-01_prt, 1.64968e-01_prt,
        1.75913e-01_prt, 1.86446e-01_prt, 1.96606e-01_prt, 2.06420e-01_prt,
        2.15917e-01_prt, 2.25118e-01_prt, 2.34045e-01_prt, 2.42715e-01_prt,
        2.51144e-01_prt, 2.59348e-01_prt, 2.67339e-01_prt, 2.75129e-01_prt,
        2.82730e-01_prt, 2.90152e-01_prt, 2.97404e-01_prt, 3.04495e-01_prt,
        3.11433e-01_prt, 3.18226e-01_prt, 3.24882e-01_prt, 3.31406e-01_prt,
        3.37806e-01_prt, 3.44088e-01_prt, 3.50257e-01_prt, 3.56319e-01_prt,
        3.62279e-01_prt, 3.68143e-01_prt, 3.73914e-01_prt, 3.79599e-01_prt,
        3.85201e-01_prt, 3.90725e-01_prt, 3.96175e-01_prt, 4.01555e-01_prt,
        4.06869e-01_prt, 4.12121e-01_prt, 4.17314e-01_prt, 4.22453e-01_prt,
        4.27541e-01_prt, 4.32581e-01_prt, 4.37578e-01_prt, 4.42533e-01_prt,
        4.47452e-01_prt, 4.52336e-01_prt, 4.57189e-01_prt, 4.62015e-01_prt,
        4.66816e-01_prt, 4.71596e-01_prt, 4.76358e-01_prt, 4.81105e-01_prt,
        4.85839e-01_prt, 4.90564e-01_prt, 4.95284e-01_prt, 5.00000e-01_prt};

    m_pair_prod_table = BW_pair_prod_table{pair_prod_params, vals};
}
//============================================
