/* Copyright 2019 Luca Fedeli, Maxence Thevenet
 *
 * This file is part of WarpX.
 *
 * License: BSD-3-Clause-LBNL
 */
#include "QuantumSyncEngineWrapper.H"

#include "Utils/TextMsg.H"

#include <AMReX.H>
#include <AMReX_BLassert.H>
#include <AMReX_GpuDevice.H>

#include "picsar_qed/physics/quantum_sync/quantum_sync_engine_tables.hpp"
//Functions needed to generate a new table
#ifdef WARPX_QED_TABLE_GEN
#   include <picsar_qed/physics/quantum_sync/quantum_sync_engine_tables_generator.hpp>
#endif
#include "picsar_qed/utils/serialization.hpp"

#include <algorithm>
#include <cmath>
#include <cstdint>
#include <initializer_list>
#include <iosfwd>
#include <vector>

using namespace std;
using namespace amrex;
namespace pxr_sr = picsar::multi_physics::utils::serialization;

//This file provides a wrapper around the quantum_sync engine
//provided by the PICSAR library

// Factory class =============================

QuantumSynchrotronGetOpticalDepth
QuantumSynchrotronEngine::build_optical_depth_functor ()
{
    return {};
}

QuantumSynchrotronEvolveOpticalDepth QuantumSynchrotronEngine::build_evolve_functor ()
{
    AMREX_ALWAYS_ASSERT(m_lookup_tables_initialized);

    return {m_dndt_table.get_view(), m_qs_minimum_chi_part};
}

QuantumSynchrotronPhotonEmission QuantumSynchrotronEngine::build_phot_em_functor ()
{
    AMREX_ALWAYS_ASSERT(m_lookup_tables_initialized);

    return {m_phot_em_table.get_view()};

}

bool QuantumSynchrotronEngine::are_lookup_tables_initialized () const
{
    return m_lookup_tables_initialized;
}

bool
QuantumSynchrotronEngine::init_lookup_tables_from_raw_data (
    const vector<char>& raw_data,
    const amrex::ParticleReal qs_minimum_chi_part)
{
    auto raw_iter = raw_data.begin();
    const auto size_first = pxr_sr::get_out<uint64_t>(raw_iter);
    if(size_first <= 0 || size_first >= raw_data.size() ) { return false; }

    const auto raw_dndt_table = vector<char>{
        raw_iter, raw_iter+static_cast<long>(size_first)};

    const auto raw_phot_em_table = vector<char>{
        raw_iter+static_cast<long>(size_first), raw_data.end()};

    m_dndt_table = QS_dndt_table{raw_dndt_table};
    m_phot_em_table = QS_phot_em_table{raw_phot_em_table};

    if (!m_dndt_table.is_init() || !m_phot_em_table.is_init()) {
        return false;
    }

    m_qs_minimum_chi_part = qs_minimum_chi_part;

    amrex::Gpu::synchronize();

    m_lookup_tables_initialized = true;

    return true;
}

void QuantumSynchrotronEngine::init_builtin_tables(
    const amrex::ParticleReal qs_minimum_chi_part)
{
    init_builtin_dndt_table();
    init_builtin_phot_em_table();
    m_qs_minimum_chi_part = qs_minimum_chi_part;

    m_lookup_tables_initialized = true;
}

vector<char> QuantumSynchrotronEngine::export_lookup_tables_data () const
{
    if(!m_lookup_tables_initialized) {
        return vector<char>{};
    }

    const auto data_dndt = m_dndt_table.serialize();
    const auto data_phot_em = m_phot_em_table.serialize();

    const uint64_t size_first = data_dndt.size();

    vector<char> res{};
    pxr_sr::put_in(size_first, res);
    for (const auto& tmp : data_dndt) {
        pxr_sr::put_in(tmp, res);
    }
    for (const auto& tmp : data_phot_em) {
        pxr_sr::put_in(tmp, res);
    }

    return res;
}

PicsarQuantumSyncCtrl
QuantumSynchrotronEngine::get_default_ctrl() const
{
    namespace pxr_qs = picsar::multi_physics::phys::quantum_sync;
    return PicsarQuantumSyncCtrl{
        pxr_qs::default_dndt_lookup_table_params<amrex::ParticleReal>,
        pxr_qs::default_photon_emission_lookup_table_params<amrex::ParticleReal>
    };
}

amrex::ParticleReal
QuantumSynchrotronEngine::get_minimum_chi_part() const
{
    return m_qs_minimum_chi_part;
}

void QuantumSynchrotronEngine::compute_lookup_tables (
    PicsarQuantumSyncCtrl ctrl,
    const amrex::ParticleReal qs_minimum_chi_part)
{
#ifdef WARPX_QED_TABLE_GEN
    m_dndt_table = QS_dndt_table{ctrl.dndt_params};
    m_dndt_table.generate(true); //Progress bar is displayed
    m_phot_em_table = QS_phot_em_table{ctrl.phot_em_params};
    m_phot_em_table.generate(true); //Progress bar is displayed
    m_qs_minimum_chi_part = qs_minimum_chi_part;

    amrex::Gpu::synchronize();

    m_lookup_tables_initialized = true;
#else
    amrex::ignore_unused(ctrl, qs_minimum_chi_part);
    WARPX_ABORT_WITH_MESSAGE("WarpX was not compiled with table generation support!");
#endif
}

void QuantumSynchrotronEngine::init_builtin_dndt_table()
{
    constexpr auto default_chi_part_min = 1.0e-3_prt;
    constexpr auto default_chi_part_max = 200.0_prt;
    constexpr auto default_chi_part_how_many = 64;

    QS_dndt_table_params dndt_params;
    dndt_params.chi_part_min = default_chi_part_min;
    dndt_params.chi_part_max = default_chi_part_max;
    dndt_params.chi_part_how_many = default_chi_part_how_many;


    const auto vals = std::vector<amrex::ParticleReal>{
        -6.13623e+00_prt, -5.94268e+00_prt, -5.74917e+00_prt, -5.55571e+00_prt,
        -5.36231e+00_prt, -5.16898e+00_prt, -4.97575e+00_prt, -4.78262e+00_prt,
        -4.58961e+00_prt, -4.39677e+00_prt, -4.20410e+00_prt, -4.01166e+00_prt,
        -3.81948e+00_prt, -3.62762e+00_prt, -3.43612e+00_prt, -3.24506e+00_prt,
        -3.05451e+00_prt, -2.86456e+00_prt, -2.67529e+00_prt, -2.48681e+00_prt,
        -2.29924e+00_prt, -2.11267e+00_prt, -1.92724e+00_prt, -1.74307e+00_prt,
        -1.56028e+00_prt, -1.37897e+00_prt, -1.19927e+00_prt, -1.02127e+00_prt,
        -8.45055e-01_prt, -6.70705e-01_prt, -4.98281e-01_prt, -3.27829e-01_prt,
        -1.59381e-01_prt, 7.04180e-03_prt, 1.71434e-01_prt, 3.33803e-01_prt,
        4.94165e-01_prt, 6.52549e-01_prt, 8.08994e-01_prt, 9.63549e-01_prt,
        1.11627e+00_prt, 1.26722e+00_prt, 1.41648e+00_prt, 1.56411e+00_prt,
        1.71020e+00_prt, 1.85484e+00_prt, 1.99811e+00_prt, 2.14009e+00_prt,
        2.28088e+00_prt, 2.42056e+00_prt, 2.55922e+00_prt, 2.69694e+00_prt,
        2.83379e+00_prt, 2.96984e+00_prt, 3.10518e+00_prt, 3.23987e+00_prt,
        3.37396e+00_prt, 3.50752e+00_prt, 3.64060e+00_prt, 3.77324e+00_prt,
        3.90549e+00_prt, 4.03740e+00_prt, 4.16899e+00_prt, 4.30031e+00_prt};

    m_dndt_table = QS_dndt_table{dndt_params, vals};
}


void QuantumSynchrotronEngine::init_builtin_phot_em_table()
{
    constexpr auto default_chi_part_min = 1.0e-3_prt;
    constexpr auto default_chi_part_max = 200.0_prt;
    constexpr auto default_frac_min = 1.0e-12_prt;
    constexpr auto default_chi_part_how_many = 64;
    constexpr auto default_frac_how_many = 64;

    QS_phot_em_table_params phot_em_params;
    phot_em_params.chi_part_min = default_chi_part_min;
    phot_em_params.chi_part_max = default_chi_part_max;
    phot_em_params.frac_min = default_frac_min;
    phot_em_params.chi_part_how_many = default_chi_part_how_many;
    phot_em_params.frac_how_many = default_frac_how_many;

    const auto vals = std::vector<amrex::ParticleReal>{
        -6.83368e+00_prt, -6.68749e+00_prt, -6.54129e+00_prt, -6.39510e+00_prt,
        -6.24890e+00_prt, -6.10271e+00_prt, -5.95651e+00_prt, -5.81031e+00_prt,
        -5.66412e+00_prt, -5.51792e+00_prt, -5.37173e+00_prt, -5.22554e+00_prt,
        -5.07934e+00_prt, -4.93315e+00_prt, -4.78696e+00_prt, -4.64076e+00_prt,
        -4.49457e+00_prt, -4.34839e+00_prt, -4.20220e+00_prt, -4.05602e+00_prt,
        -3.90984e+00_prt, -3.76367e+00_prt, -3.61751e+00_prt, -3.47136e+00_prt,
        -3.32522e+00_prt, -3.17911e+00_prt, -3.03302e+00_prt, -2.88697e+00_prt,
        -2.74097e+00_prt, -2.59504e+00_prt, -2.44919e+00_prt, -2.30347e+00_prt,
        -2.15790e+00_prt, -2.01255e+00_prt, -1.86749e+00_prt, -1.72282e+00_prt,
        -1.57867e+00_prt, -1.43521e+00_prt, -1.29270e+00_prt, -1.15146e+00_prt,
        -1.01194e+00_prt, -8.74744e-01_prt, -7.40699e-01_prt, -6.10940e-01_prt,
        -4.87011e-01_prt, -3.70993e-01_prt, -2.65630e-01_prt, -1.74333e-01_prt,
        -1.00881e-01_prt, -4.84058e-02_prt, -1.74572e-02_prt, -4.04316e-03_prt,
        -4.68433e-04_prt, -1.82946e-05_prt, -1.28390e-07_prt, -7.14633e-11_prt,
        -1.32240e-11_prt, -1.32236e-11_prt, -1.32236e-11_prt, -1.32236e-11_prt,
        -1.32236e-11_prt, -1.32236e-11_prt, -1.32236e-11_prt, 0.00000e+00_prt,
        -6.89807e+00_prt, -6.75188e+00_prt, -6.60568e+00_prt, -6.45948e+00_prt,
        -6.31329e+00_prt, -6.16709e+00_prt, -6.02090e+00_prt, -5.87470e+00_prt,
        -5.72851e+00_prt, -5.58231e+00_prt, -5.43612e+00_prt, -5.28992e+00_prt,
        -5.14373e+00_prt, -4.99753e+00_prt, -4.85134e+00_prt, -4.70515e+00_prt,
        -4.55896e+00_prt, -4.41277e+00_prt, -4.26658e+00_prt, -4.12040e+00_prt,
        -3.97422e+00_prt, -3.82804e+00_prt, -3.68188e+00_prt, -3.53572e+00_prt,
        -3.38958e+00_prt, -3.24346e+00_prt, -3.09736e+00_prt, -2.95129e+00_prt,
        -2.80526e+00_prt, -2.65930e+00_prt, -2.51341e+00_prt, -2.36763e+00_prt,
        -2.22199e+00_prt, -2.07654e+00_prt, -1.93134e+00_prt, -1.78648e+00_prt,
        -1.64208e+00_prt, -1.49829e+00_prt, -1.35533e+00_prt, -1.21348e+00_prt,
        -1.07314e+00_prt, -9.34824e-01_prt, -7.99268e-01_prt, -6.67454e-01_prt,
        -5.40730e-01_prt, -4.20925e-01_prt, -3.10479e-01_prt, -2.12523e-01_prt,
        -1.30747e-01_prt, -6.87573e-02_prt, -2.85300e-02_prt, -8.21443e-03_prt,
        -1.33651e-03_prt, -8.87127e-05_prt, -1.42937e-06_prt, -2.40958e-09_prt,
        -2.41052e-12_prt, -2.30649e-12_prt, -2.30649e-12_prt, -2.30649e-12_prt,
        -2.30649e-12_prt, -2.30649e-12_prt, -2.30649e-12_prt, 0.00000e+00_prt,
        -6.96242e+00_prt, -6.81622e+00_prt, -6.67002e+00_prt, -6.52383e+00_prt,
        -6.37763e+00_prt, -6.23144e+00_prt, -6.08524e+00_prt, -5.93905e+00_prt,
        -5.79285e+00_prt, -5.64665e+00_prt, -5.50046e+00_prt, -5.35427e+00_prt,
        -5.20807e+00_prt, -5.06188e+00_prt, -4.91568e+00_prt, -4.76949e+00_prt,
        -4.62330e+00_prt, -4.47711e+00_prt, -4.33092e+00_prt, -4.18474e+00_prt,
        -4.03856e+00_prt, -3.89238e+00_prt, -3.74621e+00_prt, -3.60005e+00_prt,
        -3.45390e+00_prt, -3.30777e+00_prt, -3.16165e+00_prt, -3.01557e+00_prt,
        -2.86953e+00_prt, -2.72353e+00_prt, -2.57761e+00_prt, -2.43178e+00_prt,
        -2.28607e+00_prt, -2.14053e+00_prt, -1.99521e+00_prt, -1.85018e+00_prt,
        -1.70556e+00_prt, -1.56148e+00_prt, -1.41813e+00_prt, -1.27574e+00_prt,
        -1.13468e+00_prt, -9.95396e-01_prt, -8.58520e-01_prt, -7.24912e-01_prt,
        -5.95749e-01_prt, -4.72630e-01_prt, -3.57712e-01_prt, -2.53820e-01_prt,
        -1.64441e-01_prt, -9.33509e-02_prt, -4.35013e-02_prt, -1.49849e-02_prt,
        -3.22456e-03_prt, -3.31351e-04_prt, -1.05911e-05_prt, -5.25639e-08_prt,
        -1.48073e-11_prt, -2.08211e-12_prt, -2.08211e-12_prt, -2.08211e-12_prt,
        -2.08211e-12_prt, -2.08211e-12_prt, -2.08211e-12_prt, 0.00000e+00_prt,
        -7.02671e+00_prt, -6.88051e+00_prt, -6.73432e+00_prt, -6.58812e+00_prt,
        -6.44193e+00_prt, -6.29573e+00_prt, -6.14954e+00_prt, -6.00334e+00_prt,
        -5.85714e+00_prt, -5.71095e+00_prt, -5.56475e+00_prt, -5.41856e+00_prt,
        -5.27236e+00_prt, -5.12617e+00_prt, -4.97998e+00_prt, -4.83378e+00_prt,
        -4.68759e+00_prt, -4.54140e+00_prt, -4.39521e+00_prt, -4.24903e+00_prt,
        -4.10284e+00_prt, -3.95666e+00_prt, -3.81049e+00_prt, -3.66433e+00_prt,
        -3.51817e+00_prt, -3.37203e+00_prt, -3.22591e+00_prt, -3.07981e+00_prt,
        -2.93375e+00_prt, -2.78773e+00_prt, -2.64177e+00_prt, -2.49590e+00_prt,
        -2.35013e+00_prt, -2.20451e+00_prt, -2.05908e+00_prt, -1.91392e+00_prt,
        -1.76911e+00_prt, -1.62477e+00_prt, -1.48106e+00_prt, -1.33822e+00_prt,
        -1.19652e+00_prt, -1.05638e+00_prt, -9.18354e-01_prt, -7.83181e-01_prt,
        -6.51887e-01_prt, -5.25873e-01_prt, -4.07032e-01_prt, -2.97887e-01_prt,
        -2.01651e-01_prt, -1.22060e-01_prt, -6.26329e-02_prt, -2.50149e-02_prt,
        -6.77615e-03_prt, -9.98061e-04_prt, -5.61030e-05_prt, -6.78360e-07_prt,
        -6.88729e-10_prt, -9.53426e-12_prt, -9.52349e-12_prt, -9.52349e-12_prt,
        -9.52349e-12_prt, -9.52349e-12_prt, -9.52349e-12_prt, 0.00000e+00_prt,
        -7.09094e+00_prt, -6.94475e+00_prt, -6.79855e+00_prt, -6.65236e+00_prt,
        -6.50616e+00_prt, -6.35997e+00_prt, -6.21377e+00_prt, -6.06757e+00_prt,
        -5.92138e+00_prt, -5.77518e+00_prt, -5.62899e+00_prt, -5.48279e+00_prt,
        -5.33660e+00_prt, -5.19040e+00_prt, -5.04421e+00_prt, -4.89802e+00_prt,
        -4.75182e+00_prt, -4.60563e+00_prt, -4.45944e+00_prt, -4.31326e+00_prt,
        -4.16707e+00_prt, -4.02089e+00_prt, -3.87471e+00_prt, -3.72855e+00_prt,
        -3.58239e+00_prt, -3.43624e+00_prt, -3.29011e+00_prt, -3.14400e+00_prt,
        -2.99792e+00_prt, -2.85188e+00_prt, -2.70589e+00_prt, -2.55998e+00_prt,
        -2.41416e+00_prt, -2.26847e+00_prt, -2.12295e+00_prt, -1.97766e+00_prt,
        -1.83268e+00_prt, -1.68811e+00_prt, -1.54410e+00_prt, -1.40085e+00_prt,
        -1.25860e+00_prt, -1.11772e+00_prt, -9.78681e-01_prt, -8.42140e-01_prt,
        -7.08987e-01_prt, -5.80444e-01_prt, -4.58170e-01_prt, -3.44398e-01_prt,
        -2.42039e-01_prt, -1.54649e-01_prt, -8.59929e-02_prt, -3.88131e-02_prt,
        -1.27086e-02_prt, -2.51783e-03_prt, -2.25371e-04_prt, -5.67634e-06_prt,
        -1.83116e-08_prt, -7.44138e-12_prt, -5.50848e-12_prt, -5.50848e-12_prt,
        -5.50848e-12_prt, -5.50848e-12_prt, -5.50848e-12_prt, 0.00000e+00_prt,
        -7.15510e+00_prt, -7.00891e+00_prt, -6.86271e+00_prt, -6.71652e+00_prt,
        -6.57032e+00_prt, -6.42413e+00_prt, -6.27793e+00_prt, -6.13173e+00_prt,
        -5.98554e+00_prt, -5.83934e+00_prt, -5.69315e+00_prt, -5.54695e+00_prt,
        -5.40076e+00_prt, -5.25456e+00_prt, -5.10837e+00_prt, -4.96218e+00_prt,
        -4.81598e+00_prt, -4.66979e+00_prt, -4.52360e+00_prt, -4.37741e+00_prt,
        -4.23123e+00_prt, -4.08504e+00_prt, -3.93887e+00_prt, -3.79269e+00_prt,
        -3.64653e+00_prt, -3.50038e+00_prt, -3.35424e+00_prt, -3.20812e+00_prt,
        -3.06202e+00_prt, -2.91597e+00_prt, -2.76995e+00_prt, -2.62400e+00_prt,
        -2.47814e+00_prt, -2.33239e+00_prt, -2.18679e+00_prt, -2.04139e+00_prt,
        -1.89626e+00_prt, -1.75149e+00_prt, -1.60722e+00_prt, -1.46360e+00_prt,
        -1.32087e+00_prt, -1.17934e+00_prt, -1.03942e+00_prt, -9.01681e-01_prt,
        -7.66906e-01_prt, -6.36155e-01_prt, -5.10880e-01_prt, -3.93046e-01_prt,
        -2.85258e-01_prt, -1.90814e-01_prt, -1.13487e-01_prt, -5.66916e-02_prt,
        -2.17036e-02_prt, -5.48859e-03_prt, -7.20753e-04_prt, -3.32360e-05_prt,
        -2.81000e-07_prt, -1.48662e-10_prt, -7.22000e-12_prt, -7.21945e-12_prt,
        -7.21945e-12_prt, -7.21945e-12_prt, -7.21945e-12_prt, 0.00000e+00_prt,
        -7.21918e+00_prt, -7.07298e+00_prt, -6.92678e+00_prt, -6.78059e+00_prt,
        -6.63439e+00_prt, -6.48820e+00_prt, -6.34200e+00_prt, -6.19581e+00_prt,
        -6.04961e+00_prt, -5.90341e+00_prt, -5.75722e+00_prt, -5.61102e+00_prt,
        -5.46483e+00_prt, -5.31863e+00_prt, -5.17244e+00_prt, -5.02625e+00_prt,
        -4.88005e+00_prt, -4.73386e+00_prt, -4.58767e+00_prt, -4.44148e+00_prt,
        -4.29529e+00_prt, -4.14911e+00_prt, -4.00293e+00_prt, -3.85675e+00_prt,
        -3.71059e+00_prt, -3.56443e+00_prt, -3.41828e+00_prt, -3.27215e+00_prt,
        -3.12605e+00_prt, -2.97997e+00_prt, -2.83394e+00_prt, -2.68796e+00_prt,
        -2.54205e+00_prt, -2.39625e+00_prt, -2.25057e+00_prt, -2.10508e+00_prt,
        -1.95982e+00_prt, -1.81488e+00_prt, -1.67037e+00_prt, -1.52644e+00_prt,
        -1.38329e+00_prt, -1.24118e+00_prt, -1.10049e+00_prt, -9.61708e-01_prt,
        -8.25515e-01_prt, -6.92837e-01_prt, -5.64941e-01_prt, -4.43549e-01_prt,
        -3.30974e-01_prt, -2.30216e-01_prt, -1.44899e-01_prt, -7.87630e-02_prt,
        -3.43106e-02_prt, -1.06083e-02_prt, -1.91011e-03_prt, -1.44988e-04_prt,
        -2.71228e-06_prt, -4.95880e-09_prt, -1.54320e-11_prt, -1.52724e-11_prt,
        -1.52724e-11_prt, -1.52724e-11_prt, -1.52724e-11_prt, 0.00000e+00_prt,
        -7.28314e+00_prt, -7.13695e+00_prt, -6.99075e+00_prt, -6.84455e+00_prt,
        -6.69836e+00_prt, -6.55216e+00_prt, -6.40597e+00_prt, -6.25977e+00_prt,
        -6.11358e+00_prt, -5.96738e+00_prt, -5.82118e+00_prt, -5.67499e+00_prt,
        -5.52879e+00_prt, -5.38260e+00_prt, -5.23640e+00_prt, -5.09021e+00_prt,
        -4.94402e+00_prt, -4.79782e+00_prt, -4.65163e+00_prt, -4.50544e+00_prt,
        -4.35925e+00_prt, -4.21307e+00_prt, -4.06689e+00_prt, -3.92071e+00_prt,
        -3.77454e+00_prt, -3.62837e+00_prt, -3.48222e+00_prt, -3.33609e+00_prt,
        -3.18997e+00_prt, -3.04388e+00_prt, -2.89783e+00_prt, -2.75182e+00_prt,
        -2.60588e+00_prt, -2.46003e+00_prt, -2.31429e+00_prt, -2.16871e+00_prt,
        -2.02334e+00_prt, -1.87825e+00_prt, -1.73353e+00_prt, -1.58933e+00_prt,
        -1.44580e+00_prt, -1.30319e+00_prt, -1.16182e+00_prt, -1.02213e+00_prt,
        -8.84697e-01_prt, -7.50336e-01_prt, -6.20152e-01_prt, -4.95651e-01_prt,
        -3.78870e-01_prt, -2.72504e-01_prt, -1.79934e-01_prt, -1.04967e-01_prt,
        -5.08903e-02_prt, -1.85694e-02_prt, -4.33586e-03_prt, -4.96245e-04_prt,
        -1.78626e-05_prt, -9.39993e-08_prt, -2.60318e-11_prt, -8.32656e-12_prt,
        -8.32656e-12_prt, -8.32656e-12_prt, -8.32656e-12_prt, 0.00000e+00_prt,
        -7.34698e+00_prt, -7.20078e+00_prt, -7.05459e+00_prt, -6.90839e+00_prt,
        -6.76220e+00_prt, -6.61600e+00_prt, -6.46980e+00_prt, -6.32361e+00_prt,
        -6.17741e+00_prt, -6.03122e+00_prt, -5.88502e+00_prt, -5.73883e+00_prt,
        -5.59263e+00_prt, -5.44644e+00_prt, -5.30024e+00_prt, -5.15405e+00_prt,
        -5.00785e+00_prt, -4.86166e+00_prt, -4.71547e+00_prt, -4.56928e+00_prt,
        -4.42309e+00_prt, -4.27690e+00_prt, -4.13072e+00_prt, -3.98454e+00_prt,
        -3.83836e+00_prt, -3.69220e+00_prt, -3.54604e+00_prt, -3.39989e+00_prt,
        -3.25377e+00_prt, -3.10767e+00_prt, -2.96159e+00_prt, -2.81557e+00_prt,
        -2.66959e+00_prt, -2.52370e+00_prt, -2.37791e+00_prt, -2.23225e+00_prt,
        -2.08678e+00_prt, -1.94155e+00_prt, -1.79666e+00_prt, -1.65221e+00_prt,
        -1.50836e+00_prt, -1.36532e+00_prt, -1.22336e+00_prt, -1.08286e+00_prt,
        -9.44344e-01_prt, -8.08515e-01_prt, -6.76333e-01_prt, -5.49115e-01_prt,
        -4.28649e-01_prt, -3.17331e-01_prt, -2.18253e-01_prt, -1.35109e-01_prt,
        -7.16006e-02_prt, -2.99558e-02_prt, -8.66338e-03_prt, -1.39061e-03_prt,
        -8.58450e-05_prt, -1.08211e-06_prt, -8.81213e-10_prt, -9.14124e-12_prt,
        -9.13691e-12_prt, -9.13691e-12_prt, -9.13691e-12_prt, 0.00000e+00_prt,
        -7.41066e+00_prt, -7.26447e+00_prt, -7.11827e+00_prt, -6.97207e+00_prt,
        -6.82588e+00_prt, -6.67968e+00_prt, -6.53349e+00_prt, -6.38729e+00_prt,
        -6.24109e+00_prt, -6.09490e+00_prt, -5.94870e+00_prt, -5.80251e+00_prt,
        -5.65631e+00_prt, -5.51012e+00_prt, -5.36392e+00_prt, -5.21773e+00_prt,
        -5.07153e+00_prt, -4.92534e+00_prt, -4.77915e+00_prt, -4.63296e+00_prt,
        -4.48677e+00_prt, -4.34058e+00_prt, -4.19439e+00_prt, -4.04821e+00_prt,
        -3.90204e+00_prt, -3.75586e+00_prt, -3.60970e+00_prt, -3.46355e+00_prt,
        -3.31742e+00_prt, -3.17130e+00_prt, -3.02522e+00_prt, -2.87917e+00_prt,
        -2.73317e+00_prt, -2.58724e+00_prt, -2.44140e+00_prt, -2.29568e+00_prt,
        -2.15012e+00_prt, -2.00477e+00_prt, -1.85972e+00_prt, -1.71506e+00_prt,
        -1.57092e+00_prt, -1.42749e+00_prt, -1.28501e+00_prt, -1.14382e+00_prt,
        -1.00436e+00_prt, -8.67244e-01_prt, -7.33318e-01_prt, -6.03728e-01_prt,
        -4.80039e-01_prt, -3.64367e-01_prt, -2.59499e-01_prt, -1.68904e-01_prt,
        -9.64162e-02_prt, -4.51735e-02_prt, -1.55832e-02_prt, -3.30468e-03_prt,
        -3.18182e-04_prt, -8.23951e-06_prt, -2.20029e-08_prt, -9.32820e-12_prt,
        -8.44891e-12_prt, -8.44891e-12_prt, -8.44891e-12_prt, 0.00000e+00_prt,
        -7.47416e+00_prt, -7.32796e+00_prt, -7.18177e+00_prt, -7.03557e+00_prt,
        -6.88938e+00_prt, -6.74318e+00_prt, -6.59698e+00_prt, -6.45079e+00_prt,
        -6.30459e+00_prt, -6.15840e+00_prt, -6.01220e+00_prt, -5.86601e+00_prt,
        -5.71981e+00_prt, -5.57362e+00_prt, -5.42742e+00_prt, -5.28123e+00_prt,
        -5.13503e+00_prt, -4.98884e+00_prt, -4.84264e+00_prt, -4.69645e+00_prt,
        -4.55026e+00_prt, -4.40407e+00_prt, -4.25789e+00_prt, -4.11170e+00_prt,
        -3.96552e+00_prt, -3.81935e+00_prt, -3.67318e+00_prt, -3.52703e+00_prt,
        -3.38089e+00_prt, -3.23476e+00_prt, -3.08866e+00_prt, -2.94260e+00_prt,
        -2.79657e+00_prt, -2.65061e+00_prt, -2.50473e+00_prt, -2.35895e+00_prt,
        -2.21331e+00_prt, -2.06786e+00_prt, -1.92267e+00_prt, -1.77782e+00_prt,
        -1.63344e+00_prt, -1.48967e+00_prt, -1.34674e+00_prt, -1.20493e+00_prt,
        -1.06464e+00_prt, -9.26404e-01_prt, -7.90955e-01_prt, -6.59295e-01_prt,
        -5.32790e-01_prt, -4.13301e-01_prt, -3.03311e-01_prt, -2.06012e-01_prt,
        -1.25166e-01_prt, -6.44255e-02_prt, -2.57044e-02_prt, -6.85590e-03_prt,
        -9.51994e-04_prt, -4.46223e-05_prt, -3.19559e-07_prt, -7.43913e-11_prt,
        -3.27816e-12_prt, -3.27816e-12_prt, -3.27816e-12_prt, 0.00000e+00_prt,
        -7.53744e+00_prt, -7.39124e+00_prt, -7.24504e+00_prt, -7.09885e+00_prt,
        -6.95265e+00_prt, -6.80646e+00_prt, -6.66026e+00_prt, -6.51407e+00_prt,
        -6.36787e+00_prt, -6.22167e+00_prt, -6.07548e+00_prt, -5.92928e+00_prt,
        -5.78309e+00_prt, -5.63689e+00_prt, -5.49070e+00_prt, -5.34450e+00_prt,
        -5.19831e+00_prt, -5.05211e+00_prt, -4.90592e+00_prt, -4.75973e+00_prt,
        -4.61354e+00_prt, -4.46735e+00_prt, -4.32116e+00_prt, -4.17497e+00_prt,
        -4.02879e+00_prt, -3.88262e+00_prt, -3.73645e+00_prt, -3.59029e+00_prt,
        -3.44414e+00_prt, -3.29801e+00_prt, -3.15189e+00_prt, -3.00581e+00_prt,
        -2.85977e+00_prt, -2.71378e+00_prt, -2.56785e+00_prt, -2.42203e+00_prt,
        -2.27632e+00_prt, -2.13078e+00_prt, -1.98547e+00_prt, -1.84046e+00_prt,
        -1.69585e+00_prt, -1.55179e+00_prt, -1.40846e+00_prt, -1.26611e+00_prt,
        -1.12510e+00_prt, -9.85880e-01_prt, -8.49102e-01_prt, -7.15634e-01_prt,
        -5.86670e-01_prt, -4.63841e-01_prt, -3.49344e-01_prt, -2.46067e-01_prt,
        -1.57574e-01_prt, -8.77228e-02_prt, -3.94737e-02_prt, -1.27172e-02_prt,
        -2.38737e-03_prt, -1.82476e-04_prt, -2.94710e-06_prt, -2.70147e-09_prt,
        -3.03868e-12_prt, -3.03213e-12_prt, -3.03213e-12_prt, 0.00000e+00_prt,
        -7.60045e+00_prt, -7.45425e+00_prt, -7.30806e+00_prt, -7.16186e+00_prt,
        -7.01567e+00_prt, -6.86947e+00_prt, -6.72328e+00_prt, -6.57708e+00_prt,
        -6.43088e+00_prt, -6.28469e+00_prt, -6.13849e+00_prt, -5.99230e+00_prt,
        -5.84610e+00_prt, -5.69991e+00_prt, -5.55371e+00_prt, -5.40752e+00_prt,
        -5.26132e+00_prt, -5.11513e+00_prt, -4.96893e+00_prt, -4.82274e+00_prt,
        -4.67655e+00_prt, -4.53036e+00_prt, -4.38417e+00_prt, -4.23798e+00_prt,
        -4.09180e+00_prt, -3.94562e+00_prt, -3.79945e+00_prt, -3.65328e+00_prt,
        -3.50713e+00_prt, -3.36099e+00_prt, -3.21487e+00_prt, -3.06877e+00_prt,
        -2.92271e+00_prt, -2.77669e+00_prt, -2.63074e+00_prt, -2.48486e+00_prt,
        -2.33910e+00_prt, -2.19348e+00_prt, -2.04806e+00_prt, -1.90291e+00_prt,
        -1.75811e+00_prt, -1.61378e+00_prt, -1.47010e+00_prt, -1.32729e+00_prt,
        -1.18564e+00_prt, -1.04556e+00_prt, -9.07623e-01_prt, -7.72576e-01_prt,
        -6.41466e-01_prt, -5.15718e-01_prt, -3.97268e-01_prt, -2.88696e-01_prt,
        -1.93301e-01_prt, -1.14918e-01_prt, -5.71378e-02_prt, -2.15106e-02_prt,
        -5.17615e-03_prt, -5.91934e-04_prt, -1.86877e-05_prt, -5.48948e-08_prt,
        -9.48919e-12_prt, -8.28793e-12_prt, -8.28793e-12_prt, 0.00000e+00_prt,
        -7.66315e+00_prt, -7.51696e+00_prt, -7.37076e+00_prt, -7.22456e+00_prt,
        -7.07837e+00_prt, -6.93217e+00_prt, -6.78598e+00_prt, -6.63978e+00_prt,
        -6.49359e+00_prt, -6.34739e+00_prt, -6.20119e+00_prt, -6.05500e+00_prt,
        -5.90880e+00_prt, -5.76261e+00_prt, -5.61641e+00_prt, -5.47022e+00_prt,
        -5.32402e+00_prt, -5.17783e+00_prt, -5.03163e+00_prt, -4.88544e+00_prt,
        -4.73925e+00_prt, -4.59306e+00_prt, -4.44687e+00_prt, -4.30068e+00_prt,
        -4.15450e+00_prt, -4.00831e+00_prt, -3.86214e+00_prt, -3.71597e+00_prt,
        -3.56981e+00_prt, -3.42366e+00_prt, -3.27753e+00_prt, -3.13143e+00_prt,
        -2.98535e+00_prt, -2.83931e+00_prt, -2.69332e+00_prt, -2.54741e+00_prt,
        -2.40159e+00_prt, -2.25591e+00_prt, -2.11039e+00_prt, -1.96511e+00_prt,
        -1.82014e+00_prt, -1.67559e+00_prt, -1.53160e+00_prt, -1.38838e+00_prt,
        -1.24617e+00_prt, -1.10535e+00_prt, -9.66389e-01_prt, -8.29963e-01_prt,
        -6.96980e-01_prt, -5.68682e-01_prt, -4.46770e-01_prt, -3.33534e-01_prt,
        -2.31969e-01_prt, -1.45746e-01_prt, -7.87448e-02_prt, -3.37167e-02_prt,
        -9.95408e-03_prt, -1.58635e-03_prt, -8.70168e-05_prt, -6.67831e-07_prt,
        -9.66782e-11_prt, -7.57749e-12_prt, -7.57749e-12_prt, 0.00000e+00_prt,
        -7.72548e+00_prt, -7.57929e+00_prt, -7.43309e+00_prt, -7.28690e+00_prt,
        -7.14070e+00_prt, -6.99450e+00_prt, -6.84831e+00_prt, -6.70211e+00_prt,
        -6.55592e+00_prt, -6.40972e+00_prt, -6.26352e+00_prt, -6.11733e+00_prt,
        -5.97113e+00_prt, -5.82494e+00_prt, -5.67874e+00_prt, -5.53255e+00_prt,
        -5.38635e+00_prt, -5.24016e+00_prt, -5.09396e+00_prt, -4.94777e+00_prt,
        -4.80158e+00_prt, -4.65539e+00_prt, -4.50920e+00_prt, -4.36301e+00_prt,
        -4.21682e+00_prt, -4.07064e+00_prt, -3.92446e+00_prt, -3.77829e+00_prt,
        -3.63212e+00_prt, -3.48597e+00_prt, -3.33983e+00_prt, -3.19371e+00_prt,
        -3.04762e+00_prt, -2.90156e+00_prt, -2.75555e+00_prt, -2.60961e+00_prt,
        -2.46374e+00_prt, -2.31800e+00_prt, -2.17240e+00_prt, -2.02701e+00_prt,
        -1.88189e+00_prt, -1.73714e+00_prt, -1.59288e+00_prt, -1.44929e+00_prt,
        -1.30660e+00_prt, -1.16512e+00_prt, -1.02527e+00_prt, -8.87642e-01_prt,
        -7.53022e-01_prt, -6.22499e-01_prt, -4.97561e-01_prt, -3.80228e-01_prt,
        -2.73191e-01_prt, -1.79866e-01_prt, -1.04173e-01_prt, -4.96236e-02_prt,
        -1.73398e-02_prt, -3.63192e-03_prt, -3.14100e-04_prt, -5.32039e-06_prt,
        -3.17026e-09_prt, -2.22048e-11_prt, -2.22045e-11_prt, 0.00000e+00_prt,
        -7.78738e+00_prt, -7.64118e+00_prt, -7.49499e+00_prt, -7.34879e+00_prt,
        -7.20259e+00_prt, -7.05640e+00_prt, -6.91020e+00_prt, -6.76401e+00_prt,
        -6.61781e+00_prt, -6.47162e+00_prt, -6.32542e+00_prt, -6.17922e+00_prt,
        -6.03303e+00_prt, -5.88683e+00_prt, -5.74064e+00_prt, -5.59444e+00_prt,
        -5.44825e+00_prt, -5.30205e+00_prt, -5.15586e+00_prt, -5.00966e+00_prt,
        -4.86347e+00_prt, -4.71728e+00_prt, -4.57109e+00_prt, -4.42490e+00_prt,
        -4.27871e+00_prt, -4.13253e+00_prt, -3.98635e+00_prt, -3.84017e+00_prt,
        -3.69400e+00_prt, -3.54785e+00_prt, -3.40170e+00_prt, -3.25557e+00_prt,
        -3.10947e+00_prt, -2.96339e+00_prt, -2.81736e+00_prt, -2.67138e+00_prt,
        -2.52548e+00_prt, -2.37968e+00_prt, -2.23401e+00_prt, -2.08852e+00_prt,
        -1.94327e+00_prt, -1.79834e+00_prt, -1.65385e+00_prt, -1.50995e+00_prt,
        -1.36683e+00_prt, -1.22477e+00_prt, -1.08414e+00_prt, -9.45461e-01_prt,
        -8.09413e-01_prt, -6.76947e-01_prt, -5.49366e-01_prt, -4.28446e-01_prt,
        -3.16580e-01_prt, -2.16891e-01_prt, -1.33169e-01_prt, -6.93177e-02_prt,
        -2.78380e-02_prt, -7.30271e-03_prt, -9.19046e-04_prt, -2.98964e-05_prt,
        -6.03839e-08_prt, -1.53989e-11_prt, -1.52887e-11_prt, 0.00000e+00_prt,
        -7.84876e+00_prt, -7.70257e+00_prt, -7.55637e+00_prt, -7.41017e+00_prt,
        -7.26398e+00_prt, -7.11778e+00_prt, -6.97159e+00_prt, -6.82539e+00_prt,
        -6.67920e+00_prt, -6.53300e+00_prt, -6.38680e+00_prt, -6.24061e+00_prt,
        -6.09441e+00_prt, -5.94822e+00_prt, -5.80202e+00_prt, -5.65583e+00_prt,
        -5.50963e+00_prt, -5.36344e+00_prt, -5.21724e+00_prt, -5.07105e+00_prt,
        -4.92485e+00_prt, -4.77866e+00_prt, -4.63247e+00_prt, -4.48628e+00_prt,
        -4.34009e+00_prt, -4.19391e+00_prt, -4.04772e+00_prt, -3.90155e+00_prt,
        -3.75538e+00_prt, -3.60921e+00_prt, -3.46306e+00_prt, -3.31693e+00_prt,
        -3.17081e+00_prt, -3.02472e+00_prt, -2.87867e+00_prt, -2.73266e+00_prt,
        -2.58673e+00_prt, -2.44087e+00_prt, -2.29514e+00_prt, -2.14957e+00_prt,
        -2.00420e+00_prt, -1.85912e+00_prt, -1.71442e+00_prt, -1.57024e+00_prt,
        -1.42675e+00_prt, -1.28418e+00_prt, -1.14288e+00_prt, -1.00328e+00_prt,
        -8.65978e-01_prt, -7.31817e-01_prt, -6.01928e-01_prt, -4.77871e-01_prt,
        -3.61762e-01_prt, -2.56416e-01_prt, -1.65391e-01_prt, -9.27046e-02_prt,
        -4.17753e-02_prt, -1.31925e-02_prt, -2.26148e-03_prt, -1.25980e-04_prt,
        -6.99647e-07_prt, -1.03404e-11_prt, 0.00000e+00_prt, 0.00000e+00_prt,
        -7.90955e+00_prt, -7.76336e+00_prt, -7.61716e+00_prt, -7.47096e+00_prt,
        -7.32477e+00_prt, -7.17857e+00_prt, -7.03238e+00_prt, -6.88618e+00_prt,
        -6.73999e+00_prt, -6.59379e+00_prt, -6.44759e+00_prt, -6.30140e+00_prt,
        -6.15520e+00_prt, -6.00901e+00_prt, -5.86281e+00_prt, -5.71662e+00_prt,
        -5.57042e+00_prt, -5.42423e+00_prt, -5.27803e+00_prt, -5.13184e+00_prt,
        -4.98564e+00_prt, -4.83945e+00_prt, -4.69326e+00_prt, -4.54707e+00_prt,
        -4.40088e+00_prt, -4.25469e+00_prt, -4.10851e+00_prt, -3.96233e+00_prt,
        -3.81615e+00_prt, -3.66999e+00_prt, -3.52383e+00_prt, -3.37769e+00_prt,
        -3.23156e+00_prt, -3.08546e+00_prt, -2.93939e+00_prt, -2.79336e+00_prt,
        -2.64739e+00_prt, -2.50150e+00_prt, -2.35571e+00_prt, -2.21006e+00_prt,
        -2.06460e+00_prt, -1.91938e+00_prt, -1.77450e+00_prt, -1.63007e+00_prt,
        -1.48625e+00_prt, -1.34325e+00_prt, -1.20135e+00_prt, -1.06094e+00_prt,
        -9.22547e-01_prt, -7.86905e-01_prt, -6.55002e-01_prt, -5.28204e-01_prt,
        -4.08376e-01_prt, -2.98032e-01_prt, -2.00444e-01_prt, -1.19547e-01_prt,
        -5.92784e-02_prt, -2.18163e-02_prt, -4.82406e-03_prt, -4.18699e-04_prt,
        -5.35426e-06_prt, -5.59711e-10_prt, 0.00000e+00_prt, 0.00000e+00_prt,
        -7.96965e+00_prt, -7.82346e+00_prt, -7.67726e+00_prt, -7.53107e+00_prt,
        -7.38487e+00_prt, -7.23867e+00_prt, -7.09248e+00_prt, -6.94628e+00_prt,
        -6.80009e+00_prt, -6.65389e+00_prt, -6.50770e+00_prt, -6.36150e+00_prt,
        -6.21530e+00_prt, -6.06911e+00_prt, -5.92291e+00_prt, -5.77672e+00_prt,
        -5.63052e+00_prt, -5.48433e+00_prt, -5.33813e+00_prt, -5.19194e+00_prt,
        -5.04574e+00_prt, -4.89955e+00_prt, -4.75336e+00_prt, -4.60717e+00_prt,
        -4.46098e+00_prt, -4.31479e+00_prt, -4.16860e+00_prt, -4.02242e+00_prt,
        -3.87624e+00_prt, -3.73007e+00_prt, -3.58391e+00_prt, -3.43776e+00_prt,
        -3.29163e+00_prt, -3.14552e+00_prt, -2.99943e+00_prt, -2.85338e+00_prt,
        -2.70739e+00_prt, -2.56146e+00_prt, -2.41562e+00_prt, -2.26990e+00_prt,
        -2.12435e+00_prt, -1.97902e+00_prt, -1.83398e+00_prt, -1.68933e+00_prt,
        -1.54522e+00_prt, -1.40183e+00_prt, -1.25941e+00_prt, -1.11829e+00_prt,
        -9.78952e-01_prt, -8.42015e-01_prt, -7.08354e-01_prt, -5.79163e-01_prt,
        -4.56081e-01_prt, -3.41341e-01_prt, -2.37906e-01_prt, -1.49508e-01_prt,
        -8.02862e-02_prt, -3.35354e-02_prt, -9.14653e-03_prt, -1.14413e-03_prt,
        -2.91028e-05_prt, -1.37503e-08_prt, 0.00000e+00_prt, 0.00000e+00_prt,
        -8.02897e+00_prt, -7.88277e+00_prt, -7.73657e+00_prt, -7.59038e+00_prt,
        -7.44418e+00_prt, -7.29799e+00_prt, -7.15179e+00_prt, -7.00560e+00_prt,
        -6.85940e+00_prt, -6.71320e+00_prt, -6.56701e+00_prt, -6.42081e+00_prt,
        -6.27462e+00_prt, -6.12842e+00_prt, -5.98223e+00_prt, -5.83603e+00_prt,
        -5.68983e+00_prt, -5.54364e+00_prt, -5.39744e+00_prt, -5.25125e+00_prt,
        -5.10506e+00_prt, -4.95886e+00_prt, -4.81267e+00_prt, -4.66648e+00_prt,
        -4.52029e+00_prt, -4.37410e+00_prt, -4.22791e+00_prt, -4.08173e+00_prt,
        -3.93555e+00_prt, -3.78937e+00_prt, -3.64321e+00_prt, -3.49705e+00_prt,
        -3.35091e+00_prt, -3.20479e+00_prt, -3.05869e+00_prt, -2.91262e+00_prt,
        -2.76660e+00_prt, -2.62064e+00_prt, -2.47476e+00_prt, -2.32899e+00_prt,
        -2.18336e+00_prt, -2.03792e+00_prt, -1.89274e+00_prt, -1.74791e+00_prt,
        -1.60355e+00_prt, -1.45981e+00_prt, -1.31693e+00_prt, -1.17520e+00_prt,
        -1.03502e+00_prt, -8.96956e-01_prt, -7.61760e-01_prt, -6.30479e-01_prt,
        -5.04555e-01_prt, -3.85958e-01_prt, -2.77349e-01_prt, -1.82185e-01_prt,
        -1.04585e-01_prt, -4.85205e-02_prt, -1.57346e-02_prt, -2.66015e-03_prt,
        -1.19304e-04_prt, -1.95626e-07_prt, 0.00000e+00_prt, 0.00000e+00_prt,
        -8.08738e+00_prt, -7.94118e+00_prt, -7.79499e+00_prt, -7.64879e+00_prt,
        -7.50260e+00_prt, -7.35640e+00_prt, -7.21020e+00_prt, -7.06401e+00_prt,
        -6.91781e+00_prt, -6.77162e+00_prt, -6.62542e+00_prt, -6.47922e+00_prt,
        -6.33303e+00_prt, -6.18683e+00_prt, -6.04064e+00_prt, -5.89444e+00_prt,
        -5.74825e+00_prt, -5.60205e+00_prt, -5.45586e+00_prt, -5.30966e+00_prt,
        -5.16347e+00_prt, -5.01727e+00_prt, -4.87108e+00_prt, -4.72489e+00_prt,
        -4.57870e+00_prt, -4.43251e+00_prt, -4.28632e+00_prt, -4.14013e+00_prt,
        -3.99395e+00_prt, -3.84778e+00_prt, -3.70161e+00_prt, -3.55545e+00_prt,
        -3.40930e+00_prt, -3.26317e+00_prt, -3.11706e+00_prt, -2.97098e+00_prt,
        -2.82493e+00_prt, -2.67894e+00_prt, -2.53302e+00_prt, -2.38720e+00_prt,
        -2.24150e+00_prt, -2.09597e+00_prt, -1.95067e+00_prt, -1.80567e+00_prt,
        -1.66109e+00_prt, -1.51706e+00_prt, -1.37378e+00_prt, -1.23150e+00_prt,
        -1.09060e+00_prt, -9.51540e-01_prt, -8.15002e-01_prt, -6.81898e-01_prt,
        -5.53491e-01_prt, -4.31519e-01_prt, -3.18353e-01_prt, -2.17146e-01_prt,
        -1.31847e-01_prt, -6.67538e-02_prt, -2.49772e-02_prt, -5.41331e-03_prt,
        -3.87555e-04_prt, -1.77093e-06_prt, 0.00000e+00_prt, 0.00000e+00_prt,
        -8.14478e+00_prt, -7.99858e+00_prt, -7.85239e+00_prt, -7.70619e+00_prt,
        -7.55999e+00_prt, -7.41380e+00_prt, -7.26760e+00_prt, -7.12141e+00_prt,
        -6.97521e+00_prt, -6.82901e+00_prt, -6.68282e+00_prt, -6.53662e+00_prt,
        -6.39043e+00_prt, -6.24423e+00_prt, -6.09804e+00_prt, -5.95184e+00_prt,
        -5.80564e+00_prt, -5.65945e+00_prt, -5.51325e+00_prt, -5.36706e+00_prt,
        -5.22086e+00_prt, -5.07467e+00_prt, -4.92848e+00_prt, -4.78228e+00_prt,
        -4.63609e+00_prt, -4.48990e+00_prt, -4.34371e+00_prt, -4.19753e+00_prt,
        -4.05134e+00_prt, -3.90517e+00_prt, -3.75899e+00_prt, -3.61283e+00_prt,
        -3.46667e+00_prt, -3.32053e+00_prt, -3.17441e+00_prt, -3.02832e+00_prt,
        -2.88226e+00_prt, -2.73624e+00_prt, -2.59029e+00_prt, -2.44442e+00_prt,
        -2.29866e+00_prt, -2.15305e+00_prt, -2.00764e+00_prt, -1.86250e+00_prt,
        -1.71772e+00_prt, -1.57343e+00_prt, -1.42980e+00_prt, -1.28705e+00_prt,
        -1.14550e+00_prt, -1.00558e+00_prt, -8.67872e-01_prt, -7.33176e-01_prt,
        -6.02602e-01_prt, -4.77680e-01_prt, -3.60513e-01_prt, -2.53947e-01_prt,
        -1.61676e-01_prt, -8.80563e-02_prt, -3.70951e-02_prt, -9.86840e-03_prt,
        -1.03910e-03_prt, -1.10564e-05_prt, -9.50173e-12_prt, 0.00000e+00_prt,
        -8.20104e+00_prt, -8.05484e+00_prt, -7.90865e+00_prt, -7.76245e+00_prt,
        -7.61626e+00_prt, -7.47006e+00_prt, -7.32387e+00_prt, -7.17767e+00_prt,
        -7.03147e+00_prt, -6.88528e+00_prt, -6.73908e+00_prt, -6.59289e+00_prt,
        -6.44669e+00_prt, -6.30049e+00_prt, -6.15430e+00_prt, -6.00810e+00_prt,
        -5.86191e+00_prt, -5.71571e+00_prt, -5.56952e+00_prt, -5.42332e+00_prt,
        -5.27713e+00_prt, -5.13093e+00_prt, -4.98474e+00_prt, -4.83855e+00_prt,
        -4.69235e+00_prt, -4.54616e+00_prt, -4.39997e+00_prt, -4.25379e+00_prt,
        -4.10760e+00_prt, -3.96142e+00_prt, -3.81525e+00_prt, -3.66908e+00_prt,
        -3.52292e+00_prt, -3.37677e+00_prt, -3.23064e+00_prt, -3.08453e+00_prt,
        -2.93846e+00_prt, -2.79242e+00_prt, -2.64644e+00_prt, -2.50053e+00_prt,
        -2.35472e+00_prt, -2.20904e+00_prt, -2.06353e+00_prt, -1.91826e+00_prt,
        -1.77331e+00_prt, -1.62879e+00_prt, -1.48485e+00_prt, -1.34169e+00_prt,
        -1.19958e+00_prt, -1.05890e+00_prt, -9.20170e-01_prt, -7.84083e-01_prt,
        -6.51617e-01_prt, -5.24118e-01_prt, -4.03448e-01_prt, -2.92155e-01_prt,
        -1.93635e-01_prt, -1.12128e-01_prt, -5.21283e-02_prt, -1.64271e-02_prt,
        -2.37796e-03_prt, -5.08169e-05_prt, -6.19466e-10_prt, 0.00000e+00_prt,
        -8.25605e+00_prt, -8.10985e+00_prt, -7.96366e+00_prt, -7.81746e+00_prt,
        -7.67126e+00_prt, -7.52507e+00_prt, -7.37887e+00_prt, -7.23268e+00_prt,
        -7.08648e+00_prt, -6.94029e+00_prt, -6.79409e+00_prt, -6.64789e+00_prt,
        -6.50170e+00_prt, -6.35550e+00_prt, -6.20931e+00_prt, -6.06311e+00_prt,
        -5.91692e+00_prt, -5.77072e+00_prt, -5.62452e+00_prt, -5.47833e+00_prt,
        -5.33213e+00_prt, -5.18594e+00_prt, -5.03975e+00_prt, -4.89355e+00_prt,
        -4.74736e+00_prt, -4.60117e+00_prt, -4.45498e+00_prt, -4.30879e+00_prt,
        -4.16260e+00_prt, -4.01642e+00_prt, -3.87024e+00_prt, -3.72407e+00_prt,
        -3.57791e+00_prt, -3.43175e+00_prt, -3.28562e+00_prt, -3.13950e+00_prt,
        -2.99341e+00_prt, -2.84735e+00_prt, -2.70134e+00_prt, -2.55540e+00_prt,
        -2.40954e+00_prt, -2.26380e+00_prt, -2.11821e+00_prt, -1.97283e+00_prt,
        -1.82773e+00_prt, -1.68301e+00_prt, -1.53880e+00_prt, -1.39527e+00_prt,
        -1.25267e+00_prt, -1.11133e+00_prt, -9.71702e-01_prt, -8.34400e-01_prt,
        -7.00282e-01_prt, -5.70534e-01_prt, -4.46800e-01_prt, -3.31354e-01_prt,
        -2.27280e-01_prt, -1.38590e-01_prt, -6.99533e-02_prt, -2.53641e-02_prt,
        -4.77509e-03_prt, -1.81428e-04_prt, -1.40005e-08_prt, 0.00000e+00_prt,
        -8.30968e+00_prt, -8.16348e+00_prt, -8.01729e+00_prt, -7.87109e+00_prt,
        -7.72490e+00_prt, -7.57870e+00_prt, -7.43250e+00_prt, -7.28631e+00_prt,
        -7.14011e+00_prt, -6.99392e+00_prt, -6.84772e+00_prt, -6.70152e+00_prt,
        -6.55533e+00_prt, -6.40913e+00_prt, -6.26294e+00_prt, -6.11674e+00_prt,
        -5.97055e+00_prt, -5.82435e+00_prt, -5.67816e+00_prt, -5.53196e+00_prt,
        -5.38576e+00_prt, -5.23957e+00_prt, -5.09338e+00_prt, -4.94718e+00_prt,
        -4.80099e+00_prt, -4.65480e+00_prt, -4.50861e+00_prt, -4.36242e+00_prt,
        -4.21623e+00_prt, -4.07005e+00_prt, -3.92387e+00_prt, -3.77769e+00_prt,
        -3.63152e+00_prt, -3.48537e+00_prt, -3.33922e+00_prt, -3.19309e+00_prt,
        -3.04699e+00_prt, -2.90092e+00_prt, -2.75489e+00_prt, -2.60891e+00_prt,
        -2.46301e+00_prt, -2.31722e+00_prt, -2.17155e+00_prt, -2.02608e+00_prt,
        -1.88085e+00_prt, -1.73595e+00_prt, -1.59150e+00_prt, -1.44765e+00_prt,
        -1.30462e+00_prt, -1.16270e+00_prt, -1.02229e+00_prt, -8.83923e-01_prt,
        -7.48362e-01_prt, -6.16650e-01_prt, -4.90241e-01_prt, -3.71154e-01_prt,
        -2.62173e-01_prt, -1.67019e-01_prt, -9.03181e-02_prt, -3.67951e-02_prt,
        -8.60747e-03_prt, -5.25902e-04_prt, -1.83903e-07_prt, 0.00000e+00_prt,
        -8.36182e+00_prt, -8.21562e+00_prt, -8.06943e+00_prt, -7.92323e+00_prt,
        -7.77703e+00_prt, -7.63084e+00_prt, -7.48464e+00_prt, -7.33845e+00_prt,
        -7.19225e+00_prt, -7.04606e+00_prt, -6.89986e+00_prt, -6.75366e+00_prt,
        -6.60747e+00_prt, -6.46127e+00_prt, -6.31508e+00_prt, -6.16888e+00_prt,
        -6.02269e+00_prt, -5.87649e+00_prt, -5.73029e+00_prt, -5.58410e+00_prt,
        -5.43790e+00_prt, -5.29171e+00_prt, -5.14551e+00_prt, -4.99932e+00_prt,
        -4.85313e+00_prt, -4.70693e+00_prt, -4.56074e+00_prt, -4.41455e+00_prt,
        -4.26836e+00_prt, -4.12218e+00_prt, -3.97600e+00_prt, -3.82982e+00_prt,
        -3.68365e+00_prt, -3.53749e+00_prt, -3.39134e+00_prt, -3.24520e+00_prt,
        -3.09908e+00_prt, -2.95300e+00_prt, -2.80695e+00_prt, -2.66095e+00_prt,
        -2.51501e+00_prt, -2.36917e+00_prt, -2.22344e+00_prt, -2.07788e+00_prt,
        -1.93253e+00_prt, -1.78748e+00_prt, -1.64282e+00_prt, -1.49869e+00_prt,
        -1.35529e+00_prt, -1.21286e+00_prt, -1.07176e+00_prt, -9.32463e-01_prt,
        -7.95642e-01_prt, -6.62215e-01_prt, -5.33469e-01_prt, -4.11198e-01_prt,
        -2.97898e-01_prt, -1.96981e-01_prt, -1.12883e-01_prt, -5.06787e-02_prt,
        -1.41908e-02_prt, -1.28359e-03_prt, -1.56011e-06_prt, 0.00000e+00_prt,
        -8.41236e+00_prt, -8.26616e+00_prt, -8.11996e+00_prt, -7.97377e+00_prt,
        -7.82757e+00_prt, -7.68138e+00_prt, -7.53518e+00_prt, -7.38899e+00_prt,
        -7.24279e+00_prt, -7.09659e+00_prt, -6.95040e+00_prt, -6.80420e+00_prt,
        -6.65801e+00_prt, -6.51181e+00_prt, -6.36561e+00_prt, -6.21942e+00_prt,
        -6.07322e+00_prt, -5.92703e+00_prt, -5.78083e+00_prt, -5.63464e+00_prt,
        -5.48844e+00_prt, -5.34225e+00_prt, -5.19605e+00_prt, -5.04986e+00_prt,
        -4.90366e+00_prt, -4.75747e+00_prt, -4.61128e+00_prt, -4.46509e+00_prt,
        -4.31890e+00_prt, -4.17271e+00_prt, -4.02653e+00_prt, -3.88035e+00_prt,
        -3.73417e+00_prt, -3.58801e+00_prt, -3.44185e+00_prt, -3.29571e+00_prt,
        -3.14958e+00_prt, -3.00348e+00_prt, -2.85742e+00_prt, -2.71139e+00_prt,
        -2.56543e+00_prt, -2.41954e+00_prt, -2.27376e+00_prt, -2.12812e+00_prt,
        -1.98267e+00_prt, -1.83748e+00_prt, -1.69264e+00_prt, -1.54827e+00_prt,
        -1.40453e+00_prt, -1.26166e+00_prt, -1.11995e+00_prt, -9.79846e-01_prt,
        -8.41924e-01_prt, -7.07000e-01_prt, -5.76214e-01_prt, -4.51160e-01_prt,
        -3.34072e-01_prt, -2.28051e-01_prt, -1.37260e-01_prt, -6.68403e-02_prt,
        -2.17291e-02_prt, -2.71824e-03_prt, -9.23596e-06_prt, 0.00000e+00_prt,
        -8.46119e+00_prt, -8.31500e+00_prt, -8.16880e+00_prt, -8.02261e+00_prt,
        -7.87641e+00_prt, -7.73021e+00_prt, -7.58402e+00_prt, -7.43782e+00_prt,
        -7.29163e+00_prt, -7.14543e+00_prt, -6.99923e+00_prt, -6.85304e+00_prt,
        -6.70684e+00_prt, -6.56065e+00_prt, -6.41445e+00_prt, -6.26826e+00_prt,
        -6.12206e+00_prt, -5.97586e+00_prt, -5.82967e+00_prt, -5.68347e+00_prt,
        -5.53728e+00_prt, -5.39108e+00_prt, -5.24489e+00_prt, -5.09869e+00_prt,
        -4.95250e+00_prt, -4.80631e+00_prt, -4.66011e+00_prt, -4.51392e+00_prt,
        -4.36773e+00_prt, -4.22154e+00_prt, -4.07536e+00_prt, -3.92918e+00_prt,
        -3.78300e+00_prt, -3.63683e+00_prt, -3.49067e+00_prt, -3.34452e+00_prt,
        -3.19839e+00_prt, -3.05228e+00_prt, -2.90619e+00_prt, -2.76015e+00_prt,
        -2.61415e+00_prt, -2.46823e+00_prt, -2.32240e+00_prt, -2.17669e+00_prt,
        -2.03116e+00_prt, -1.88585e+00_prt, -1.74084e+00_prt, -1.59626e+00_prt,
        -1.45223e+00_prt, -1.30897e+00_prt, -1.16674e+00_prt, -1.02592e+00_prt,
        -8.87032e-01_prt, -7.50800e-01_prt, -6.18233e-01_prt, -4.90749e-01_prt,
        -3.70346e-01_prt, -2.59826e-01_prt, -1.63042e-01_prt, -8.50082e-02_prt,
        -3.12914e-02_prt, -5.11928e-03_prt, -4.06448e-05_prt, 0.00000e+00_prt,
        -8.50824e+00_prt, -8.36205e+00_prt, -8.21585e+00_prt, -8.06965e+00_prt,
        -7.92346e+00_prt, -7.77726e+00_prt, -7.63107e+00_prt, -7.48487e+00_prt,
        -7.33868e+00_prt, -7.19248e+00_prt, -7.04628e+00_prt, -6.90009e+00_prt,
        -6.75389e+00_prt, -6.60770e+00_prt, -6.46150e+00_prt, -6.31530e+00_prt,
        -6.16911e+00_prt, -6.02291e+00_prt, -5.87672e+00_prt, -5.73052e+00_prt,
        -5.58433e+00_prt, -5.43813e+00_prt, -5.29194e+00_prt, -5.14574e+00_prt,
        -4.99955e+00_prt, -4.85335e+00_prt, -4.70716e+00_prt, -4.56097e+00_prt,
        -4.41478e+00_prt, -4.26859e+00_prt, -4.12240e+00_prt, -3.97622e+00_prt,
        -3.83004e+00_prt, -3.68387e+00_prt, -3.53770e+00_prt, -3.39155e+00_prt,
        -3.24541e+00_prt, -3.09928e+00_prt, -2.95319e+00_prt, -2.80712e+00_prt,
        -2.66111e+00_prt, -2.51515e+00_prt, -2.36928e+00_prt, -2.22351e+00_prt,
        -2.07790e+00_prt, -1.93248e+00_prt, -1.78734e+00_prt, -1.64256e+00_prt,
        -1.49828e+00_prt, -1.35467e+00_prt, -1.21198e+00_prt, -1.07054e+00_prt,
        -9.30813e-01_prt, -7.93438e-01_prt, -6.59315e-01_prt, -5.29710e-01_prt,
        -4.06408e-01_prt, -2.91936e-01_prt, -1.89828e-01_prt, -1.04850e-01_prt,
        -4.28169e-02_prt, -8.74979e-03_prt, -1.40043e-04_prt, 0.00000e+00_prt,
        -8.55343e+00_prt, -8.40723e+00_prt, -8.26104e+00_prt, -8.11484e+00_prt,
        -7.96864e+00_prt, -7.82245e+00_prt, -7.67625e+00_prt, -7.53006e+00_prt,
        -7.38386e+00_prt, -7.23766e+00_prt, -7.09147e+00_prt, -6.94527e+00_prt,
        -6.79908e+00_prt, -6.65288e+00_prt, -6.50668e+00_prt, -6.36049e+00_prt,
        -6.21429e+00_prt, -6.06810e+00_prt, -5.92190e+00_prt, -5.77571e+00_prt,
        -5.62951e+00_prt, -5.48332e+00_prt, -5.33712e+00_prt, -5.19093e+00_prt,
        -5.04473e+00_prt, -4.89854e+00_prt, -4.75234e+00_prt, -4.60615e+00_prt,
        -4.45996e+00_prt, -4.31377e+00_prt, -4.16758e+00_prt, -4.02140e+00_prt,
        -3.87522e+00_prt, -3.72904e+00_prt, -3.58287e+00_prt, -3.43671e+00_prt,
        -3.29057e+00_prt, -3.14443e+00_prt, -2.99833e+00_prt, -2.85225e+00_prt,
        -2.70621e+00_prt, -2.56022e+00_prt, -2.41431e+00_prt, -2.26849e+00_prt,
        -2.12281e+00_prt, -1.97730e+00_prt, -1.83204e+00_prt, -1.68709e+00_prt,
        -1.54259e+00_prt, -1.39868e+00_prt, -1.25559e+00_prt, -1.11360e+00_prt,
        -9.73133e-01_prt, -8.34759e-01_prt, -6.99278e-01_prt, -5.67819e-01_prt,
        -4.41985e-01_prt, -3.24052e-01_prt, -2.17240e-01_prt, -1.26006e-01_prt,
        -5.61371e-02_prt, -1.38001e-02_prt, -3.94274e-04_prt, 0.00000e+00_prt,
        -8.59669e+00_prt, -8.45049e+00_prt, -8.30429e+00_prt, -8.15810e+00_prt,
        -8.01190e+00_prt, -7.86571e+00_prt, -7.71951e+00_prt, -7.57332e+00_prt,
        -7.42712e+00_prt, -7.28092e+00_prt, -7.13473e+00_prt, -6.98853e+00_prt,
        -6.84234e+00_prt, -6.69614e+00_prt, -6.54994e+00_prt, -6.40375e+00_prt,
        -6.25755e+00_prt, -6.11136e+00_prt, -5.96516e+00_prt, -5.81897e+00_prt,
        -5.67277e+00_prt, -5.52657e+00_prt, -5.38038e+00_prt, -5.23418e+00_prt,
        -5.08799e+00_prt, -4.94180e+00_prt, -4.79560e+00_prt, -4.64941e+00_prt,
        -4.50322e+00_prt, -4.35703e+00_prt, -4.21084e+00_prt, -4.06465e+00_prt,
        -3.91847e+00_prt, -3.77229e+00_prt, -3.62612e+00_prt, -3.47996e+00_prt,
        -3.33380e+00_prt, -3.18766e+00_prt, -3.04154e+00_prt, -2.89545e+00_prt,
        -2.74939e+00_prt, -2.60338e+00_prt, -2.45744e+00_prt, -2.31158e+00_prt,
        -2.16583e+00_prt, -2.02024e+00_prt, -1.87487e+00_prt, -1.72978e+00_prt,
        -1.58508e+00_prt, -1.44091e+00_prt, -1.29746e+00_prt, -1.15499e+00_prt,
        -1.01388e+00_prt, -8.74633e-01_prt, -7.37964e-01_prt, -6.04888e-01_prt,
        -4.76843e-01_prt, -3.55882e-01_prt, -2.44931e-01_prt, -1.48114e-01_prt,
        -7.10067e-02_prt, -2.03603e-02_prt, -9.39472e-04_prt, 0.00000e+00_prt,
        -8.63797e+00_prt, -8.49178e+00_prt, -8.34558e+00_prt, -8.19939e+00_prt,
        -8.05319e+00_prt, -7.90699e+00_prt, -7.76080e+00_prt, -7.61460e+00_prt,
        -7.46841e+00_prt, -7.32221e+00_prt, -7.17602e+00_prt, -7.02982e+00_prt,
        -6.88362e+00_prt, -6.73743e+00_prt, -6.59123e+00_prt, -6.44504e+00_prt,
        -6.29884e+00_prt, -6.15264e+00_prt, -6.00645e+00_prt, -5.86025e+00_prt,
        -5.71406e+00_prt, -5.56786e+00_prt, -5.42167e+00_prt, -5.27547e+00_prt,
        -5.12928e+00_prt, -4.98308e+00_prt, -4.83689e+00_prt, -4.69070e+00_prt,
        -4.54450e+00_prt, -4.39831e+00_prt, -4.25212e+00_prt, -4.10594e+00_prt,
        -3.95975e+00_prt, -3.81357e+00_prt, -3.66740e+00_prt, -3.52123e+00_prt,
        -3.37507e+00_prt, -3.22892e+00_prt, -3.08280e+00_prt, -2.93669e+00_prt,
        -2.79062e+00_prt, -2.64458e+00_prt, -2.49861e+00_prt, -2.35271e+00_prt,
        -2.20691e+00_prt, -2.06125e+00_prt, -1.91578e+00_prt, -1.77056e+00_prt,
        -1.62569e+00_prt, -1.48129e+00_prt, -1.33753e+00_prt, -1.19464e+00_prt,
        -1.05297e+00_prt, -9.12953e-01_prt, -7.75246e-01_prt, -6.40757e-01_prt,
        -5.10782e-01_prt, -3.87174e-01_prt, -2.72594e-01_prt, -1.70827e-01_prt,
        -8.71357e-02_prt, -2.84153e-02_prt, -1.95018e-03_prt, 0.00000e+00_prt,
        -8.67726e+00_prt, -8.53106e+00_prt, -8.38486e+00_prt, -8.23867e+00_prt,
        -8.09247e+00_prt, -7.94628e+00_prt, -7.80008e+00_prt, -7.65389e+00_prt,
        -7.50769e+00_prt, -7.36149e+00_prt, -7.21530e+00_prt, -7.06910e+00_prt,
        -6.92291e+00_prt, -6.77671e+00_prt, -6.63051e+00_prt, -6.48432e+00_prt,
        -6.33812e+00_prt, -6.19193e+00_prt, -6.04573e+00_prt, -5.89954e+00_prt,
        -5.75334e+00_prt, -5.60714e+00_prt, -5.46095e+00_prt, -5.31475e+00_prt,
        -5.16856e+00_prt, -5.02236e+00_prt, -4.87617e+00_prt, -4.72998e+00_prt,
        -4.58378e+00_prt, -4.43759e+00_prt, -4.29140e+00_prt, -4.14521e+00_prt,
        -3.99903e+00_prt, -3.85285e+00_prt, -3.70667e+00_prt, -3.56050e+00_prt,
        -3.41433e+00_prt, -3.26818e+00_prt, -3.12205e+00_prt, -2.97593e+00_prt,
        -2.82984e+00_prt, -2.68379e+00_prt, -2.53779e+00_prt, -2.39185e+00_prt,
        -2.24601e+00_prt, -2.10029e+00_prt, -1.95473e+00_prt, -1.80940e+00_prt,
        -1.66438e+00_prt, -1.51978e+00_prt, -1.37574e+00_prt, -1.23249e+00_prt,
        -1.09032e+00_prt, -9.49633e-01_prt, -8.11019e-01_prt, -6.75295e-01_prt,
        -5.43634e-01_prt, -4.17716e-01_prt, -2.99957e-01_prt, -1.93826e-01_prt,
        -1.04216e-01_prt, -3.78569e-02_prt, -3.61159e-03_prt, 0.00000e+00_prt,
        -8.71451e+00_prt, -8.56832e+00_prt, -8.42212e+00_prt, -8.27593e+00_prt,
        -8.12973e+00_prt, -7.98354e+00_prt, -7.83734e+00_prt, -7.69114e+00_prt,
        -7.54495e+00_prt, -7.39875e+00_prt, -7.25256e+00_prt, -7.10636e+00_prt,
        -6.96016e+00_prt, -6.81397e+00_prt, -6.66777e+00_prt, -6.52158e+00_prt,
        -6.37538e+00_prt, -6.22918e+00_prt, -6.08299e+00_prt, -5.93679e+00_prt,
        -5.79060e+00_prt, -5.64440e+00_prt, -5.49821e+00_prt, -5.35201e+00_prt,
        -5.20582e+00_prt, -5.05962e+00_prt, -4.91343e+00_prt, -4.76723e+00_prt,
        -4.62104e+00_prt, -4.47485e+00_prt, -4.32866e+00_prt, -4.18247e+00_prt,
        -4.03628e+00_prt, -3.89010e+00_prt, -3.74392e+00_prt, -3.59774e+00_prt,
        -3.45158e+00_prt, -3.30542e+00_prt, -3.15928e+00_prt, -3.01315e+00_prt,
        -2.86705e+00_prt, -2.72098e+00_prt, -2.57496e+00_prt, -2.42899e+00_prt,
        -2.28311e+00_prt, -2.13733e+00_prt, -1.99170e+00_prt, -1.84627e+00_prt,
        -1.70112e+00_prt, -1.55633e+00_prt, -1.41206e+00_prt, -1.26849e+00_prt,
        -1.12588e+00_prt, -9.84610e-01_prt, -8.45201e-01_prt, -7.08398e-01_prt,
        -5.75265e-01_prt, -4.47328e-01_prt, -3.26792e-01_prt, -2.16825e-01_prt,
        -1.21941e-01_prt, -4.85071e-02_prt, -6.08473e-03_prt, 0.00000e+00_prt,
        -8.74974e+00_prt, -8.60355e+00_prt, -8.45735e+00_prt, -8.31115e+00_prt,
        -8.16496e+00_prt, -8.01876e+00_prt, -7.87257e+00_prt, -7.72637e+00_prt,
        -7.58018e+00_prt, -7.43398e+00_prt, -7.28778e+00_prt, -7.14159e+00_prt,
        -6.99539e+00_prt, -6.84920e+00_prt, -6.70300e+00_prt, -6.55680e+00_prt,
        -6.41061e+00_prt, -6.26441e+00_prt, -6.11822e+00_prt, -5.97202e+00_prt,
        -5.82583e+00_prt, -5.67963e+00_prt, -5.53343e+00_prt, -5.38724e+00_prt,
        -5.24104e+00_prt, -5.09485e+00_prt, -4.94865e+00_prt, -4.80246e+00_prt,
        -4.65627e+00_prt, -4.51007e+00_prt, -4.36388e+00_prt, -4.21769e+00_prt,
        -4.07151e+00_prt, -3.92532e+00_prt, -3.77914e+00_prt, -3.63296e+00_prt,
        -3.48679e+00_prt, -3.34063e+00_prt, -3.19448e+00_prt, -3.04835e+00_prt,
        -2.90223e+00_prt, -2.75615e+00_prt, -2.61010e+00_prt, -2.46411e+00_prt,
        -2.31819e+00_prt, -2.17237e+00_prt, -2.02667e+00_prt, -1.88116e+00_prt,
        -1.73589e+00_prt, -1.59095e+00_prt, -1.44646e+00_prt, -1.30261e+00_prt,
        -1.15962e+00_prt, -1.01784e+00_prt, -8.77732e-01_prt, -7.39986e-01_prt,
        -6.05565e-01_prt, -4.75867e-01_prt, -3.52903e-01_prt, -2.39575e-01_prt,
        -1.40023e-01_prt, -6.01435e-02_prt, -9.47701e-03_prt, 0.00000e+00_prt,
        -8.78295e+00_prt, -8.63675e+00_prt, -8.49055e+00_prt, -8.34436e+00_prt,
        -8.19816e+00_prt, -8.05197e+00_prt, -7.90577e+00_prt, -7.75957e+00_prt,
        -7.61338e+00_prt, -7.46718e+00_prt, -7.32099e+00_prt, -7.17479e+00_prt,
        -7.02860e+00_prt, -6.88240e+00_prt, -6.73620e+00_prt, -6.59001e+00_prt,
        -6.44381e+00_prt, -6.29762e+00_prt, -6.15142e+00_prt, -6.00522e+00_prt,
        -5.85903e+00_prt, -5.71283e+00_prt, -5.56664e+00_prt, -5.42044e+00_prt,
        -5.27425e+00_prt, -5.12805e+00_prt, -4.98186e+00_prt, -4.83566e+00_prt,
        -4.68947e+00_prt, -4.54328e+00_prt, -4.39708e+00_prt, -4.25089e+00_prt,
        -4.10470e+00_prt, -3.95852e+00_prt, -3.81233e+00_prt, -3.66616e+00_prt,
        -3.51998e+00_prt, -3.37382e+00_prt, -3.22766e+00_prt, -3.08152e+00_prt,
        -2.93540e+00_prt, -2.78930e+00_prt, -2.64324e+00_prt, -2.49722e+00_prt,
        -2.35127e+00_prt, -2.20540e+00_prt, -2.05965e+00_prt, -1.91406e+00_prt,
        -1.76869e+00_prt, -1.62361e+00_prt, -1.47894e+00_prt, -1.33484e+00_prt,
        -1.19152e+00_prt, -1.04928e+00_prt, -9.08571e-01_prt, -7.69999e-01_prt,
        -6.34454e-01_prt, -5.03216e-01_prt, -3.78133e-01_prt, -2.61864e-01_prt,
        -1.58201e-01_prt, -7.25228e-02_prt, -1.38263e-02_prt, 0.00000e+00_prt,
        -8.81414e+00_prt, -8.66795e+00_prt, -8.52175e+00_prt, -8.37556e+00_prt,
        -8.22936e+00_prt, -8.08316e+00_prt, -7.93697e+00_prt, -7.79077e+00_prt,
        -7.64458e+00_prt, -7.49838e+00_prt, -7.35218e+00_prt, -7.20599e+00_prt,
        -7.05979e+00_prt, -6.91360e+00_prt, -6.76740e+00_prt, -6.62121e+00_prt,
        -6.47501e+00_prt, -6.32881e+00_prt, -6.18262e+00_prt, -6.03642e+00_prt,
        -5.89023e+00_prt, -5.74403e+00_prt, -5.59783e+00_prt, -5.45164e+00_prt,
        -5.30544e+00_prt, -5.15925e+00_prt, -5.01305e+00_prt, -4.86686e+00_prt,
        -4.72067e+00_prt, -4.57447e+00_prt, -4.42828e+00_prt, -4.28209e+00_prt,
        -4.13590e+00_prt, -3.98971e+00_prt, -3.84353e+00_prt, -3.69735e+00_prt,
        -3.55117e+00_prt, -3.40500e+00_prt, -3.25884e+00_prt, -3.11269e+00_prt,
        -2.96656e+00_prt, -2.82045e+00_prt, -2.67438e+00_prt, -2.52834e+00_prt,
        -2.38236e+00_prt, -2.23645e+00_prt, -2.09065e+00_prt, -1.94500e+00_prt,
        -1.79953e+00_prt, -1.65433e+00_prt, -1.50951e+00_prt, -1.36518e+00_prt,
        -1.22157e+00_prt, -1.07894e+00_prt, -9.37694e-01_prt, -7.98399e-01_prt,
        -6.61871e-01_prt, -5.29289e-01_prt, -4.02354e-01_prt, -2.83515e-01_prt,
        -1.76243e-01_prt, -8.54001e-02_prt, -1.90999e-02_prt, 0.00000e+00_prt,
        -8.84336e+00_prt, -8.69717e+00_prt, -8.55097e+00_prt, -8.40477e+00_prt,
        -8.25858e+00_prt, -8.11238e+00_prt, -7.96619e+00_prt, -7.81999e+00_prt,
        -7.67380e+00_prt, -7.52760e+00_prt, -7.38140e+00_prt, -7.23521e+00_prt,
        -7.08901e+00_prt, -6.94282e+00_prt, -6.79662e+00_prt, -6.65042e+00_prt,
        -6.50423e+00_prt, -6.35803e+00_prt, -6.21184e+00_prt, -6.06564e+00_prt,
        -5.91945e+00_prt, -5.77325e+00_prt, -5.62705e+00_prt, -5.48086e+00_prt,
        -5.33466e+00_prt, -5.18847e+00_prt, -5.04227e+00_prt, -4.89608e+00_prt,
        -4.74988e+00_prt, -4.60369e+00_prt, -4.45750e+00_prt, -4.31131e+00_prt,
        -4.16512e+00_prt, -4.01893e+00_prt, -3.87274e+00_prt, -3.72656e+00_prt,
        -3.58038e+00_prt, -3.43421e+00_prt, -3.28804e+00_prt, -3.14189e+00_prt,
        -2.99575e+00_prt, -2.84963e+00_prt, -2.70354e+00_prt, -2.55749e+00_prt,
        -2.41148e+00_prt, -2.26554e+00_prt, -2.11970e+00_prt, -1.97398e+00_prt,
        -1.82844e+00_prt, -1.68314e+00_prt, -1.53817e+00_prt, -1.39365e+00_prt,
        -1.24978e+00_prt, -1.10681e+00_prt, -9.65094e-01_prt, -8.25166e-01_prt,
        -6.87778e-01_prt, -5.54020e-01_prt, -4.25469e-01_prt, -3.04385e-01_prt,
        -1.93953e-01_prt, -9.85428e-02_prt, -2.52054e-02_prt, 0.00000e+00_prt,
        -8.87064e+00_prt, -8.72445e+00_prt, -8.57825e+00_prt, -8.43206e+00_prt,
        -8.28586e+00_prt, -8.13966e+00_prt, -7.99347e+00_prt, -7.84727e+00_prt,
        -7.70108e+00_prt, -7.55488e+00_prt, -7.40868e+00_prt, -7.26249e+00_prt,
        -7.11629e+00_prt, -6.97010e+00_prt, -6.82390e+00_prt, -6.67770e+00_prt,
        -6.53151e+00_prt, -6.38531e+00_prt, -6.23912e+00_prt, -6.09292e+00_prt,
        -5.94673e+00_prt, -5.80053e+00_prt, -5.65433e+00_prt, -5.50814e+00_prt,
        -5.36194e+00_prt, -5.21575e+00_prt, -5.06955e+00_prt, -4.92336e+00_prt,
        -4.77716e+00_prt, -4.63097e+00_prt, -4.48478e+00_prt, -4.33858e+00_prt,
        -4.19239e+00_prt, -4.04620e+00_prt, -3.90002e+00_prt, -3.75383e+00_prt,
        -3.60765e+00_prt, -3.46148e+00_prt, -3.31531e+00_prt, -3.16915e+00_prt,
        -3.02301e+00_prt, -2.87688e+00_prt, -2.73077e+00_prt, -2.58470e+00_prt,
        -2.43868e+00_prt, -2.29271e+00_prt, -2.14683e+00_prt, -2.00106e+00_prt,
        -1.85545e+00_prt, -1.71005e+00_prt, -1.56496e+00_prt, -1.42028e+00_prt,
        -1.27618e+00_prt, -1.13290e+00_prt, -9.90777e-01_prt, -8.50295e-01_prt,
        -7.12155e-01_prt, -5.77370e-01_prt, -4.47406e-01_prt, -3.24362e-01_prt,
        -2.11168e-01_prt, -1.11740e-01_prt, -3.20077e-02_prt, 0.00000e+00_prt,
        -8.89603e+00_prt, -8.74984e+00_prt, -8.60364e+00_prt, -8.45745e+00_prt,
        -8.31125e+00_prt, -8.16505e+00_prt, -8.01886e+00_prt, -7.87266e+00_prt,
        -7.72647e+00_prt, -7.58027e+00_prt, -7.43407e+00_prt, -7.28788e+00_prt,
        -7.14168e+00_prt, -6.99549e+00_prt, -6.84929e+00_prt, -6.70309e+00_prt,
        -6.55690e+00_prt, -6.41070e+00_prt, -6.26451e+00_prt, -6.11831e+00_prt,
        -5.97212e+00_prt, -5.82592e+00_prt, -5.67972e+00_prt, -5.53353e+00_prt,
        -5.38733e+00_prt, -5.24114e+00_prt, -5.09494e+00_prt, -4.94875e+00_prt,
        -4.80255e+00_prt, -4.65636e+00_prt, -4.51017e+00_prt, -4.36397e+00_prt,
        -4.21778e+00_prt, -4.07159e+00_prt, -3.92540e+00_prt, -3.77922e+00_prt,
        -3.63303e+00_prt, -3.48686e+00_prt, -3.34069e+00_prt, -3.19452e+00_prt,
        -3.04837e+00_prt, -2.90224e+00_prt, -2.75612e+00_prt, -2.61004e+00_prt,
        -2.46399e+00_prt, -2.31800e+00_prt, -2.17209e+00_prt, -2.02627e+00_prt,
        -1.88060e+00_prt, -1.73512e+00_prt, -1.58992e+00_prt, -1.44509e+00_prt,
        -1.30079e+00_prt, -1.15725e+00_prt, -1.01477e+00_prt, -8.73798e-01_prt,
        -7.34999e-01_prt, -5.99315e-01_prt, -4.68118e-01_prt, -3.43362e-01_prt,
        -2.27755e-01_prt, -1.24807e-01_prt, -3.93469e-02_prt, 0.00000e+00_prt,
        -8.91959e+00_prt, -8.77339e+00_prt, -8.62720e+00_prt, -8.48100e+00_prt,
        -8.33481e+00_prt, -8.18861e+00_prt, -8.04241e+00_prt, -7.89622e+00_prt,
        -7.75002e+00_prt, -7.60383e+00_prt, -7.45763e+00_prt, -7.31143e+00_prt,
        -7.16524e+00_prt, -7.01904e+00_prt, -6.87285e+00_prt, -6.72665e+00_prt,
        -6.58046e+00_prt, -6.43426e+00_prt, -6.28806e+00_prt, -6.14187e+00_prt,
        -5.99567e+00_prt, -5.84948e+00_prt, -5.70328e+00_prt, -5.55709e+00_prt,
        -5.41089e+00_prt, -5.26469e+00_prt, -5.11850e+00_prt, -4.97230e+00_prt,
        -4.82611e+00_prt, -4.67991e+00_prt, -4.53372e+00_prt, -4.38753e+00_prt,
        -4.24134e+00_prt, -4.09514e+00_prt, -3.94896e+00_prt, -3.80277e+00_prt,
        -3.65658e+00_prt, -3.51040e+00_prt, -3.36423e+00_prt, -3.21806e+00_prt,
        -3.07191e+00_prt, -2.92577e+00_prt, -2.77964e+00_prt, -2.63355e+00_prt,
        -2.48748e+00_prt, -2.34147e+00_prt, -2.19553e+00_prt, -2.04967e+00_prt,
        -1.90394e+00_prt, -1.75840e+00_prt, -1.61309e+00_prt, -1.46814e+00_prt,
        -1.32367e+00_prt, -1.17989e+00_prt, -1.03709e+00_prt, -8.95699e-01_prt,
        -7.56322e-01_prt, -6.19852e-01_prt, -4.87576e-01_prt, -3.61326e-01_prt,
        -2.43612e-01_prt, -1.37588e-01_prt, -4.70555e-02_prt, 0.00000e+00_prt,
        -8.94138e+00_prt, -8.79518e+00_prt, -8.64898e+00_prt, -8.50279e+00_prt,
        -8.35659e+00_prt, -8.21040e+00_prt, -8.06420e+00_prt, -7.91801e+00_prt,
        -7.77181e+00_prt, -7.62561e+00_prt, -7.47942e+00_prt, -7.33322e+00_prt,
        -7.18703e+00_prt, -7.04083e+00_prt, -6.89463e+00_prt, -6.74844e+00_prt,
        -6.60224e+00_prt, -6.45605e+00_prt, -6.30985e+00_prt, -6.16366e+00_prt,
        -6.01746e+00_prt, -5.87126e+00_prt, -5.72507e+00_prt, -5.57887e+00_prt,
        -5.43268e+00_prt, -5.28648e+00_prt, -5.14029e+00_prt, -4.99409e+00_prt,
        -4.84790e+00_prt, -4.70170e+00_prt, -4.55551e+00_prt, -4.40931e+00_prt,
        -4.26312e+00_prt, -4.11693e+00_prt, -3.97074e+00_prt, -3.82455e+00_prt,
        -3.67837e+00_prt, -3.53218e+00_prt, -3.38601e+00_prt, -3.23984e+00_prt,
        -3.09368e+00_prt, -2.94753e+00_prt, -2.80140e+00_prt, -2.65529e+00_prt,
        -2.50921e+00_prt, -2.36318e+00_prt, -2.21721e+00_prt, -2.07132e+00_prt,
        -1.92555e+00_prt, -1.77993e+00_prt, -1.63455e+00_prt, -1.48948e+00_prt,
        -1.34486e+00_prt, -1.20087e+00_prt, -1.05780e+00_prt, -9.16033e-01_prt,
        -7.76151e-01_prt, -6.38993e-01_prt, -5.05773e-01_prt, -3.78216e-01_prt,
        -2.58665e-01_prt, -1.49958e-01_prt, -5.49705e-02_prt, 0.00000e+00_prt,
        -8.96147e+00_prt, -8.81527e+00_prt, -8.66907e+00_prt, -8.52288e+00_prt,
        -8.37668e+00_prt, -8.23049e+00_prt, -8.08429e+00_prt, -7.93809e+00_prt,
        -7.79190e+00_prt, -7.64570e+00_prt, -7.49951e+00_prt, -7.35331e+00_prt,
        -7.20712e+00_prt, -7.06092e+00_prt, -6.91472e+00_prt, -6.76853e+00_prt,
        -6.62233e+00_prt, -6.47614e+00_prt, -6.32994e+00_prt, -6.18374e+00_prt,
        -6.03755e+00_prt, -5.89135e+00_prt, -5.74516e+00_prt, -5.59896e+00_prt,
        -5.45277e+00_prt, -5.30657e+00_prt, -5.16037e+00_prt, -5.01418e+00_prt,
        -4.86798e+00_prt, -4.72179e+00_prt, -4.57560e+00_prt, -4.42940e+00_prt,
        -4.28321e+00_prt, -4.13702e+00_prt, -3.99083e+00_prt, -3.84464e+00_prt,
        -3.69845e+00_prt, -3.55227e+00_prt, -3.40609e+00_prt, -3.25992e+00_prt,
        -3.11375e+00_prt, -2.96760e+00_prt, -2.82146e+00_prt, -2.67534e+00_prt,
        -2.52925e+00_prt, -2.38320e+00_prt, -2.23721e+00_prt, -2.09129e+00_prt,
        -1.94547e+00_prt, -1.79981e+00_prt, -1.65434e+00_prt, -1.50918e+00_prt,
        -1.36442e+00_prt, -1.22026e+00_prt, -1.07694e+00_prt, -9.34847e-01_prt,
        -7.94522e-01_prt, -6.56761e-01_prt, -5.22715e-01_prt, -3.94018e-01_prt,
        -2.72862e-01_prt, -1.61817e-01_prt, -6.29425e-02_prt, 0.00000e+00_prt,
        -8.97993e+00_prt, -8.83374e+00_prt, -8.68754e+00_prt, -8.54135e+00_prt,
        -8.39515e+00_prt, -8.24895e+00_prt, -8.10276e+00_prt, -7.95656e+00_prt,
        -7.81037e+00_prt, -7.66417e+00_prt, -7.51797e+00_prt, -7.37178e+00_prt,
        -7.22558e+00_prt, -7.07939e+00_prt, -6.93319e+00_prt, -6.78700e+00_prt,
        -6.64080e+00_prt, -6.49460e+00_prt, -6.34841e+00_prt, -6.20221e+00_prt,
        -6.05602e+00_prt, -5.90982e+00_prt, -5.76362e+00_prt, -5.61743e+00_prt,
        -5.47123e+00_prt, -5.32504e+00_prt, -5.17884e+00_prt, -5.03265e+00_prt,
        -4.88645e+00_prt, -4.74026e+00_prt, -4.59406e+00_prt, -4.44787e+00_prt,
        -4.30168e+00_prt, -4.15548e+00_prt, -4.00929e+00_prt, -3.86310e+00_prt,
        -3.71691e+00_prt, -3.57073e+00_prt, -3.42455e+00_prt, -3.27837e+00_prt,
        -3.13220e+00_prt, -2.98605e+00_prt, -2.83990e+00_prt, -2.69378e+00_prt,
        -2.54768e+00_prt, -2.40161e+00_prt, -2.25560e+00_prt, -2.10965e+00_prt,
        -1.96380e+00_prt, -1.81808e+00_prt, -1.67256e+00_prt, -1.52730e+00_prt,
        -1.38243e+00_prt, -1.23810e+00_prt, -1.09458e+00_prt, -9.52195e-01_prt,
        -8.11482e-01_prt, -6.73192e-01_prt, -5.38425e-01_prt, -4.08729e-01_prt,
        -2.86173e-01_prt, -1.73092e-01_prt, -7.08416e-02_prt, 0.00000e+00_prt,
        -8.99686e+00_prt, -8.85067e+00_prt, -8.70447e+00_prt, -8.55827e+00_prt,
        -8.41208e+00_prt, -8.26588e+00_prt, -8.11969e+00_prt, -7.97349e+00_prt,
        -7.82729e+00_prt, -7.68110e+00_prt, -7.53490e+00_prt, -7.38871e+00_prt,
        -7.24251e+00_prt, -7.09631e+00_prt, -6.95012e+00_prt, -6.80392e+00_prt,
        -6.65773e+00_prt, -6.51153e+00_prt, -6.36534e+00_prt, -6.21914e+00_prt,
        -6.07294e+00_prt, -5.92675e+00_prt, -5.78055e+00_prt, -5.63436e+00_prt,
        -5.48816e+00_prt, -5.34197e+00_prt, -5.19577e+00_prt, -5.04957e+00_prt,
        -4.90338e+00_prt, -4.75718e+00_prt, -4.61099e+00_prt, -4.46480e+00_prt,
        -4.31860e+00_prt, -4.17241e+00_prt, -4.02622e+00_prt, -3.88003e+00_prt,
        -3.73384e+00_prt, -3.58765e+00_prt, -3.44147e+00_prt, -3.29529e+00_prt,
        -3.14912e+00_prt, -3.00296e+00_prt, -2.85681e+00_prt, -2.71067e+00_prt,
        -2.56456e+00_prt, -2.41849e+00_prt, -2.27245e+00_prt, -2.12648e+00_prt,
        -1.98060e+00_prt, -1.83484e+00_prt, -1.68926e+00_prt, -1.54392e+00_prt,
        -1.39895e+00_prt, -1.25449e+00_prt, -1.11077e+00_prt, -9.68140e-01_prt,
        -8.27086e-01_prt, -6.88334e-01_prt, -5.52934e-01_prt, -4.22365e-01_prt,
        -2.98586e-01_prt, -1.83733e-01_prt, -7.85591e-02_prt, 0.00000e+00_prt,
        -9.01233e+00_prt, -8.86614e+00_prt, -8.71994e+00_prt, -8.57375e+00_prt,
        -8.42755e+00_prt, -8.28135e+00_prt, -8.13516e+00_prt, -7.98896e+00_prt,
        -7.84277e+00_prt, -7.69657e+00_prt, -7.55037e+00_prt, -7.40418e+00_prt,
        -7.25798e+00_prt, -7.11179e+00_prt, -6.96559e+00_prt, -6.81939e+00_prt,
        -6.67320e+00_prt, -6.52700e+00_prt, -6.38081e+00_prt, -6.23461e+00_prt,
        -6.08842e+00_prt, -5.94222e+00_prt, -5.79602e+00_prt, -5.64983e+00_prt,
        -5.50363e+00_prt, -5.35744e+00_prt, -5.21124e+00_prt, -5.06505e+00_prt,
        -4.91885e+00_prt, -4.77266e+00_prt, -4.62646e+00_prt, -4.48027e+00_prt,
        -4.33407e+00_prt, -4.18788e+00_prt, -4.04169e+00_prt, -3.89550e+00_prt,
        -3.74931e+00_prt, -3.60312e+00_prt, -3.45694e+00_prt, -3.31076e+00_prt,
        -3.16458e+00_prt, -3.01842e+00_prt, -2.87226e+00_prt, -2.72612e+00_prt,
        -2.58000e+00_prt, -2.43391e+00_prt, -2.28786e+00_prt, -2.14187e+00_prt,
        -1.99596e+00_prt, -1.85017e+00_prt, -1.70453e+00_prt, -1.55913e+00_prt,
        -1.41406e+00_prt, -1.26948e+00_prt, -1.12560e+00_prt, -9.82749e-01_prt,
        -8.41396e-01_prt, -7.02238e-01_prt, -5.66284e-01_prt, -4.34952e-01_prt,
        -3.10105e-01_prt, -1.93707e-01_prt, -8.60082e-02_prt, 0.00000e+00_prt,
        -9.02644e+00_prt, -8.88024e+00_prt, -8.73404e+00_prt, -8.58785e+00_prt,
        -8.44165e+00_prt, -8.29546e+00_prt, -8.14926e+00_prt, -8.00306e+00_prt,
        -7.85687e+00_prt, -7.71067e+00_prt, -7.56448e+00_prt, -7.41828e+00_prt,
        -7.27209e+00_prt, -7.12589e+00_prt, -6.97969e+00_prt, -6.83350e+00_prt,
        -6.68730e+00_prt, -6.54111e+00_prt, -6.39491e+00_prt, -6.24871e+00_prt,
        -6.10252e+00_prt, -5.95632e+00_prt, -5.81013e+00_prt, -5.66393e+00_prt,
        -5.51774e+00_prt, -5.37154e+00_prt, -5.22534e+00_prt, -5.07915e+00_prt,
        -4.93295e+00_prt, -4.78676e+00_prt, -4.64056e+00_prt, -4.49437e+00_prt,
        -4.34817e+00_prt, -4.20198e+00_prt, -4.05579e+00_prt, -3.90960e+00_prt,
        -3.76341e+00_prt, -3.61722e+00_prt, -3.47103e+00_prt, -3.32485e+00_prt,
        -3.17867e+00_prt, -3.03251e+00_prt, -2.88635e+00_prt, -2.74020e+00_prt,
        -2.59407e+00_prt, -2.44798e+00_prt, -2.30191e+00_prt, -2.15590e+00_prt,
        -2.00997e+00_prt, -1.86414e+00_prt, -1.71846e+00_prt, -1.57300e+00_prt,
        -1.42785e+00_prt, -1.28316e+00_prt, -1.13914e+00_prt, -9.96094e-01_prt,
        -8.54478e-01_prt, -7.14965e-01_prt, -5.78526e-01_prt, -4.46526e-01_prt,
        -3.20745e-01_prt, -2.03002e-01_prt, -9.31232e-02_prt, 0.00000e+00_prt,
        -9.03926e+00_prt, -8.89306e+00_prt, -8.74686e+00_prt, -8.60067e+00_prt,
        -8.45447e+00_prt, -8.30828e+00_prt, -8.16208e+00_prt, -8.01589e+00_prt,
        -7.86969e+00_prt, -7.72349e+00_prt, -7.57730e+00_prt, -7.43110e+00_prt,
        -7.28491e+00_prt, -7.13871e+00_prt, -6.99251e+00_prt, -6.84632e+00_prt,
        -6.70012e+00_prt, -6.55393e+00_prt, -6.40773e+00_prt, -6.26153e+00_prt,
        -6.11534e+00_prt, -5.96914e+00_prt, -5.82295e+00_prt, -5.67675e+00_prt,
        -5.53056e+00_prt, -5.38436e+00_prt, -5.23816e+00_prt, -5.09197e+00_prt,
        -4.94577e+00_prt, -4.79958e+00_prt, -4.65338e+00_prt, -4.50719e+00_prt,
        -4.36099e+00_prt, -4.21480e+00_prt, -4.06861e+00_prt, -3.92241e+00_prt,
        -3.77622e+00_prt, -3.63004e+00_prt, -3.48385e+00_prt, -3.33767e+00_prt,
        -3.19149e+00_prt, -3.04532e+00_prt, -2.89915e+00_prt, -2.75300e+00_prt,
        -2.60687e+00_prt, -2.46076e+00_prt, -2.31469e+00_prt, -2.16866e+00_prt,
        -2.02270e+00_prt, -1.87684e+00_prt, -1.73113e+00_prt, -1.58561e+00_prt,
        -1.44040e+00_prt, -1.29562e+00_prt, -1.15147e+00_prt, -1.00825e+00_prt,
        -8.66403e-01_prt, -7.26578e-01_prt, -5.89713e-01_prt, -4.57128e-01_prt,
        -3.30531e-01_prt, -2.11616e-01_prt, -9.98571e-02_prt, 0.00000e+00_prt,
        -9.05088e+00_prt, -8.90469e+00_prt, -8.75849e+00_prt, -8.61229e+00_prt,
        -8.46610e+00_prt, -8.31990e+00_prt, -8.17371e+00_prt, -8.02751e+00_prt,
        -7.88132e+00_prt, -7.73512e+00_prt, -7.58892e+00_prt, -7.44273e+00_prt,
        -7.29653e+00_prt, -7.15034e+00_prt, -7.00414e+00_prt, -6.85794e+00_prt,
        -6.71175e+00_prt, -6.56555e+00_prt, -6.41936e+00_prt, -6.27316e+00_prt,
        -6.12696e+00_prt, -5.98077e+00_prt, -5.83457e+00_prt, -5.68838e+00_prt,
        -5.54218e+00_prt, -5.39599e+00_prt, -5.24979e+00_prt, -5.10359e+00_prt,
        -4.95740e+00_prt, -4.81120e+00_prt, -4.66501e+00_prt, -4.51881e+00_prt,
        -4.37262e+00_prt, -4.22643e+00_prt, -4.08023e+00_prt, -3.93404e+00_prt,
        -3.78785e+00_prt, -3.64166e+00_prt, -3.49547e+00_prt, -3.34929e+00_prt,
        -3.20311e+00_prt, -3.05693e+00_prt, -2.91077e+00_prt, -2.76461e+00_prt,
        -2.61847e+00_prt, -2.47236e+00_prt, -2.32627e+00_prt, -2.18023e+00_prt,
        -2.03425e+00_prt, -1.88837e+00_prt, -1.74262e+00_prt, -1.59706e+00_prt,
        -1.45178e+00_prt, -1.30692e+00_prt, -1.16266e+00_prt, -1.01929e+00_prt,
        -8.77243e-01_prt, -7.37144e-01_prt, -5.99906e-01_prt, -4.66808e-01_prt,
        -3.39497e-01_prt, -2.19560e-01_prt, -1.06179e-01_prt, 0.00000e+00_prt,
        -9.06140e+00_prt, -8.91520e+00_prt, -8.76901e+00_prt, -8.62281e+00_prt,
        -8.47662e+00_prt, -8.33042e+00_prt, -8.18422e+00_prt, -8.03803e+00_prt,
        -7.89183e+00_prt, -7.74564e+00_prt, -7.59944e+00_prt, -7.45324e+00_prt,
        -7.30705e+00_prt, -7.16085e+00_prt, -7.01466e+00_prt, -6.86846e+00_prt,
        -6.72227e+00_prt, -6.57607e+00_prt, -6.42987e+00_prt, -6.28368e+00_prt,
        -6.13748e+00_prt, -5.99129e+00_prt, -5.84509e+00_prt, -5.69889e+00_prt,
        -5.55270e+00_prt, -5.40650e+00_prt, -5.26031e+00_prt, -5.11411e+00_prt,
        -4.96792e+00_prt, -4.82172e+00_prt, -4.67553e+00_prt, -4.52933e+00_prt,
        -4.38314e+00_prt, -4.23694e+00_prt, -4.09075e+00_prt, -3.94455e+00_prt,
        -3.79836e+00_prt, -3.65217e+00_prt, -3.50598e+00_prt, -3.35980e+00_prt,
        -3.21362e+00_prt, -3.06744e+00_prt, -2.92127e+00_prt, -2.77511e+00_prt,
        -2.62897e+00_prt, -2.48285e+00_prt, -2.33675e+00_prt, -2.19070e+00_prt,
        -2.04471e+00_prt, -1.89880e+00_prt, -1.75302e+00_prt, -1.60742e+00_prt,
        -1.46209e+00_prt, -1.31715e+00_prt, -1.17279e+00_prt, -1.02929e+00_prt,
        -8.87070e-01_prt, -7.46731e-01_prt, -6.09166e-01_prt, -4.75619e-01_prt,
        -3.47682e-01_prt, -2.26853e-01_prt, -1.12073e-01_prt, 0.00000e+00_prt,
        -9.07089e+00_prt, -8.92470e+00_prt, -8.77850e+00_prt, -8.63230e+00_prt,
        -8.48611e+00_prt, -8.33991e+00_prt, -8.19372e+00_prt, -8.04752e+00_prt,
        -7.90133e+00_prt, -7.75513e+00_prt, -7.60893e+00_prt, -7.46274e+00_prt,
        -7.31654e+00_prt, -7.17035e+00_prt, -7.02415e+00_prt, -6.87795e+00_prt,
        -6.73176e+00_prt, -6.58556e+00_prt, -6.43937e+00_prt, -6.29317e+00_prt,
        -6.14697e+00_prt, -6.00078e+00_prt, -5.85458e+00_prt, -5.70839e+00_prt,
        -5.56219e+00_prt, -5.41600e+00_prt, -5.26980e+00_prt, -5.12360e+00_prt,
        -4.97741e+00_prt, -4.83121e+00_prt, -4.68502e+00_prt, -4.53882e+00_prt,
        -4.39263e+00_prt, -4.24643e+00_prt, -4.10024e+00_prt, -3.95405e+00_prt,
        -3.80785e+00_prt, -3.66166e+00_prt, -3.51547e+00_prt, -3.36929e+00_prt,
        -3.22310e+00_prt, -3.07693e+00_prt, -2.93075e+00_prt, -2.78459e+00_prt,
        -2.63844e+00_prt, -2.49231e+00_prt, -2.34621e+00_prt, -2.20015e+00_prt,
        -2.05414e+00_prt, -1.90821e+00_prt, -1.76241e+00_prt, -1.61677e+00_prt,
        -1.47139e+00_prt, -1.32639e+00_prt, -1.18195e+00_prt, -1.03834e+00_prt,
        -8.95956e-01_prt, -7.55406e-01_prt, -6.17555e-01_prt, -4.83614e-01_prt,
        -3.55130e-01_prt, -2.33523e-01_prt, -1.17533e-01_prt, 0.00000e+00_prt,
        -9.07944e+00_prt, -8.93325e+00_prt, -8.78705e+00_prt, -8.64085e+00_prt,
        -8.49466e+00_prt, -8.34846e+00_prt, -8.20227e+00_prt, -8.05607e+00_prt,
        -7.90988e+00_prt, -7.76368e+00_prt, -7.61748e+00_prt, -7.47129e+00_prt,
        -7.32509e+00_prt, -7.17890e+00_prt, -7.03270e+00_prt, -6.88650e+00_prt,
        -6.74031e+00_prt, -6.59411e+00_prt, -6.44792e+00_prt, -6.30172e+00_prt,
        -6.15552e+00_prt, -6.00933e+00_prt, -5.86313e+00_prt, -5.71694e+00_prt,
        -5.57074e+00_prt, -5.42455e+00_prt, -5.27835e+00_prt, -5.13215e+00_prt,
        -4.98596e+00_prt, -4.83976e+00_prt, -4.69357e+00_prt, -4.54737e+00_prt,
        -4.40118e+00_prt, -4.25498e+00_prt, -4.10879e+00_prt, -3.96260e+00_prt,
        -3.81640e+00_prt, -3.67021e+00_prt, -3.52402e+00_prt, -3.37783e+00_prt,
        -3.23165e+00_prt, -3.08547e+00_prt, -2.93930e+00_prt, -2.79313e+00_prt,
        -2.64698e+00_prt, -2.50084e+00_prt, -2.35473e+00_prt, -2.20866e+00_prt,
        -2.06264e+00_prt, -1.91670e+00_prt, -1.77087e+00_prt, -1.62520e+00_prt,
        -1.47978e+00_prt, -1.33472e+00_prt, -1.19020e+00_prt, -1.04649e+00_prt,
        -9.03973e-01_prt, -7.63238e-01_prt, -6.25135e-01_prt, -4.90849e-01_prt,
        -3.61886e-01_prt, -2.39599e-01_prt, -1.22564e-01_prt, 0.00000e+00_prt,
        -9.08713e+00_prt, -8.94093e+00_prt, -8.79474e+00_prt, -8.64854e+00_prt,
        -8.50234e+00_prt, -8.35615e+00_prt, -8.20995e+00_prt, -8.06376e+00_prt,
        -7.91756e+00_prt, -7.77136e+00_prt, -7.62517e+00_prt, -7.47897e+00_prt,
        -7.33278e+00_prt, -7.18658e+00_prt, -7.04038e+00_prt, -6.89419e+00_prt,
        -6.74799e+00_prt, -6.60180e+00_prt, -6.45560e+00_prt, -6.30941e+00_prt,
        -6.16321e+00_prt, -6.01701e+00_prt, -5.87082e+00_prt, -5.72462e+00_prt,
        -5.57843e+00_prt, -5.43223e+00_prt, -5.28603e+00_prt, -5.13984e+00_prt,
        -4.99364e+00_prt, -4.84745e+00_prt, -4.70125e+00_prt, -4.55506e+00_prt,
        -4.40886e+00_prt, -4.26267e+00_prt, -4.11647e+00_prt, -3.97028e+00_prt,
        -3.82409e+00_prt, -3.67789e+00_prt, -3.53170e+00_prt, -3.38552e+00_prt,
        -3.23933e+00_prt, -3.09315e+00_prt, -2.94697e+00_prt, -2.80080e+00_prt,
        -2.65465e+00_prt, -2.50851e+00_prt, -2.36239e+00_prt, -2.21631e+00_prt,
        -2.07028e+00_prt, -1.92432e+00_prt, -1.77847e+00_prt, -1.63278e+00_prt,
        -1.48732e+00_prt, -1.34221e+00_prt, -1.19763e+00_prt, -1.05383e+00_prt,
        -9.11190e-01_prt, -7.70293e-01_prt, -6.31969e-01_prt, -4.97380e-01_prt,
        -3.67997e-01_prt, -2.45116e-01_prt, -1.27177e-01_prt, 0.00000e+00_prt,
        -9.09402e+00_prt, -8.94783e+00_prt, -8.80163e+00_prt, -8.65543e+00_prt,
        -8.50924e+00_prt, -8.36304e+00_prt, -8.21685e+00_prt, -8.07065e+00_prt,
        -7.92445e+00_prt, -7.77826e+00_prt, -7.63206e+00_prt, -7.48587e+00_prt,
        -7.33967e+00_prt, -7.19347e+00_prt, -7.04728e+00_prt, -6.90108e+00_prt,
        -6.75489e+00_prt, -6.60869e+00_prt, -6.46250e+00_prt, -6.31630e+00_prt,
        -6.17010e+00_prt, -6.02391e+00_prt, -5.87771e+00_prt, -5.73152e+00_prt,
        -5.58532e+00_prt, -5.43912e+00_prt, -5.29293e+00_prt, -5.14673e+00_prt,
        -5.00054e+00_prt, -4.85434e+00_prt, -4.70815e+00_prt, -4.56195e+00_prt,
        -4.41576e+00_prt, -4.26956e+00_prt, -4.12337e+00_prt, -3.97717e+00_prt,
        -3.83098e+00_prt, -3.68479e+00_prt, -3.53860e+00_prt, -3.39241e+00_prt,
        -3.24622e+00_prt, -3.10004e+00_prt, -2.95386e+00_prt, -2.80769e+00_prt,
        -2.66153e+00_prt, -2.51539e+00_prt, -2.36927e+00_prt, -2.22318e+00_prt,
        -2.07714e+00_prt, -1.93116e+00_prt, -1.78529e+00_prt, -1.63958e+00_prt,
        -1.49409e+00_prt, -1.34894e+00_prt, -1.20430e+00_prt, -1.06041e+00_prt,
        -9.17673e-01_prt, -7.76634e-01_prt, -6.38116e-01_prt, -5.03261e-01_prt,
        -3.73511e-01_prt, -2.50110e-01_prt, -1.31387e-01_prt, 0.00000e+00_prt,
        -9.10020e+00_prt, -8.95400e+00_prt, -8.80780e+00_prt, -8.66161e+00_prt,
        -8.51541e+00_prt, -8.36922e+00_prt, -8.22302e+00_prt, -8.07682e+00_prt,
        -7.93063e+00_prt, -7.78443e+00_prt, -7.63824e+00_prt, -7.49204e+00_prt,
        -7.34585e+00_prt, -7.19965e+00_prt, -7.05345e+00_prt, -6.90726e+00_prt,
        -6.76106e+00_prt, -6.61487e+00_prt, -6.46867e+00_prt, -6.32247e+00_prt,
        -6.17628e+00_prt, -6.03008e+00_prt, -5.88389e+00_prt, -5.73769e+00_prt,
        -5.59149e+00_prt, -5.44530e+00_prt, -5.29910e+00_prt, -5.15291e+00_prt,
        -5.00671e+00_prt, -4.86052e+00_prt, -4.71432e+00_prt, -4.56813e+00_prt,
        -4.42193e+00_prt, -4.27574e+00_prt, -4.12954e+00_prt, -3.98335e+00_prt,
        -3.83715e+00_prt, -3.69096e+00_prt, -3.54477e+00_prt, -3.39858e+00_prt,
        -3.25239e+00_prt, -3.10621e+00_prt, -2.96003e+00_prt, -2.81386e+00_prt,
        -2.66769e+00_prt, -2.52155e+00_prt, -2.37542e+00_prt, -2.22933e+00_prt,
        -2.08327e+00_prt, -1.93729e+00_prt, -1.79141e+00_prt, -1.64567e+00_prt,
        -1.50015e+00_prt, -1.35496e+00_prt, -1.21027e+00_prt, -1.06632e+00_prt,
        -9.23486e-01_prt, -7.82322e-01_prt, -6.43634e-01_prt, -5.08546e-01_prt,
        -3.78473e-01_prt, -2.54618e-01_prt, -1.35215e-01_prt, 0.00000e+00_prt,
        -9.10572e+00_prt, -8.95952e+00_prt, -8.81333e+00_prt, -8.66713e+00_prt,
        -8.52093e+00_prt, -8.37474e+00_prt, -8.22854e+00_prt, -8.08235e+00_prt,
        -7.93615e+00_prt, -7.78995e+00_prt, -7.64376e+00_prt, -7.49756e+00_prt,
        -7.35137e+00_prt, -7.20517e+00_prt, -7.05897e+00_prt, -6.91278e+00_prt,
        -6.76658e+00_prt, -6.62039e+00_prt, -6.47419e+00_prt, -6.32799e+00_prt,
        -6.18180e+00_prt, -6.03560e+00_prt, -5.88941e+00_prt, -5.74321e+00_prt,
        -5.59702e+00_prt, -5.45082e+00_prt, -5.30462e+00_prt, -5.15843e+00_prt,
        -5.01223e+00_prt, -4.86604e+00_prt, -4.71984e+00_prt, -4.57365e+00_prt,
        -4.42745e+00_prt, -4.28126e+00_prt, -4.13506e+00_prt, -3.98887e+00_prt,
        -3.84267e+00_prt, -3.69648e+00_prt, -3.55029e+00_prt, -3.40410e+00_prt,
        -3.25791e+00_prt, -3.11172e+00_prt, -2.96554e+00_prt, -2.81937e+00_prt,
        -2.67321e+00_prt, -2.52706e+00_prt, -2.38092e+00_prt, -2.23482e+00_prt,
        -2.08877e+00_prt, -1.94277e+00_prt, -1.79687e+00_prt, -1.65112e+00_prt,
        -1.50557e+00_prt, -1.36035e+00_prt, -1.21562e+00_prt, -1.07160e+00_prt,
        -9.28689e-01_prt, -7.87415e-01_prt, -6.48578e-01_prt, -5.13286e-01_prt,
        -3.82930e-01_prt, -2.58677e-01_prt, -1.38684e-01_prt, 0.00000e+00_prt,
        -9.11065e+00_prt, -8.96445e+00_prt, -8.81825e+00_prt, -8.67206e+00_prt,
        -8.52586e+00_prt, -8.37967e+00_prt, -8.23347e+00_prt, -8.08727e+00_prt,
        -7.94108e+00_prt, -7.79488e+00_prt, -7.64869e+00_prt, -7.50249e+00_prt,
        -7.35629e+00_prt, -7.21010e+00_prt, -7.06390e+00_prt, -6.91771e+00_prt,
        -6.77151e+00_prt, -6.62531e+00_prt, -6.47912e+00_prt, -6.33292e+00_prt,
        -6.18673e+00_prt, -6.04053e+00_prt, -5.89434e+00_prt, -5.74814e+00_prt,
        -5.60194e+00_prt, -5.45575e+00_prt, -5.30955e+00_prt, -5.16336e+00_prt,
        -5.01716e+00_prt, -4.87096e+00_prt, -4.72477e+00_prt, -4.57857e+00_prt,
        -4.43238e+00_prt, -4.28618e+00_prt, -4.13999e+00_prt, -3.99379e+00_prt,
        -3.84760e+00_prt, -3.70141e+00_prt, -3.55521e+00_prt, -3.40902e+00_prt,
        -3.26283e+00_prt, -3.11665e+00_prt, -2.97047e+00_prt, -2.82429e+00_prt,
        -2.67813e+00_prt, -2.53197e+00_prt, -2.38584e+00_prt, -2.23973e+00_prt,
        -2.09367e+00_prt, -1.94766e+00_prt, -1.80175e+00_prt, -1.65598e+00_prt,
        -1.51042e+00_prt, -1.36517e+00_prt, -1.22039e+00_prt, -1.07633e+00_prt,
        -9.33339e-01_prt, -7.91968e-01_prt, -6.53000e-01_prt, -5.17528e-01_prt,
        -3.86924e-01_prt, -2.62323e-01_prt, -1.41817e-01_prt, 0.00000e+00_prt,
        -9.11504e+00_prt, -8.96884e+00_prt, -8.82265e+00_prt, -8.67645e+00_prt,
        -8.53025e+00_prt, -8.38406e+00_prt, -8.23786e+00_prt, -8.09167e+00_prt,
        -7.94547e+00_prt, -7.79928e+00_prt, -7.65308e+00_prt, -7.50688e+00_prt,
        -7.36069e+00_prt, -7.21449e+00_prt, -7.06830e+00_prt, -6.92210e+00_prt,
        -6.77590e+00_prt, -6.62971e+00_prt, -6.48351e+00_prt, -6.33732e+00_prt,
        -6.19112e+00_prt, -6.04492e+00_prt, -5.89873e+00_prt, -5.75253e+00_prt,
        -5.60634e+00_prt, -5.46014e+00_prt, -5.31395e+00_prt, -5.16775e+00_prt,
        -5.02155e+00_prt, -4.87536e+00_prt, -4.72916e+00_prt, -4.58297e+00_prt,
        -4.43677e+00_prt, -4.29058e+00_prt, -4.14438e+00_prt, -3.99819e+00_prt,
        -3.85199e+00_prt, -3.70580e+00_prt, -3.55961e+00_prt, -3.41342e+00_prt,
        -3.26723e+00_prt, -3.12104e+00_prt, -2.97486e+00_prt, -2.82868e+00_prt,
        -2.68251e+00_prt, -2.53636e+00_prt, -2.39022e+00_prt, -2.24411e+00_prt,
        -2.09804e+00_prt, -1.95203e+00_prt, -1.80610e+00_prt, -1.66032e+00_prt,
        -1.51474e+00_prt, -1.36946e+00_prt, -1.22465e+00_prt, -1.08054e+00_prt,
        -9.37487e-01_prt, -7.96032e-01_prt, -6.56948e-01_prt, -5.21319e-01_prt,
        -3.90498e-01_prt, -2.65591e-01_prt, -1.44639e-01_prt, 0.00000e+00_prt,
        -9.11895e+00_prt, -8.97275e+00_prt, -8.82656e+00_prt, -8.68036e+00_prt,
        -8.53417e+00_prt, -8.38797e+00_prt, -8.24177e+00_prt, -8.09558e+00_prt,
        -7.94938e+00_prt, -7.80319e+00_prt, -7.65699e+00_prt, -7.51079e+00_prt,
        -7.36460e+00_prt, -7.21840e+00_prt, -7.07221e+00_prt, -6.92601e+00_prt,
        -6.77982e+00_prt, -6.63362e+00_prt, -6.48742e+00_prt, -6.34123e+00_prt,
        -6.19503e+00_prt, -6.04884e+00_prt, -5.90264e+00_prt, -5.75644e+00_prt,
        -5.61025e+00_prt, -5.46405e+00_prt, -5.31786e+00_prt, -5.17166e+00_prt,
        -5.02547e+00_prt, -4.87927e+00_prt, -4.73307e+00_prt, -4.58688e+00_prt,
        -4.44068e+00_prt, -4.29449e+00_prt, -4.14829e+00_prt, -4.00210e+00_prt,
        -3.85590e+00_prt, -3.70971e+00_prt, -3.56352e+00_prt, -3.41733e+00_prt,
        -3.27114e+00_prt, -3.12495e+00_prt, -2.97876e+00_prt, -2.83259e+00_prt,
        -2.68642e+00_prt, -2.54026e+00_prt, -2.39412e+00_prt, -2.24800e+00_prt,
        -2.10193e+00_prt, -1.95591e+00_prt, -1.80998e+00_prt, -1.66418e+00_prt,
        -1.51858e+00_prt, -1.37328e+00_prt, -1.22844e+00_prt, -1.08429e+00_prt,
        -9.41184e-01_prt, -7.99654e-01_prt, -6.60469e-01_prt, -5.24702e-01_prt,
        -3.93689e-01_prt, -2.68515e-01_prt, -1.47175e-01_prt, 0.00000e+00_prt,
        -9.12243e+00_prt, -8.97623e+00_prt, -8.83004e+00_prt, -8.68384e+00_prt,
        -8.53764e+00_prt, -8.39145e+00_prt, -8.24525e+00_prt, -8.09906e+00_prt,
        -7.95286e+00_prt, -7.80666e+00_prt, -7.66047e+00_prt, -7.51427e+00_prt,
        -7.36808e+00_prt, -7.22188e+00_prt, -7.07569e+00_prt, -6.92949e+00_prt,
        -6.78329e+00_prt, -6.63710e+00_prt, -6.49090e+00_prt, -6.34471e+00_prt,
        -6.19851e+00_prt, -6.05231e+00_prt, -5.90612e+00_prt, -5.75992e+00_prt,
        -5.61373e+00_prt, -5.46753e+00_prt, -5.32133e+00_prt, -5.17514e+00_prt,
        -5.02894e+00_prt, -4.88275e+00_prt, -4.73655e+00_prt, -4.59036e+00_prt,
        -4.44416e+00_prt, -4.29797e+00_prt, -4.15177e+00_prt, -4.00558e+00_prt,
        -3.85938e+00_prt, -3.71319e+00_prt, -3.56699e+00_prt, -3.42080e+00_prt,
        -3.27461e+00_prt, -3.12842e+00_prt, -2.98224e+00_prt, -2.83606e+00_prt,
        -2.68989e+00_prt, -2.54373e+00_prt, -2.39759e+00_prt, -2.25147e+00_prt,
        -2.10539e+00_prt, -1.95936e+00_prt, -1.81343e+00_prt, -1.66762e+00_prt,
        -1.52200e+00_prt, -1.37668e+00_prt, -1.23181e+00_prt, -1.08763e+00_prt,
        -9.44473e-01_prt, -8.02878e-01_prt, -6.63604e-01_prt, -5.27715e-01_prt,
        -3.96535e-01_prt, -2.71126e-01_prt, -1.49447e-01_prt, 0.00000e+00_prt,
        -9.12552e+00_prt, -8.97932e+00_prt, -8.83313e+00_prt, -8.68693e+00_prt,
        -8.54073e+00_prt, -8.39454e+00_prt, -8.24834e+00_prt, -8.10215e+00_prt,
        -7.95595e+00_prt, -7.80975e+00_prt, -7.66356e+00_prt, -7.51736e+00_prt,
        -7.37117e+00_prt, -7.22497e+00_prt, -7.07877e+00_prt, -6.93258e+00_prt,
        -6.78638e+00_prt, -6.64019e+00_prt, -6.49399e+00_prt, -6.34780e+00_prt,
        -6.20160e+00_prt, -6.05540e+00_prt, -5.90921e+00_prt, -5.76301e+00_prt,
        -5.61682e+00_prt, -5.47062e+00_prt, -5.32442e+00_prt, -5.17823e+00_prt,
        -5.03203e+00_prt, -4.88584e+00_prt, -4.73964e+00_prt, -4.59345e+00_prt,
        -4.44725e+00_prt, -4.30105e+00_prt, -4.15486e+00_prt, -4.00866e+00_prt,
        -3.86247e+00_prt, -3.71628e+00_prt, -3.57008e+00_prt, -3.42389e+00_prt,
        -3.27770e+00_prt, -3.13151e+00_prt, -2.98533e+00_prt, -2.83915e+00_prt,
        -2.69297e+00_prt, -2.54681e+00_prt, -2.40067e+00_prt, -2.25455e+00_prt,
        -2.10846e+00_prt, -1.96243e+00_prt, -1.81649e+00_prt, -1.67067e+00_prt,
        -1.52504e+00_prt, -1.37970e+00_prt, -1.23481e+00_prt, -1.09059e+00_prt,
        -9.47396e-01_prt, -8.05743e-01_prt, -6.66391e-01_prt, -5.30396e-01_prt,
        -3.99069e-01_prt, -2.73455e-01_prt, -1.51480e-01_prt, 0.00000e+00_prt,
        -9.12826e+00_prt, -8.98206e+00_prt, -8.83587e+00_prt, -8.68967e+00_prt,
        -8.54348e+00_prt, -8.39728e+00_prt, -8.25108e+00_prt, -8.10489e+00_prt,
        -7.95869e+00_prt, -7.81250e+00_prt, -7.66630e+00_prt, -7.52010e+00_prt,
        -7.37391e+00_prt, -7.22771e+00_prt, -7.08152e+00_prt, -6.93532e+00_prt,
        -6.78912e+00_prt, -6.64293e+00_prt, -6.49673e+00_prt, -6.35054e+00_prt,
        -6.20434e+00_prt, -6.05814e+00_prt, -5.91195e+00_prt, -5.76575e+00_prt,
        -5.61956e+00_prt, -5.47336e+00_prt, -5.32717e+00_prt, -5.18097e+00_prt,
        -5.03477e+00_prt, -4.88858e+00_prt, -4.74238e+00_prt, -4.59619e+00_prt,
        -4.44999e+00_prt, -4.30380e+00_prt, -4.15760e+00_prt, -4.01141e+00_prt,
        -3.86521e+00_prt, -3.71902e+00_prt, -3.57282e+00_prt, -3.42663e+00_prt,
        -3.28044e+00_prt, -3.13425e+00_prt, -2.98807e+00_prt, -2.84188e+00_prt,
        -2.69571e+00_prt, -2.54955e+00_prt, -2.40340e+00_prt, -2.25728e+00_prt,
        -2.11119e+00_prt, -1.96515e+00_prt, -1.81920e+00_prt, -1.67337e+00_prt,
        -1.52773e+00_prt, -1.38238e+00_prt, -1.23747e+00_prt, -1.09322e+00_prt,
        -9.49991e-01_prt, -8.08288e-01_prt, -6.68868e-01_prt, -5.32779e-01_prt,
        -4.01323e-01_prt, -2.75528e-01_prt, -1.53296e-01_prt, 0.00000e+00_prt,
        -9.13069e+00_prt, -8.98449e+00_prt, -8.83830e+00_prt, -8.69210e+00_prt,
        -8.54591e+00_prt, -8.39971e+00_prt, -8.25351e+00_prt, -8.10732e+00_prt,
        -7.96112e+00_prt, -7.81493e+00_prt, -7.66873e+00_prt, -7.52253e+00_prt,
        -7.37634e+00_prt, -7.23014e+00_prt, -7.08395e+00_prt, -6.93775e+00_prt,
        -6.79155e+00_prt, -6.64536e+00_prt, -6.49916e+00_prt, -6.35297e+00_prt,
        -6.20677e+00_prt, -6.06058e+00_prt, -5.91438e+00_prt, -5.76818e+00_prt,
        -5.62199e+00_prt, -5.47579e+00_prt, -5.32960e+00_prt, -5.18340e+00_prt,
        -5.03720e+00_prt, -4.89101e+00_prt, -4.74481e+00_prt, -4.59862e+00_prt,
        -4.45242e+00_prt, -4.30623e+00_prt, -4.16003e+00_prt, -4.01384e+00_prt,
        -3.86764e+00_prt, -3.72145e+00_prt, -3.57525e+00_prt, -3.42906e+00_prt,
        -3.28287e+00_prt, -3.13668e+00_prt, -2.99049e+00_prt, -2.84431e+00_prt,
        -2.69814e+00_prt, -2.55197e+00_prt, -2.40582e+00_prt, -2.25970e+00_prt,
        -2.11361e+00_prt, -1.96757e+00_prt, -1.82161e+00_prt, -1.67577e+00_prt,
        -1.53013e+00_prt, -1.38476e+00_prt, -1.23983e+00_prt, -1.09556e+00_prt,
        -9.52293e-01_prt, -8.10546e-01_prt, -6.71065e-01_prt, -5.34894e-01_prt,
        -4.03325e-01_prt, -2.77372e-01_prt, -1.54914e-01_prt, 0.00000e+00_prt,
        -9.13284e+00_prt, -8.98665e+00_prt, -8.84045e+00_prt, -8.69425e+00_prt,
        -8.54806e+00_prt, -8.40186e+00_prt, -8.25567e+00_prt, -8.10947e+00_prt,
        -7.96327e+00_prt, -7.81708e+00_prt, -7.67088e+00_prt, -7.52469e+00_prt,
        -7.37849e+00_prt, -7.23229e+00_prt, -7.08610e+00_prt, -6.93990e+00_prt,
        -6.79371e+00_prt, -6.64751e+00_prt, -6.50132e+00_prt, -6.35512e+00_prt,
        -6.20892e+00_prt, -6.06273e+00_prt, -5.91653e+00_prt, -5.77034e+00_prt,
        -5.62414e+00_prt, -5.47794e+00_prt, -5.33175e+00_prt, -5.18555e+00_prt,
        -5.03936e+00_prt, -4.89316e+00_prt, -4.74697e+00_prt, -4.60077e+00_prt,
        -4.45457e+00_prt, -4.30838e+00_prt, -4.16218e+00_prt, -4.01599e+00_prt,
        -3.86979e+00_prt, -3.72360e+00_prt, -3.57740e+00_prt, -3.43121e+00_prt,
        -3.28502e+00_prt, -3.13883e+00_prt, -2.99264e+00_prt, -2.84646e+00_prt,
        -2.70029e+00_prt, -2.55412e+00_prt, -2.40797e+00_prt, -2.26184e+00_prt,
        -2.11575e+00_prt, -1.96971e+00_prt, -1.82374e+00_prt, -1.67790e+00_prt,
        -1.53224e+00_prt, -1.38687e+00_prt, -1.24192e+00_prt, -1.09763e+00_prt,
        -9.54333e-01_prt, -8.12547e-01_prt, -6.73013e-01_prt, -5.36770e-01_prt,
        -4.05101e-01_prt, -2.79009e-01_prt, -1.56354e-01_prt, 0.00000e+00_prt};

    m_phot_em_table = QS_phot_em_table{phot_em_params, vals};
}

//============================================
